unit CCOWRPCBrokerCAPRI;

interface

uses Classes, Messages, SysUtils, WinProcs, WinTypes, Windows, Dialogs, Forms,
       {VA}
     CCOWRpcBroker,
     XWBut1, {RpcbEdtr,} MFunStr, Hash,
     ComObj, ActiveX, OleCtrls, trpcb,
     VERGENCECONTEXTORLib_TLB, VCEdit{support ANUVCNew variable};

type
  TCCOWRPCBrokerCAPRI = Class(TCCOWRPCBroker)
  private
    FstrServer: String;
    FstrPort: String;
    FRPCServer: String;
    FRPCPort: String;
    FRPCVolume: String;
    FRPCUCI: String;
    FUserSpecs: String;
    FstrPortHome: String;
    FstrServerHome: String;
    FstrDivisionHome: String;
    FAccessCodeHome: String;
    FVerifyCodeHome: String;
    FBSEToken: String;
    FTestMode: Boolean;
    FFHIEConnection: Boolean;
    FVerifyCode: String;
    FOnBeforeCall: TNotifyEvent;
    FOnAfterCall: TNotifyEvent;
    procedure GetServerAV;
  protected
    procedure SetConnected(Value: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Call; override;
    function BSEReconnect(aContext: String = ''): Boolean;
    function Connect2HomeServer(anAccess, aVerify, aServer, aPort: String; Is   ug: Boolean): Boolean;
    function Connect2RemoteServer(aServer, aPort: String; Is   ug: Boolean; IsShowError: Boolean = True; aStationNo: String = ''): Boolean;
    function CallServer(RPC: String = ''; RPCParams: TParams = nil; IsTry2Reconnect: Boolean = True): Boolean;
    property ANUstrServer: String read FstrServer write FstrServer;
    property ANUstrPort: String read FstrPort write FstrPort;
    property ANURPCServer: String read FRPCServer;  {read-only}
    property ANURPCPort: String read FRPCPort;  {read-only}
    property ANURPCVolume: String read FRPCVolume;  {read-only}
    property ANURPCUCI: String read FRPCUCI;  {read-only}
    property ANUUserSpecs: String read FUserSpecs write FUserSpecs;
    property ANUstrServerHome: String read FstrServerHome;  {read-only}
    property ANUstrPortHome: String read FstrPortHome;  {read-only}
    property ANUstrDivisionHome: String read FstrDivisionHome write FstrDivisionHome;
    property ANUAccessCodeHome: String read FAccessCodeHome;  {read-only}
    property ANUVerifyCodeHome: String read FVerifyCodeHome;  {read-only}
    property ANUBSEToken: String read FBSEToken write FBSEToken;
    property ANUVerifyCode: String read FVerifyCode write FVerifyCode;
    property ANUTestMode: Boolean read FTestMode write FTestMode;
    property ANUFHIEConnection: Boolean read FFHIEConnection write FFHIEConnection;
  published
    property OnBeforeCall: TNotifyEvent read FOnBeforeCall write FOnBeforeCall;
    property OnAfterCall: TNotifyEvent read FOnAfterCall write FOnAfterCall;
  end;

var
  ANURemoteProcedureCallInProgress: Boolean;
  ESSOVersion: Boolean;
  ANULocalMods: Boolean;

procedure AuthenticateUser(ConnectingBroker: TCCOWRPCBrokerCAPRI);
procedure Register;

implementation
{$R CCOWRPCBrokerCAPRI.res}

{--------------------------------------------------------------------------------
 Uses clause copied from CCOWRPCBroker unit to support AuthenticateUser procedure
 --------------------------------------------------------------------------------}
uses
  Loginfrm, RpcbErr, WSockc, SelDiv{p8}, RpcSLogin{p13}, fRPCBErrMsg,
  CCOW_const;

{---------------------------------------------------------------------------
 Var copied from CCOWRPCBroker unit to support AuthenticateUser procedure
 ---------------------------------------------------------------------------}
var
  CCOWToken: String;
  Domain: String;
  PassCode1: String;
  PassCode2: String;

constructor TCCOWRPCBrokerCAPRI.Create(AOwner: TComponent);
begin
  inherited;
  FstrServer:= '';
  FstrPort:= '';
  FRPCServer:= '';
  FRPCPort:= '';
  FRPCVolume:= '';
  FRPCUCI:= '';
  FstrPortHome:= '';
  FstrServerHome:= '';
  FAccessCodeHome:= '';
  FVerifyCodeHome:= '';
  FBSEToken := '';
  FTestMode := False;
end;

destructor TCCOWRPCBrokerCAPRI.Destroy;
begin
  inherited;
end;

{--------------------------- CallServer --------------------------------
 This function acts as an RPC call wrapper
 Input:
   RPC : String - Remote Procedure Call name
   RPCParams: TParams - Populated RPC parameters
   IsTry2Reconnect: Boolean - Indicates that function should attempt to
                              reconnect when the RPCBroker is not connected.

 Output:
   Boolean result returns TRUE when RPCBroker is connected and RPC call
   succeeds; otherwise returns FALSE.
 -----------------------------------------------------------------------}
function TCCOWRPCBrokerCAPRI.CallServer(RPC: String = ''; RPCParams: TParams = nil; IsTry2Reconnect: Boolean = True): Boolean;
begin
  Result := False;
  if (socket = -1) or (not Connected) then
  begin
    if IsTry2Reconnect and (not BSEReconnect) then
    begin
      ShowMessage('CAPRI lost connection to server and '
                  + #13#10 + 'failed to re-establish the connection');
      Exit;   {Short-circuit}
    end;
  end;

  if Connected then
  begin
    try
      if RPC <> '' then
        RemoteProcedure := RPC;
      if RPCParams <> nil then
        Param.Assign(RPCParams);
      Call;
      Result := True;
    except on E: EBrokerError do
      begin
        ShowMessage('CAPRI failed to communicate with VistA'
                     + #13#10 + 'Error: ' + E.Message);
        Connected := False;
      end;
    end;
  end;
end;

{--------------------------- Connect2HomeServer ------------------------
 This function attempts to connect to the 'home' server and retrieves a
 BSE token for use in remote connections.  Silent Login is used when
 the Access/Verify parameters are populated.

 Results:  True on success; False on failure.
 -------------------------------------------------------------------------}
function TCCOWRPCBrokerCAPRI.Connect2HomeServer(anAccess, aVerify, aServer, aPort: String;
                                                Is   ug: Boolean): Boolean;
var
  SaveContext: String;
  IsKernelLogin: Boolean;
begin
  Result := False;
  Server := aServer;
  ListenerPort := StrToInt(aPort);
  IsKernelLogin := KernelLogIn;
     ugMode := Is   ug;

  {setup Silent Login when credentials are received}
  if (anAccess <> '') then
  begin
    if(aVerify <> '') or (Piece(anAccess, ';', 2) <> '') then
    begin
      LogIn.Mode := lmAVCodes;
      LogIn.AccessCode := anAccess;
      LogIn.VerifyCode := aVerify;
      LogIn.Division := FstrDivisionHome;
      KernelLogIn := False;
      LogIn.PromptDivision := False;
    end;
  end;

  try
    Connected := True;
    if Connected then
    begin
      {protect CurrentContext before setting to XUS SIGNON.  XUS SIGNON
       contains XUS SET VISITOR and XUS DIVISION SET (called by SelDiv.SetDiv)}
      SaveContext := CurrentContext;
      CreateContext('XUS SIGNON');

      {force Division because silent login skips it}
      if LogIn.Division <> '' then
        SelDiv.SetDiv(Login.Division, TRPCBroker(Self));

      //Temporarily attach XUS SET VISITOR to DVBAB CAPRI GUI per
      //Joel Ivey of the Broker development team  - rpm 3/11/09
      CreateContext('DVBA CAPRI GUI');
      {retrieve BSE token}
      RemoteProcedure := 'XUS SET VISITOR';
      Call;
      FBSEToken := Results[0];

      {restore CurrentContext}
      CreateContext(SaveContext);

      {initialize 'Home' properties on first login}
      if FstrServerHome = '' then
      begin
        FstrPortHome:= IntToStr(ListenerPort);
        FstrServerHome:= Server;
        FstrDivisionHome:= Piece(User.Division, '^', 3);
        FAccessCodeHome:= Login.AccessCode;
        FVerifyCodeHome:= Login.VerifyCode;
      end;
    end;
    Result := True;
  except
    On E: EBrokerError do
      begin
        if E.Code <> XWB_BADSIGNON then
          ShowMessage('Could not connect to home server.'
                      + #13#10 + 'Error: ' + E.Message);
        Connected := False;
      end;
  end;
  KernelLogIn := IsKernelLogin;
end;

{--------------------------- Connect2RemoteServer ------------------------
 This function attempts to connect to a server using a BSE token.

 Results:  True on success; False on failure.
 -------------------------------------------------------------------------}
function TCCOWRPCBrokerCAPRI.Connect2RemoteServer(aServer, aPort: String;
                                                  Is   ug: Boolean;
                                                  IsShowError: Boolean = True;
                                                  aStationNo: String = ''): Boolean;
const
  Key = 'Never put off until tomorrow what you can do the day after tomorrow. -- Mark Twain';
  TestKey = 'Nothing is impossible for the man who does not have to do it himself.';
  FHIEKey = 'Courage is resistance to fear, mastery of fear - not absence of fear. -- Mark Twain';
begin
  Result := False;
  Server := aServer;
  ListenerPort := StrToInt(aPort);
     ugMode := Is   ug;
  if (NOT EssoVersion) and FFHIEConnection then
    SecurityPhrase := FHIEKey + '^' + FBSEToken + '^' + aStationNo + '^' + FstrPortHome
  else
  begin
    if FTestMode then
      SecurityPhrase := TestKey + '^' + FBSEToken
    else
      SecurityPhrase := Key + '^' + FBSEToken;
  end;
  try
    Connected := True;
    if Connected then
    begin
      Result := True;
      FBSEToken := '';  //clear the token
    end;
  except
    on E: EBrokerError do
      begin
        if IsShowError then
          ShowMessage('Could not connect to remote server.'
                      + #13#10 + 'Error: ' + E.Message);
        Connected := False;
      end;
  end;
end;


{-------------------- GetServerAV --------------------
 This procedure is called by AuthenticateUser to
 populate the AV codes, server, volume, uci, and port
 -----------------------------------------------------}
procedure TCCOWRPCBrokerCAPRI.GetServerAV;
begin
  LogIn.AccessCode := frmSignon.accessCode.Text;
  LogIn.VerifyCode := frmSignon.verifyCode.Text;
  if VCEdit.ANUVCNew <> '' then
  begin
    Login.VerifyCode := VCEdit.ANUVCNew;
    VCEdit.ANUVCNew := '';
  end;
  FRPCServer := frmSignon.lblServer.Caption;
  FRPCVolume := frmSignon.lblVolume.Caption;
  FRPCUCI := frmSignon.lblUCI.Caption;
  FRPCPort := frmSignon.lblPort.Caption;
end;

{---------------- BSEReconnect ------------------------
 This function attempts to reconnect the RPCBroker when
 the connection is lost.
 ------------------------------------------------------}
function TCCOWRPCBrokerCAPRI.BSEReconnect(aContext: String = ''): Boolean;
var
  SaveClearParmeters, SaveClearResults: boolean;
  SaveParam: TParams;
  SaveRemoteProcedure: string;
  SaveResults: TStrings;
  SaveKernelLogin: boolean;
  SaveVistaLogin: TVistaLogin;
begin
  Result := False;

  SaveParam := TParams.Create(nil);
  SaveParam.Assign(Param);                  //save off settings
  SaveRemoteProcedure := RemoteProcedure;
  SaveResults := Results;
  SaveClearParmeters := ClearParameters;
  SaveClearResults := ClearResults;
  ClearParameters := True;
  ClearResults := True;
  SaveKernelLogin := KernelLogin;
  SaveVistaLogin := Login;
  try
    if (FstrServerHome = '') or
       (FServer = FstrServerHome) then
        Connect2HomeServer(FAccessCodeHome, FVerifyCodeHome,
                           FstrServerHome, FstrPortHome,    ugMode)
    else
    begin
      if FBSEToken = '' then  //must return "home" and retrieve token
        Connect2HomeServer(FAccessCodeHome, FVerifyCodeHome,
                           FstrServerHome, FstrPortHome,    ugMode);
      Connect2RemoteServer(FstrServer, FstrPort,    ugMode);
    end;
    if Connected then
    begin
      if aContext <> '' then
        Result := CreateContext(aContext)
      else
        Result := CreateContext('DVBA CAPRI GUI');
    end;
  finally
    ClearParameters := SaveClearParmeters;
    ClearResults := SaveClearResults;
    Param.Assign(SaveParam);                  //restore settings
    SaveParam.Free;
    RemoteProcedure := SaveRemoteProcedure;
    Results := SaveResults;
    FKernelLogin := SaveKernelLogin;
    FLogin := SaveVistaLogin;
  end;
end;

{------------------------ AuthenticateUser --------------------------------
 This procedure is copied directly from the CCOWRPCBroker unit and contains
 the released code base Version 1.1 Patch 47 (Jun. 17, 2008) except where
 delimited by the comments:
      //DVBA CAPRI modification begin
      //DVBA CAPRI modification end
---------------------------------------------------------------------------}
procedure AuthenticateUser(ConnectingBroker: TCCOWRPCBrokerCAPRI);
var
  SaveClearParmeters, SaveClearResults: boolean;
  SaveParam: TParams;
  SaveRemoteProcedure, SaveRpcVersion: string;
  SaveResults: TStrings;
  blnSignedOn: boolean;
  SaveKernelLogin: boolean;
  SaveVistaLogin: TVistaLogin;
  OldExceptionHandler: TExceptionEvent;
  OldHandle: THandle;
begin
  With ConnectingBroker do
  begin
    SaveParam := TParams.Create(nil);
    SaveParam.Assign(Param);                  //save off settings
    SaveRemoteProcedure := RemoteProcedure;
    SaveRpcVersion := RpcVersion;
    SaveResults := Results;
    SaveClearParmeters := ClearParameters;
    SaveClearResults := ClearResults;
    ClearParameters := True;                  //set'em as I need'em
    ClearResults := True;
    SaveKernelLogin := KernelLogin;     //  p13
    SaveVistaLogin := Login;            //  p13
  end;

  blnSignedOn := False;                       //initialize to bad sign-on

  if ConnectingBroker.AccessVerifyCodes <> '' then   // p13 handle as AVCode single signon
  begin
    ConnectingBroker.Login.AccessCode := Piece(ConnectingBroker.AccessVerifyCodes, ';', 1);
    ConnectingBroker.Login.VerifyCode := Piece(ConnectingBroker.AccessVerifyCodes, ';', 2);
    ConnectingBroker.Login.Mode := lmAVCodes;
    ConnectingBroker.KernelLogIn := False;
  end;

    //CCOW start
    if ConnectingBroker.KernelLogIn and (not (ConnectingBroker.Contextor = nil)) then
    begin
      CCOWtoken := ConnectingBroker.GetCCOWtoken(ConnectingBroker.Contextor);
      if length(CCOWtoken)>0 then
        begin
          ConnectingBroker.FKernelLogIn := false;
          ConnectingBroker.Login.Mode := lmAppHandle;
          ConnectingBroker.Login.LogInHandle := CCOWtoken;
        end;
     end;
     //CCOW end
   //CCOW Start                                // p13  following section for silent signon
  if not ConnectingBroker.FKernelLogIn then
    if ConnectingBroker.FLogin <> nil then     //the user.  vistalogin contains login info
    begin
      blnsignedon := SilentLogin(ConnectingBroker);    // RpcSLogin unit
      if not blnSignedOn then
      begin     //Switch back to Kernel Login
        ConnectingBroker.FKernelLogIn := true;
        ConnectingBroker.Login.Mode := lmAVCodes;
      end;
    end;
   //CCOW end

  if ConnectingBroker.FKernelLogIn then
  begin   //p13
    CCOWToken := '';  //  061201 JLI if can't sign on with Token clear it so can get new one
    if Assigned(Application.OnException) then
      OldExceptionHandler := Application.OnException
    else
      OldExceptionHandler := nil;
    Application.OnException := TfrmErrMsg.RPCBShowException;
    frmSignon := TfrmSignon.Create(Application);
    try

  //    ShowApplicationAndFocusOK(Application);
      OldHandle := GetForegroundWindow;

      //DVBA CAPRI modification begin
      if ConnectingBroker.FSecurityPhrase = '' then
      //DVBA CAPRI modification end
        SetForegroundWindow(frmSignon.Handle);
      PrepareSignonForm(ConnectingBroker);
      if SetUpSignOn then                       //SetUpSignOn in loginfrm unit.
      begin                                     //True if signon needed

        if frmSignOn.lblServer.Caption <> '' then
        begin
          frmSignOn.ShowModal;                    //do interactive logon   // p13
          if frmSignOn.Tag = 1 then               //Tag=1 for good logon
            blnSignedOn := True;                   //Successfull logon
        end
      end
      else                                      //False when no logon needed
        blnSignedOn := NoSignOnNeeded;          //Returns True always (for now!)
      if blnSignedOn then                       //P6 If logged on, retrieve user info.
      begin
        GetBrokerInfo(ConnectingBroker);
        if not SelDiv.ChooseDiv('',ConnectingBroker) then
        begin
          blnSignedOn := False;//P8
          {Select division if multi-division user.  First parameter is 'userid'
          (DUZ or username) for future use. (P8)}
          ConnectingBroker.Login.ErrorText := 'Failed to select Division';  // p13 set some text indicating problem
        end;
      end;
      //DVBA CAPRI modification begin
      if blnSignedOn then
        ConnectingBroker.GetServerAV; //Populate AV codes, server, vol, uci, port
      //DVBA CAPRI modification end

      //DVBA CAPRI modification begin
      if ConnectingBroker.FSecurityPhrase = '' then
      //DVBA CAPRI modification end
        SetForegroundWindow(OldHandle);
    finally
      frmSignon.Free;
//      frmSignon.Release;                        //get rid of signon form

//      if ConnectingBroker.Owner is TForm then
//        SetForegroundWindow(TForm(ConnectingBroker.Owner).Handle)
//      else
//        SetForegroundWindow(ActiveWindow);

      //DVBA CAPRI modification begin
      if ConnectingBroker.FSecurityPhrase = '' then
      //DVBA CAPRI modification end
        ShowApplicationAndFocusOK(Application);
    end ; //try
    if Assigned(OldExceptionHandler) then
      Application.OnException := OldExceptionHandler;
   end;   //if kernellogin
                                                 // p13  following section for silent signon
  if (not ConnectingBroker.KernelLogIn) and (not blnsignedon) then     // was doing the signon twice if already true
    if ConnectingBroker.Login <> nil then     //the user.  vistalogin contains login info
      blnsignedon := SilentLogin(ConnectingBroker);    // RpcSLogin unit
  if not blnsignedon then
  begin
//    ConnectingBroker.Login.FailedLogin(ConnectingBroker.Login);
    TXWBWinsock(ConnectingBroker.XWBWinsock).NetworkDisconnect(ConnectingBroker.Socket);
  end
  else
    GetBrokerInfo(ConnectingBroker);

  //reset the Broker
  with ConnectingBroker do
  begin
    ClearParameters := SaveClearParmeters;
    ClearResults := SaveClearResults;
    Param.Assign(SaveParam);                  //restore settings
    SaveParam.Free;
    RemoteProcedure := SaveRemoteProcedure;
    RpcVersion := SaveRpcVersion;
    Results := SaveResults;
    FKernelLogin := SaveKernelLogin;         // p13
    FLogin := SaveVistaLogin;                // p13
  end;

  if not blnSignedOn then                     //Flag for unsuccessful signon.
    TXWBWinsock(ConnectingBroker.XWBWinsock).NetError('',XWB_BadSignOn);               //Will raise error.

end;

{------------------------ SetConnected ------------------------------------
This method procedure is copied directly from the CCOWRPCBroker unit in the
released code base Version 1.1 Patch 47 (Jun. 17, 2008) and contains no
modifications.  The procedure exists only to support the local version of
AuthenticateUser.
---------------------------------------------------------------------------}
procedure TCCOWRPCBrokerCAPRI.SetConnected(Value: Boolean);
var
  BrokerDir, Str1, Str2, Str3 :string;
  RPCBContextor: TContextorControl;
begin
  RPCBError := '';
  Login.ErrorText := '';
  if (Connected <> Value) and not(csReading in ComponentState) then begin
    if Value and (FConnecting <> Value) then begin                 {connect}
      FSocket := ExistingSocket(Self);
      FConnecting := True; // FConnected := True;
      try
        if FSocket = 0  then
        begin
          {Execute Client Agent from directory in Registry.}
          BrokerDir := ReadRegData(HKLM, REG_BROKER, 'BrokerDr');
          if BrokerDir <> '' then
            ProcessExecute(BrokerDir + '\ClAgent.Exe', sw_ShowNoActivate)
          else
            ProcessExecute('ClAgent.Exe', sw_ShowNoActivate);
          if    ugMode and (not OldConnectionOnly) then
          begin
            Str1 := 'Control of debugging FOR UCX OR NON-CALLBACK CONNECTIONS has been moved from the client to the server. To start a    ug session, do the following:'+#13#10#13#10;
            Str2 := '1. On the server, set initial breakpoints where desired.'+#13#10+'2. DO DEBUG^XWBTCPM.'+#13#10+'3. Enter a unique Listener port number (i.e., a port number not in general use).'+#13#10;
            Str3 := '4. Connect the client application using the port number entered in Step #3.';
            ShowMessage(Str1 + Str2 + Str3);
          end;
          TXWBWinsock(XWBWinsock).IsBackwardsCompatible := IsBackwardCompatibleConnection;
          TXWBWinsock(XWBWinsock).OldConnectionOnly := OldConnectionOnly;
          FSocket := TXWBWinsock(XWBWinsock).NetworkConnect(   ugMode, FServer,
                                    ListenerPort, FRPCTimeLimit);
              //  060919 Prefix added to handle multiple brokers including old and new
              Prefix := TXWBWinsock(XWBWinsock).Prefix;
              FIsNewStyleConnection := TXWBWinsock(XWBWinsock).IsNewStyle;
          AuthenticateUser(Self);
          StoreConnection(Self);  //MUST store connection before CreateContext()
          //CCOW start
          if (FContextor <> nil) and (length(CCOWtoken) = 0) then
          begin
          //Get new CCOW token
            CCOWToken := GetCCOWHandle(Self);
            if Length(CCOWToken) > 0 then
            begin
              try
                RPCBContextor := TContextorControl.Create(Application);
                RPCBContextor.Run('BrokerLoginModule#', PassCode1+PassCode2, TRUE, '*');
                CCOWsetUser(user.name, CCOWToken, Domain, user.Vpid, RPCBContextor);  //Clear token
                FCCOWLogonIDName := CCOW_LOGON_ID;
                FCCOWLogonIdValue := Domain;
                FCCOWLogonName := CCOW_LOGON_NAME;
                FCCOWLogonNameValue := user.name;
                if user.name <> '' then
                  FWasUserDefined := True;
                FCCOWLogonVpid := CCOW_LOGON_VPID;
                FCCOWLogonVpidValue := user.Vpid;
                RPCBContextor.Free;
                RPCBContextor := nil;
              except
                ShowMessage('Problem with Contextor.Run');
                FreeAndNil(RPCBContextor);
              end;
            end;   // if Length(CCOWToken) > 0
          end;  //if
          //CCOW end
          FPulse.Enabled := True; //P6 Start heartbeat.
          CreateContext('');      //Closes XUS SIGNON context.
        end
        else
        begin                     //p13
          StoreConnection(Self);
          FPulse.Enabled := True; //p13
        end;                      //p13
        FConnected := True;         // jli mod 12/17/01
        FConnecting := False;
      except
        on E: EBrokerError do begin
          if E.Code = XWB_BadSignOn then
            TXWBWinsock(XWBWinsock).NetworkDisconnect(FSocket);
          FSocket := 0;
          FConnected := False;
          FConnecting := False;
          FRPCBError := E.Message;               // p13  handle errors as specified
          if Login.ErrorText <> '' then
            FRPCBError := E.Message + chr(10) + Login.ErrorText;
          if Assigned(FOnRPCBFailure) then       // p13
            FOnRPCBFailure(Self)                 // p13
          else if ShowErrorMsgs = semRaise then
            Raise;                               // p13
//          raise;   {this is where I would do OnNetError}
        end{on};
      end{try};
    end{if}
    else if not Value then
    begin                           //p13
      FConnected := False;          //p13
      FPulse.Enabled := False;      //p13
      if RemoveConnection(Self) = NoMore then begin
        {FPulse.Enabled := False;  ///P6;p13 }
        TXWBWinsock(XWBWinsock).NetworkDisconnect(Socket);   {actually disconnect from server}
        FSocket := 0;                {store internal}
        //FConnected := False;      //p13
      end{if};
    end; {else}
  end{if};
end;

procedure TCCOWRPCBrokerCAPRI.Call;
begin
  if ANURemoteProcedureCallInProgress and (not FConnecting) then
    Exit;
  ANURemoteProcedureCallInProgress := True;
  try
    if assigned(OnBeforeCall) then
      OnBeforeCall(Self);
    inherited;
    if assigned(OnAfterCall) then
      OnAfterCall(Self);
  finally
    ANURemoteProcedureCallInProgress := False;
  end;
end;

{------------------------ Register --------------------
 Register the component in the component palette
 ------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('Kernel', [TCCOWRPCBrokerCAPRI]);
end;

end.
