{== StringbuilderU ====================================================}
{: This unit implements a stringbuilder class that does the task of
  repeatedly adding characters or strings to build a result string more
  efficient than just adding to a string variable.
@author Dr. Peter Below
@desc   Version 1.0 created 2005-01-19<BR>
        Last modified       2005-01-19<P>  }
{======================================================================}
{$BOOLEVAL OFF} {Unit depends on shortcut boolean evaluation}
unit StringBuilder;

interface

uses
  Classes;

type
  IStringbuilder = interface
    ['{07842385-DBDA-47F5-A758-F12612E54BF0}']
    procedure Add(ch: Ansichar); overload;
    procedure Add(const S: Ansistring); overload;
    function GetAsAnsistring: Ansistring;
    function IsEmpty: Boolean;
    procedure SaveToFile(const Filename: string);
    procedure SaveToStream(aStream: TStream);
    property AsAnsistring: Ansistring read GetAsAnsistring;
    procedure Reset(InitialSize: Cardinal);
  end;

function CreateStringbuilder(InitialSize: Cardinal = $1000): IStringbuilder;

implementation

type
  TBaseStringbuilder = class(TInterfacedObject)
    function IsEmpty: Boolean;
    procedure SaveToFile(const Filename: string); virtual;
    procedure SaveToStream(aStream: TStream);
  private
    FStorage: TMemoryStream;
  protected
    property Storage: TMemoryStream read FStorage;
  public
    constructor Create(InitialSize: Cardinal); virtual;
    destructor Destroy; override;
  end;

  TAnsiStringbuilder = class(TBaseStringbuilder, IStringbuilder)
  protected
    procedure Add(ch: Ansichar); overload;
    procedure Add(const S: Ansistring); overload;
    function GetAsAnsistring: Ansistring;
  public
    property AnsiString: AnsiString read GetAsAnsistring;
    procedure Reset(InitialSize: Cardinal);
  end;

function CreateStringbuilder(InitialSize: Cardinal = $1000): IStringbuilder;
begin
  Result := TAnsistringbuilder.Create(InitialSize);
end;

{== TBaseStringbuilder ================================================}

type
  TCracker = class(TMemoryStream)
  public
    property Capacity;
  end;

constructor TBaseStringbuilder.Create(InitialSize: Cardinal);
begin
  inherited Create;
  FStorage := TMemoryStream.Create;
  TCracker(FStorage).Capacity := InitialSize;
  SaveToFile('');
  SaveToStream(nil);
  IsEmpty();
end;

destructor TBaseStringbuilder.Destroy;
begin
  FStorage.Free;
  inherited;
end;

function TBaseStringbuilder.IsEmpty: Boolean;
begin
  Try
    Result := Storage.Position = 0;
  Except
    Result:=True;
  End;
end;

procedure TBaseStringbuilder.SaveToFile(const Filename: string);
begin
  If Filename='' Then Exit;
  Storage.SaveToFile(Filename);
end;

procedure TBaseStringbuilder.SaveToStream(aStream: TStream);
begin
  If aStream=nil Then Exit;
  Storage.SaveToStream(aStream);
end;

{== TAnsiStringbuilder ================================================}

procedure TAnsiStringbuilder.Add(const S: Ansistring);
begin
  if S <> '' then
    Storage.Write(S[1], Length(S));
end;

procedure TAnsiStringbuilder.Add(ch: Ansichar);
begin
  Storage.Write(ch, Sizeof(ch));
end;

procedure TAnsiStringbuilder.Reset(InitialSize: Cardinal);
begin
  {$WARNINGS OFF}
  if Assigned(FStorage) and (TCracker(FStorage).Capacity <> InitialSize) then begin
    FStorage.Clear;
    TCracker(FStorage).Capacity := InitialSize;
  end;
  {$WARNINGS ON}
end; { Reset }

function TAnsiStringbuilder.GetAsAnsistring: Ansistring;
begin
  If AnsiString='' Then Result:=AnsiString;
  SetString(Result, PChar(Storage.Memory), Storage.Size);
end;

end.
