unit UnitExamInstanceBase;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  CPWMObjects,
  StdCtrls,
  ORCtrls,
  ExtCtrls,
  CPWMRadioGroup,
  ColorCapriBtn;

type
  TFormExamInstanceBase = class(TForm)
  private

    { Private declarations }
  public
    FormLoading: boolean;
    UpdatingSiblings: boolean;
    procedure xGenericChange(Sender: TObject);
    procedure xButtonGenericClick(Sender: TObject);
    procedure xGenericHint(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  end;

var
  FormExamInstanceBase: TFormExamInstanceBase;
  currentExam: TCPWMExam;

implementation

uses pncscoderun,
  pncsrunreport,
  UnitMemoText,
  Logger,
  hintdisplay,
  main,
  PNCSMain,
  UnitSiblings;

{$R *.DFM}

procedure TFormExamInstanceBase.xGenericChange(Sender: TObject);
begin
  if FormLoading then
    exit; // Disallow this event while form is loading

  if UpdatingSiblings then // Ignore changes that are happening during the update siblings
    exit;

  {rpm 4/1/09 - Defect #2165 - replace timestamp/tag method of determining
                replication status of control with TList.}
  UpdatingControls := TList.Create;
  try
    UpdatingSiblings := true;
    //rpm 4/1/09 - Defect #2165
    //(Sender as TControl).Tag := DateTimeToTimeStamp(now).Time;
    UpdatingControls.Add(TControl(Sender));
    UpdateSiblings2(pncsForm.xpanelbasecontrol, Sender as TControl);
  finally
    UpdatingSiblings := false;
    FreeAndNil(UpdatingControls);    //rpm 4/1/09 - Defect #2165
  end;
  if PNCSEvent = True then
    PNCSEventRepeat := True;
  Log('pncsEvent from ' + (Sender as TControl).Name, leInformation, 'TFormExamBase', 'xGenericChange', 246);
end;

procedure TFormExamInstanceBase.xGenericHint(Sender: TObject; Shift: TShiftState; X, Y: Integer);

  function actualposition(tempControl: tControl; which: string): integer;
  var
    pos: integer;
    last: integer;
  begin
    which := uppercase(which);
    last:=0;
    pos:=0;
    if which = 'LEFT' then begin
      pos := tempControl.left;
      if tempControl.Parent <> nil then
        repeat
          tempControl := tempControl.Parent;
          if tempControl <> nil then begin
            last := tempControl.left;
            pos := pos + last;
          end;
        until tempControl = nil;
      pos := pos - last // remove off main form's top,left properties
    end;
    if which = 'TOP' then begin
      pos := tempControl.top;
      if tempControl.Parent <> nil then
        repeat
          tempControl := tempControl.Parent;
          if tempControl <> nil then begin
            last := tempControl.top;
            pos := pos + last;
          end;
        until tempControl = nil;
      pos := pos - last // remove off main form's top,top properties
    end;
    result := pos;
  end;

begin
  if pncsform.xbuttondetails.visible = true then
    exit;

  if uppercase(Sender.ClassName) = 'TMEMO' then begin

    if Assigned(TempLastMemo) and (PNCSForm.xbuttonDetails.visible = false) then begin
      PNCSForm.xbuttonDetails.top := actualposition(tempLastMemo, 'top') + templastmemo.height + 2;
      PNCSForm.xbuttonDetails.left := actualposition(tempLastMemo, 'left');
      PNCSForm.xbuttonDetails.width := tempLastMemo.Width - 2;
      PNCSForm.xbuttonDetails.BringToFront;
      PNCSForm.xbuttonDetails.Visible := true;
      hOldWnd := WindowFromPoint(Mouse.CursorPos);
    end;
  end;
end;

procedure TFormExamInstanceBase.xButtonGenericClick(Sender: TObject);
var
  tempTButton: TButton;
  tempTColorCAPRIBtn: TColorCAPRIBtn;
begin
  if FormLoading then
    exit; // Disallow this event while form is loading

  if UpdatingSiblings then begin
    exit;
  end;
  UpdatingControls := TList.Create;  //rpm 4/1/09 - Defect #2165
  try
    UpdatingSiblings := true;

    if Sender is TMemo then begin
      if assigned(formCustomHint) then begin
        formCustomHint.TimerShowHint.Enabled := False;
        formCustomHint.timerShowHint.Interval := 999999999;
        if formCustomHint.Visible = true then
          formCustomHint.Close;
      end;
    end;

    if Sender is TColorCAPRIBtn then begin
      tempTColorCAPRIBtn := Sender as TColorCAPRIBtn;
      ButtonClicked(tempTColorCAPRIBtn.name); // in pncsCodeRun module
    end
    else
      if Sender is TButton then begin
        tempTButton := Sender as TButton;
        ButtonClicked(tempTButton.name); // in pncsCodeRun module
      end
      else
        if Sender is TListBox then begin
          //rpm 4/1/09 - Defect #2165
          //(Sender as TControl).Tag := DateTimeToTimeStamp(now).Time;
          UpdatingControls.Add(Sender as TControl);
          UpdateSiblings2(pncsForm.xpanelbasecontrol, Sender as TControl);
        end
        else
          if Sender is TORListBox then begin
            //rpm 4/1/09 - Defect #2165
            //(Sender as TControl).Tag := DateTimeToTimeStamp(now).Time;
            UpdatingControls.Add(Sender as TControl);
            UpdateSiblings2(pncsForm.xpanelbasecontrol, Sender as TControl);
          end
          else
            if Sender is TRadioGroup then begin
              //rpm 4/1/09 - Defect #2165
              //(Sender as TControl).Tag := DateTimeToTimeStamp(now).Time;
              UpdatingControls.Add(Sender as TControl);
              UpdateSiblings2(pncsForm.xpanelbasecontrol, Sender as TControl);
            end
            else
              if Sender is TCPWMRadioGroup then begin
                //rpm 4/1/09 - Defect #2165
                //(Sender as TControl).Tag := DateTimeToTimeStamp(now).Time;
                UpdatingControls.Add(Sender as TControl);
                UpdateSiblings2(pncsForm.xpanelbasecontrol, Sender as TControl);
              end
              else
                if Sender is TImage then begin
                  //rpm 4/1/09 - Defect #2165
                  //(Sender as TControl).Tag := DateTimeToTimeStamp(now).Time;
                  UpdatingControls.Add(Sender as TControl);
                  UpdateSiblings2(pncsForm.xpanelbasecontrol, Sender as TControl);
                end
                else
                  if Sender is TCheckBox then begin
                    //rpm 4/1/09 - Defect #2165
                    //(Sender as TControl).Tag := DateTimeToTimeStamp(now).Time;
                    UpdatingControls.Add(Sender as TControl);
                    UpdateSiblings2(pncsForm.xpanelbasecontrol, Sender as TControl);
                  end;
  finally
    UpdatingSiblings := false;
    FreeAndNil(UpdatingControls);  //rpm 4/1/09 - Defect #2165
  end;
  if PNCSEvent = True then
    PNCSEventRepeat := True;
end;

end.

