unit WSSE;

(*
   Delphi does not natively support web service security headers.  This unit
   is a modification of code provided by Jean-Marie Babet in a response on the
   Embarcadero discussion forum to a post entitled "WS with WS-Security and
   Delphi 2007" from Oct, 5, 2010.

   Direct quote from post:
   "This issue has come up before and back then I put together a file, WSSE.pas,
    that manually constructed the types. I'm attaching a copy below. You will
    probably need to tweak it to match your needs: I put this together to match
    what someone wanted and it was a little different from what you're after but
    similar.
    Please, let me know if I can be of any help.

    Cheers,

    Bruneau"

   URL: https://forums.embarcadero.com/thread.jspa?threadID=44370
 *)

interface

uses SysUtils, InvokeRegistry, SOAPHTTPClient, Types, XSBuiltIns;

const
  IS_OPTN=$0001;
  IS_ATTR=$0010;
  IS_TEXT=$0020;
  IS_REF =$0080;
  IS_QUAL=$0100;

type

  AttributedString=class(TRemotable)
  private
    FText: WideString;
    FId: WideString;
    FId_Specified:boolean;
    procedure SetId(Index:Integer; const AId: WideString);
    function Id_Specified(Index:Integer):boolean;
  published
    property Text:WideString Index (IS_TEXT) read FText write FText;
    property Id:WideString Index (IS_ATTR or IS_OPTN) read FId write SetId
stored Id_Specified;
  end;

  PasswordString=class(AttributedString)
  private
    FType_:WideString;
    FType__Specified:boolean;
    procedure SetType_(Index:Integer; const AWideString:WideString);
    function Type__Specified(Index:Integer):boolean;
  published
    property Type_:WideString Index (IS_ATTR or IS_OPTN) read FType_ write
SetType_ stored Type__Specified;
  end;


  //************************************************************************ //
  // XML       : UsernameTokenType, global, <complexType>
  // Namespace : http://schemas.xmlsoap.org/ws/2002/07/secext
  //************************************************************************ //
  UsernameTokenType = class(TRemotable)
  private
    FId: WideString;
    FId_Specified: boolean;
    FUsername: WideString;
    FPassword: PasswordString;
    procedure SetId(Index: Integer; const AId: WideString);
    function  Id_Specified(Index: Integer): boolean;
  public
    destructor Destroy; override;
  published
    property Id:       WideString  Index (IS_ATTR or IS_OPTN) read FId write
SetId stored Id_Specified;
    property Username: WideString  read FUsername write FUsername;
    property Password: PasswordString  read FPassword write FPassword;
  end;


  //************************************************************************ //
  // XML       : UsernameToken, global, <element>
  // Namespace : http://schemas.xmlsoap.org/ws/2002/07/secext
  //************************************************************************ //
  UsernameToken = class(UsernameTokenType)
  private
  published
  end;

  Security=class(TSOAPHeader)
  private
    FUserNameToken:UserNameToken;
  public
    destructor Destroy; override;
  published
    property UsernameToken:UsernameToken index (IS_REF) read FUserNameToken
write FUserNameToken;
  end;

implementation

procedure AttributedString.SetId(Index:Integer; const AId: WideString);
begin
  FId:=AId;
  FId_Specified:=True;
end;

function AttributedString.Id_Specified(Index:Integer):boolean;
begin
  Result:=FId_Specified;
end;

procedure PasswordString.SetType_(Index:Integer; const
AWideString:WideString);
begin
  FType_:=AWideString;
  FType__Specified:=True;
end;

function PasswordString.Type__Specified(Index:Integer):boolean;
begin
  Result:=FType__Specified;
end;

destructor Security.Destroy;
begin
  FreeAndNIL(FUserNameToken);
  inherited Destroy;
end;

const
  NS_SECEXT='http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd';
  NS_UTILITY='http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd';

{ UsernameTokenType }

destructor UsernameTokenType.Destroy;
begin
  SysUtils.FreeAndNil(FPassword);
  inherited Destroy;
end;

procedure UsernameTokenType.SetId(Index: Integer; const AId: WideString);
begin
  FId := AId;
  FId_Specified := True;
end;

function UsernameTokenType.Id_Specified(Index: Integer): boolean;
begin
  Result := FId_Specified;
end;


initialization
  RemClassRegistry.RegisterXSClass(Security, NS_SECEXT, 'Security');
  RemClassRegistry.RegisterXSClass(UsernameToken, NS_SECEXT,
'UsernameToken');

  RemClassRegistry.RegisterXSClass(UsernameTokenType, NS_SECEXT,
'UsernameTokenType');
  RemClassRegistry.RegisterExternalPropName(TypeInfo(UsernameTokenType),
'Id', '[Namespace="'+NS_UTILITY+'"]');
  RemClassRegistry.RegisterXSClass(PasswordString, NS_SECEXT, 'Password');
  RemClassRegistry.RegisterExternalPropName(TypeInfo(PasswordString),
'Type_', 'Type');

  RemClassRegistry.RegisterXSClass(AttributedString, NS_SECEXT,
'AttributedString');
  RemClassRegistry.RegisterExternalPropName(TypeInfo(AttributedString),
'Id', '[Namespace="'+NS_UTILITY+'"]');
end.


