unit VlerDasClaim;

interface

uses Classes, XMLIntf, XMLDoc, XsdEngineLite, XmlConfig, XsdComplexType,
     XmlComponentBindings, ExtCtrls, SysUtils, IdCoderMIME
     , untConstVals
     ;

type
  TVlerDasClaim = class
    //const        // DONE : LMS moved to a consts unit
      //CLIENT_ASSIGNING_AUTHORITY_SSA = '2.16.840.1.113883.3.184';
      //CLIENT_ASSIGNING_AUTHORITY_VA = '2.16.840.1.113883.4.349';
    private
      AttachmentsList: TList; //of TVlerDasAttachment
      NamespacePrefix: String;
      NamespaceURI: String;
      XsdEngine: TXsdEngineLite;
      XmlConfig: TXmlConfig;
      RootPanel: TPanel;
      procedure RenderMedicalDataXml(XmlNode: IXMLNode);
      function FixDelphiXmlBugs(xmlDocument: IXMLDocument): IXMLDocument;
      function ReformatNamespaceIfNeeded(CurrentNamespace : String) : String;   // CodeCR563 JRL 9/2/14
    public
      DocumentCreationDate: String;
      DocumentTypeVersion: String;
      ClaimID: String;
      DocumentTitleText: String;
      DocumentVersion: String;
      DocumentIdentificationID: String;
      ClientIdentificationID: String;
      ClientAssigningAuthority: String;
      ServiceProviderIdentificationID: String;
      ServiceProviderAssigningAuthority: String;
      ServiceProviderRoleTitle: String;
      PatientDateOfBirth: String;
      PatientFirstName: String;
      PatientMiddleName: String;
      PatientLastName: String;
      PatientGender: String;
      PatientSSN: String;
      ProviderFirstName: String;
      ProviderMiddleName: String;
      ProviderLastName: String;
      FacilityIdentificationID: String;
      FacilityIdentificationText: String;
      FacilityName: String;
      ApproverID : String;                      // CodeCR563 JRL 6/9/14
      ApproverAssigningAuthority : String;      // CodeCR563 JRL 6/9/14
      ApproverLastName : String;                // CodeCR563 JRL 6/9/14
      ApproverFirstName : String;               // CodeCR563 JRL 6/9/14
      ApproverMiddleName : String;              // CodeCR563 JRL 6/9/14
      PhysicalSource : String;                  // CodeCR563 JRL 6/9/14
      ExamID : String;                          // CodeCR563 JRL 6/9/14
      ExaminerType : String;                    // CodeCR563 JRL 6/9/14
      ContractType : String;                    // CodeCR563 JRL 6/9/14
      StatusDate : String;                      // CodeCR563 JRL 6/9/14
      Comments : String;                        // CodeCR563 JRL 6/9/14
      StatusText : String;                      // CodeCR563 JRL 6/10/14
      procedure SetXmlNamespace(prefix: String; URI: String);
      procedure SetAttachments(listOfAttachments: TList); //of TVlerDasAttachment
      procedure SetRootPanel(theRootPanel: TPanel);
      function ToXml: IXMLDocument;
    published
      constructor Create(theXsdEngine: TXsdEngineLite; theXmlConfig: TXmlConfig);
  end;

type
  TVlerDasAttachment = class
    public
      // DecodeFileNM: String;
      BinaryBase64Object: WideString;
      BinaryDescriptionText: String;
      BinaryFormatStandardName: String;
      BinaryLocationURI: String;
      BinarySizeValue: String;
      BinaryCategoryText: String;
      class function EncodeToBase64(stream: TStream): WideString;
      class function DecodeToBase64(TextIn: WideString): TStream;
  end;

implementation

uses forms;

constructor TVlerDasClaim.Create(theXsdEngine: TXsdEngineLite; theXmlConfig: TXmlConfig);
begin
  XsdEngine := theXsdEngine;
  XmlConfig := theXmlConfig;
end;

procedure TVlerDasClaim.SetXmlNamespace(prefix: String; URI: String);
begin
  NamespacePrefix := prefix;
  NamespaceURI := URI;
end;

procedure TVlerDasClaim.SetRootPanel(theRootPanel: TPanel);
begin
  RootPanel := theRootPanel;
end;

procedure TVlerDasClaim.SetAttachments(listOfAttachments: TList);
begin
  AttachmentsList := listOfAttachments;
end;

function TVlerDasClaim.ToXml: IXMLDocument;
const
// Patch 187 notes for in-house exams:  In Patch 186 (version 1.0 DBQs), schema validation of medical data by VBMS was not done.  Patch 187 was originally going to
// send data that would pass VBMS validation.  However, schema validation requires updating header information in CAPRI, and that cannot happen on the same
// timeline that the BEST team needs to update their medical DBQs to comply with legal or business.  So the decision was made to alter CAPRI to send 1.0 DBQ data
// so VBMS will ignore validation.  (Shortest timeline, VBMS could not make the change for months).  In discussions with all the groups, it appears version number
// is not the key to schema validation, it is the constant defined below as NS_CLD.  So the constant NS_CLD will be changed to send the value used in the 1.0 DBQs and
// this should eliminate schema validation and allow VBMS to process all DBQs as well as the BEST team to make periodic updates that dont have to coincide with
// CAPRI releases.  All other changes including new fields and other XML changes are still included in the XML  the namespace NS_CLD will be the only change to
// eliminate schema validation but other data will appear the same as what is in the Claims IEPD 1.1.2.
//NS_CLD = 'http://vler.domain/vler/schemas/health/clinicalDocuments/clinicalAssessments/cpExams/Claim/1.0';  // JRL 10/17/14
//NS_CLD = 'http://vler.domain/vler/schemas/health/clinicalDocuments/clinicalAssessments/cpExams/Claim/1.1.2';  // CodeCR563 JRL 8/4/14
//NS_VLER = 'http://domain/vler/schemas/vlerSupersetSchema/0.9/vler';           // CodeCR563 JRL 7/29/14
//DOCUMENT_TYPE_VERSION = '1.1';                                                // CodeCR563 JRL 8/1/14
  NS_CLD = 'http://vler.domain/vler/schemas/health/clinicalDocuments/clinicalAssessments/cpExams/1.3';  // revert to patch 186 namespace so schema validation is ignored
  NS_NC = 'http://niem.gov/niem/niem-core/2.0';
  NS_VLER = 'http://domain/vler/schemas/vlerSupersetSchema/0.7/vler';           // JRL 11/3/14 Revert to 1.0 headers
  NS_NIEM_XSD = 'http://niem.gov/niem/proxy/xsd/2.0';
  NS_S = 'http://niem.gov/niem/structures/2.0';
  NS_XSI = 'http://www.w3.org/2001/XMLSchema-instance';
  DOCUMENT_TYPE_VERSION = '1.0';                                                // JRL 11/3/14 Revert to 1.0 headers

var
  xml: IXMLDocument;

  xmlNode: IXMLNode;

  commonData: IXMLNode;
  medicalData: IXMLNode;
  attachments: IXMLNode;

  i: Integer;
  anAttachment: TVlerDasAttachment;
begin
  xml := TXMLDocument.Create(Nil);
  xml.Active := True;
  xml.Version := '1.0';                                                         // CodeCR563 JRL 7/29/14
  xml.Encoding := 'UTF-8';
  xml.DocumentElement := xml.CreateNode('cld:Claim', ntElement, NS_CLD);
  xml.DocumentElement.DeclareNamespace('nc', NS_NC);
  xml.DocumentElement.DeclareNamespace('vler', NS_VLER);
  xml.DocumentElement.DeclareNamespace('niem-xsd',NS_NIEM_XSD);
  xml.DocumentElement.DeclareNamespace('s', NS_S);
  xml.DocumentElement.DeclareNamespace('xsi', NS_XSI);
  if Length(NamespacePrefix) > 0 then
  begin
//  NamespaceURI := ReformatNamespaceIfNeeded(NamespaceURI);                    // CodeCR563 JRL 9/2/14
                                                                                // Removed 11/3/14 to revert to 1.0 headers
    xml.DocumentElement.DeclareNamespace(NamespacePrefix, NamespaceURI);
  end;

  // remoaeb code to get the version from the dbq files.  Just hard code it above.  If the common data
  // section changes, then the constants above (path and version) need to be updated too.
  //  xml.DocumentElement.AddChild('DocumentTypeVersion').NodeValue := DocumentTypeVersion; 
  xml.DocumentElement.AddChild('DocumentTypeVersion').NodeValue := DOCUMENT_TYPE_VERSION;
  xml.DocumentElement.AddChild('ClaimID').NodeValue := ClaimID;

  commonData := xml.DocumentElement.AddChild('CommonData');
  xmlNode := commonData.AddChild('nc:Document');
  xmlNode.AddChild('nc:DocumentCreationDate').AddChild('nc:DateTime').NodeValue := DocumentCreationDate;
  xmlNode.AddChild('nc:DocumentDescriptionText').NodeValue := DocumentTitleText;
  xmlNode.AddChild('nc:DocumentIdentification').AddChild('nc:IdentificationID').NodeValue := DocumentIdentificationID;
  xmlNode.AddChild('nc:DocumentStatus').AddChild('nc:StatusText').NodeValue := StatusText;  // Removed hardcoding of "Completed" CodeCR563 JRL 6/13/14
  xmlNode.AddChild('nc:DocumentTitleText').NodeValue := DocumentTitleText;

  // add in exam detail    
  xmlNode := commonData.AddChild('vler:ExamDetail');                            // CodeCR563 JRL 6/9/14
  xmlNode.AddChild('vler:PhysicalSource').NodeValue := PhysicalSource;          // CodeCR563 JRL 6/9/14
  xmlNode.AddChild('vler:ExamId').NodeValue := ExamID;                          // CodeCR563 JRL 6/9/14
  xmlNode.AddChild('vler:ExaminerType').NodeValue := ExaminerType;              // CodeCR563 JRL 6/9/14
  xmlNode.AddChild('vler:ContractType').NodeValue := ContractType;              // CodeCR563 JRL 6/9/14
  with xmlNode.AddChild('vler:StatusDate') do                                   // CodeCR563 JRL 8/4/14
  begin                                                                         // CodeCR563 JRL 8/4/14
    AddChild('nc:DateTime').NodeValue := StatusDate;                            // CodeCR563 JRL 8/4/14
  end;                                                                          // CodeCR563 JRL 8/4/14
//  xmlNode.AddChild('vler:StatusDate').NodeValue := StatusDate;                  // CodeCR563 JRL 6/9/14
  xmlNode.AddChild('vler:Comments').NodeValue := Comments;                      // CodeCR563 JRL 6/9/14
  
  xmlNode := commonData.AddChild('vler:Client');
  xmlNode.Attributes['s:id'] := 'client';
  with xmlNode.AddChild('vler:ClientIdentifier') do
  begin
    AddChild('nc:IdentificationID').NodeValue := ClientIdentificationID;
    AddChild('vler:AssigningAuthority').NodeValue := ClientAssigningAuthority;
  end;
  xmlNode.AddChild('vler:ClientStatus').AddChild('vler:ClientStatusText').NodeValue := 'OK';
  xmlNode.AddChild('nc:RoleOfPersonReference').Attributes['s:ref'] := 'one';

  xmlNode := commonData.AddChild('vler:ServiceProvider');
  xmlNode.Attributes['s:id'] := 'service_provider';
  with xmlNode.AddChild('vler:PersonIdentifier') do
  begin
    AddChild('nc:IdentificationID').NodeValue := ServiceProviderIdentificationID;
    AddChild('vler:AssigningAuthority').NodeValue := ServiceProviderAssigningAuthority;
    AddChild('nc:IdentificationJurisdictionText').NodeValue := 'CAPRI User IEN';
  end;
  xmlNode.AddChild('vler:ServiceProviderRoleTitle').NodeValue := 'Physician';
  xmlNode.AddChild('nc:RoleOfPersonReference').Attributes['s:ref'] := 'two';

  // add in approver detail
  xmlNode := commonData.AddChild('vler:Approver');                              // CodeCR563 JRL 6/9/14
  xmlNode.Attributes['s:id'] := 'approver';                                     // CodeCR563 JRL 6/9/14
  xmlNode.AddChild('PersonIdentifier').NodeValue := '';                         // CodeCR563 JRL 6/9/14
  xmlNode.AddChild('vler:ServiceProviderRoleTitle').NodeValue := 'EXAMINER';    // CodeCR563 JRL 6/9/14
  xmlNode.AddChild('nc:RoleOfPersonReference').Attributes['s:ref'] := 'three';  // CodeCR563 JRL 6/9/14

  xmlNode := commonData.AddChild('nc:Person');                                  // CodeCR563 JRL 6/9/14
  xmlNode.Attributes['s:id'] := 'three';                                        // CodeCR563 JRL 6/9/14
  with xmlNode.AddChild('nc:PersonName') do                                     // CodeCR563 JRL 6/9/14
  begin                                                                         // CodeCR563 JRL 6/9/14
    AddChild('nc:PersonGivenName').NodeValue := ApproverFirstName;              // CodeCR563 JRL 6/9/14
    AddChild('nc:PersonMiddleName').NodeValue := ApproverMiddleName;            // CodeCR563 JRL 6/9/14
    AddChild('nc:PersonSurName').NodeValue := ApproverLastName;                 // CodeCR563 JRL 6/9/14
  end;                                                                          // CodeCR563 JRL 6/9/14

  xmlNode := commonData.AddChild('nc:Person');
  xmlNode.Attributes['s:id'] := 'two';
  with xmlNode.AddChild('nc:PersonName') do
  begin
    AddChild('nc:PersonGivenName').NodeValue := ProviderFirstName;
    AddChild('nc:PersonMiddleName').NodeValue := ProviderMiddleName;
    AddChild('nc:PersonSurName').NodeValue := ProviderLastName;
  end;

  xmlNode := commonData.AddChild('nc:Person');
  xmlNode.Attributes['s:id'] := 'one';
  xmlNode.AddChild('nc:PersonBirthDate').AddChild('nc:Date').NodeValue := PatientDateOfBirth;
  with xmlNode.AddChild('nc:PersonName') do
  begin
    AddChild('nc:PersonGivenName').NodeValue := PatientFirstName;
    AddChild('nc:PersonMiddleName').NodeValue := PatientMiddleName;
    AddChild('nc:PersonSurName').NodeValue := PatientLastName;
  end;
  xmlNode.AddChild('nc:PersonSexCode').NodeValue := PatientGender;
  xmlNode.AddChild('nc:PersonSSNIdentification').AddChild('nc:IdentificationID').NodeValue := PatientSSN;

  xmlNode := commonData.AddChild('nc:Facility');
  with xmlNode.AddChild('nc:FacilityIdentification') do
  begin
    AddChild('nc:IdentificationID').NodeValue := FacilityIdentificationID;
    AddChild('nc:IdentificationJurisdictionText').NodeValue := FacilityIdentificationText;
  end;
  xmlNode.AddChild('nc:FacilityName').NodeValue := FacilityName;

  with commonData.AddChild('vler:ServiceProviderClientAssociation') do
  begin
    AddChild('vler:ServiceProviderReference').Attributes['s:ref'] := 'service_provider';
    AddChild('vler:ClientReference').Attributes['s:ref'] := 'client';
  end;
  

  medicalData := xml.DocumentElement.AddChild('MedicalData');
  RenderMedicalDataXml(medicalData);

  attachments := xml.DocumentElement.AddChild('Attachments');

  if (AttachmentsList <> Nil) and (AttachmentsList.Count > 0) then
  begin
    for i := 0 to AttachmentsList.Count - 1 do
    begin
      anAttachment := AttachmentsList[i];
      with attachments.AddChild('nc:Attachment') do
      begin
        AddChild('nc:BinaryBase64Object').NodeValue := anAttachment.BinaryBase64Object;
        AddChild('nc:BinaryDescriptionText').NodeValue := anAttachment.BinaryDescriptionText;
        AddChild('nc:BinaryFormatStandardName').NodeValue := anAttachment.BinaryFormatStandardName;
        AddChild('nc:BinaryLocationURI').NodeValue := anAttachment.BinaryLocationURI;
        AddChild('nc:BinarySizeValue').NodeValue := anAttachment.BinarySizeValue;
        AddChild('nc:BinaryCategoryText').NodeValue := anAttachment.BinaryCategoryText;
      end;
    end;
  end;

  Result := FixDelphiXmlBugs(xml);
end;

{
This helper function currently fixes:
  #1 - Inappropriate tagging of <ncsd:Selected> ; changes the tag <ncsd:Selected> to <Selected>
}
function TVlerDasClaim.FixDelphiXmlBugs(xmlDocument: IXMLDocument): IXMLDocument;
var
  stringStream: TStringStream;
  xmlString: WideString;
begin
  stringStream := TStringStream.Create('');
  xmlDocument.SaveToStream(stringStream);

  // #1 - change <ncsd:Selected> to <Selected>
  xmlString := StringReplace(stringStream.DataString, 'ncsd:Selected>', 'Selected>',[rfReplaceAll]);

  FreeAndNil(stringStream);
  stringStream := TStringStream.Create(xmlString);

  Result := TXMLDocument.Create(Nil);
  Result.LoadFromStream(stringStream);

  Result := (Result as IXMLDocument);
end;

procedure TVlerDasClaim.RenderMedicalDataXml(XmlNode: IXMLNode);
var
  i: Integer;
  component: TComponent;
  componentName: String;
  complexType: TXsdComplexType;
  xmlBindings: TXmlComponentBindings;

  nsPrefix: String;
  rootTag: String;

  parentNode: IXmlNode;
begin
  if RootPanel = Nil then
    Exit;

  nsPrefix := XmlConfig.GetValue(TXmlConfig.NAMESPACE_PREFIX);
  rootTag := XmlConfig.GetValue(TXmlConfig.ELEMENT_NAME);

  parentNode := XmlNode.AddChild(nsPrefix + ':' + rootTag);
  xmlBindings := TXmlComponentBindings.Create(XmlConfig, parentNode);

  for i := 0 to rootPanel.ComponentCount - 1 do
  begin
    component := rootPanel.Components[i];
    componentName := component.Name;

    if XsdEngine.ContainsComponent(componentName) = False then
      continue;

    complexType := XsdEngine.GetXsdComplexType(componentName);

    xmlBindings.AddXmlNode(component, complexType);
  end; // for i := 0 to rootPanel.ComponentCount - 1 do

end; // function TXmlEngine.RenderMedicalDataXml(rootPanel: TPanel): TStringList;

class function TVlerDasAttachment.EncodeToBase64(stream: TStream): WideString;
var
  encoder: TIdEncoderMime;
begin
  try
    encoder := TIdEncoderMIME.Create(Application);           // LMS : Create with owner so App will free in case its left over
    Result := encoder.Encode(stream, stream.Size);
  except
    on E: Exception do
      Result := '';
  end;

  if encoder <> Nil then
    FreeAndNil(encoder);
end;

class function TVlerDasAttachment.DecodeToBase64(TextIn : WideString): TStream;
var
  Decoder: TIdDecoderMime;
  DecodedStream: TMemoryStream;
begin
  Result := nil;
  try
    DecodedStream:= TMemoryStream.Create;
    Decoder := TIdDecoderMime.Create(Application);
    Decoder.DecodeBegin(DecodedStream);
    Decoder.Decode(TextIn);
    Decoder.DecodeEnd;
    Result := DecodedStream;
  except
    on E: Exception do
  end;

  if Decoder <> Nil then
    FreeAndNil(Decoder);
end;

//------------------------------------------------------------------------------
// Modify the URI if the URI is in the format of:  
//   "http://vler.domain/vler/schemas/health/clinicalDocuments/clinicalAssessments/cpExams/1.3/HearingLossAndTinnitus"
//
// From: xmlns:hlat="http://vler.domain/vler/schemas/health/clinicalDocuments/clinicalAssessments/cpExams/1.3/HearingLossAndTinnitus">
//   To: xmlns:hlat="http://vler.domain/vler/schemas/health/clinicalDocuments/clinicalAssessments/cpExams/HearingLossAndTinnitus/1.3">
//------------------------------------------------------------------------------
function TVlerDasClaim.ReformatNamespaceIfNeeded(CurrentNamespace: String): String;
var
  cpExamsPosition,VersionPosition, DBQNamePosition, OutValue : Integer;
  DBQNameStr, VersionStr, UpdatedStr : string;
begin
  cpExamsPosition := Pos('cpExams',CurrentNamespace);
  VersionPosition := cpExamsPosition + 8;  // go to first character past next slash
  if TryStrToInt(Copy(CurrentNamespace,VersionPosition,1),OutValue) then
  begin // convert
    DBQNamePosition := VersionPosition + 4;
    VersionStr := Copy(CurrentNamespace,VersionPosition,3);
    DBQNameStr := Copy(CurrentNamespace,DBQNamePosition,(length(CurrentNameSpace)-DBQNamePosition+1));
    UpdatedStr := Copy(CurrentNamespace,1,(VersionPosition-1)) + DBQNameStr + '/' + VersionStr;
    Result := UpdatedStr;
  end
  else
    Result := CurrentNamespace;
end;
  
end.
