unit VocRehabClass;
//------------------------------------------------------------------------------
// CodeCR347 - JRL June, 2012 - Added unit
// Vocational Rehab Form 28-8861
//
// This class is comprised of properties and routines to support the addition
// of an electronic Vocational Rehab form 28-8861.
//------------------------------------------------------------------------------
interface

uses SysUtils, Classes, Dialogs, DateUtils, CommCtrl, Types, Graphics,
     Controls, StrUtils,
     DiTypLib,
     Trpcb,
     CAPRISupport,
     dialogsCAPRI,
     Fmcntrls,
     Fmcmpnts,
     VAUtils;

type 
  TVocRehab = Class
CONST
    DATE_RANGE = 'Date Range';
    NO_RECORD_FOUND = 'No record found';
    NOT_COMPLETED = '[Not Completed]';
    FIELD2COL = 20;  // used to format dates on VocRehab Tab selection screen
    FIELD3COL = 40;  // used to format dates on VocRehab Tab selection screen
  private
    TVocRehab_DateOptions : TStringList;
    sPOC_Name : string;             // See property POC_Name - Point of Contact
    sPOC_Email : string;            // See property POC_Email
    sPOC_Phone : string;            // See property POC_Phone
    sPOC_Location : string;         // See property POC_Location
    sPOC_LocationIEN : string;      // See property POC_LocationIEN
    sPOC_IEN : string;              // See property POC_IEN
    // VHA Coordinator Role flag
    bVHAAccess : Boolean;           // See property VHAAccess
    // Request Status
    bRequestCancelled : Boolean;    // See property RequestCancelled
    bRequestCompleted : Boolean;    // See property RequestCompleted
    // report variables
    bReportIsDelimited : Boolean;   // See property ReportIsDelimited
    sReportRO : string;             // See property ReportRO
    sReportStatus : char;           // See property ReportStatus
    dtCurrentDate : TDateTime;
    dtReportStartDate : TDateTime;
    dtReportStopDate : TDateTime;
    FiscalYearCurrent_StartDate : TDateTime;
    FiscalYearCurrent_EndDate : TDateTime;
    FiscalYearCurrentText : string;
    FiscalYearPrevious1_StartDate : TDateTime;
    FiscalYearPrevious1_EndDate : TDateTime;
    FiscalYearPrevious1Text : string;
    FiscalYearPrevious2_StartDate : TDateTime;
    FiscalYearPrevious2_EndDate : TDateTime;
    FiscalYearPrevious2Text : string;
    FiscalYearAll_StartDate : TDateTime;
    FiscalYearAll_EndDate : TDateTime;
    FiscalYearAllText : string;
    CalendarYearCurrent_StartDate : TDateTime;
    CalendarYearCurrent_EndDate : TDateTime;
    CalendarYearCurrentText : string;
    CalendarYearPrevious1_StartDate : TDateTime;
    CalendarYearPrevious1_EndDate : TDateTime;
    CalendarYearPrevious1Text : string;
    CalendarYearPrevious2_StartDate : TDateTime;
    CalendarYearPrevious2_EndDate : TDateTime;
    CalendarYearPrevious2Text : string;
    CalendarYearAll_StartDate : TDateTime;
    CalendarYearAll_EndDate : TDateTime;
    CalendarYearAllText : string;
    ErrorDate : TDateTime;
    class function GetAddExam: Boolean;
    class procedure SetAddExam(const Value: Boolean);
    function  GetThirtyMonthCap: TDateTime;
    function  GetDateOptionList: TStringList;
    procedure SetupCalendarYearCurrent(wYear : Word);
    procedure SetupCalendarYearPrevious1(wYear : Word);
    procedure SetupCalendarYearPrevious2(wMonth, wYear : Word);
    procedure SetupCalendarYearAll;
    procedure SetupFiscalYearCurrent(wYear : Word);
    procedure SetupFiscalYearPrevious1(wYear : Word);
    procedure SetupFiscalYearPrevious2(wMonth, wYear : Word);
    procedure SetupFiscalYearAll;
    procedure CalculateDates;
    function  GetVAClaimNumber : String;
  public  
  published
    constructor Create; Overload;
    Destructor  Destroy; override; 
    procedure EnableReasonForReferral;
    procedure GetDateRanges(strRange : String; out dtStart, dtEnd : TDateTime);
    procedure GetReportStatusList;
    function  ValidateDateRange(StartDate, EndDate : TDateTime) : Boolean;
    procedure SetupMedicalServicesList;
    function  PopulatePOCInfo(sPOC_IEN : string) : Boolean;
    function  GetPointOfContactIEN(NewRecord : Boolean) : string;
    procedure SendMailManMessage(MsgType : string);
    procedure RunForm8861StatusReport;
    property  AddExam : Boolean   // Used to determine "Add" or "Edit" exam
       Read   GetAddExam
       Write  SetAddExam;
    property  ThirtyMonthCap : TDateTime     // Report history end date - no reports exceed 30 months in length
       Read   GetThirtyMonthCap;
    property  DateOptionList : TStringList
       Read   GetDateOptionList;
    property  CurrentDate : TDateTime
       Read   dtCurrentDate
       Write  dtCurrentDate;
    property  ReportStartDate : TDateTime    // Report Start Date used in VocRehabReportSetup.pas
       Read   dtReportStartDate
       Write  dtReportStartDate;
    property  ReportStopDate : TDateTime     // Report Stop Date used in VocRehabReportSetup.pas
       Read   dtReportStopDate
       Write  dtReportStopDate;
    property  ReportStatus : char            // Report Status (ALL, NEW, PENDING, CANCELLED, COMPLETE) used in VocRehabReportSetup.pas
       Read   sReportStatus
       Write  sReportStatus;
    property  ReportIsDelimited : Boolean    // Report Is Delimited flag used in VocRehabReportSetup.pas
       Read   bReportIsDelimited
       Write  bReportIsDelimited;
    property  ReportRegionalOffice : string  // Report Regional Office
       Read   sReportRO
       Write  sReportRO;
    property  POC_Name : string   // Point of Contact Name
       Read   sPOC_Name
       Write  sPOC_Name;
    property  POC_Email : string  // Point of Contact Email Address
       Read   sPOC_Email
       Write  sPOC_Email;
    property  POC_Phone : string  // Point of Contact Phone
       Read   sPOC_Phone
       Write  sPOC_Phone;
    property  POC_Location : string  // Point of Contact Location
       Read   sPOC_Location
       Write  sPOC_Location;
    property  POC_LocationIEN : string  // Point of Contact Location IEN
       Read   sPOC_LocationIEN
       Write  sPOC_LocationIEN;
    property  POC_IEN : string  // Point of Contact IEN
       Read   sPOC_IEN
       Write  sPOC_IEN;
    property  VAClaimNumber : string  // display only 
       Read   GetVAClaimNumber;
    property  VHAAccess : Boolean  // security keys access for VHA Coordinator Use
       Read   bVHAAccess
       Write  bVHAAccess;
    property  RequestCancelled : Boolean  // flag to indicate if request has been cancelled
       Read   bRequestCancelled
       Write  bRequestCancelled;
    property  RequestCompleted : Boolean  // flag to indicate if request has been completed
       Read   bRequestCompleted
       Write  bRequestCompleted;
End;                           

implementation
{ TVocRehab }

uses
  Main, VocRehabMedicalRequest, ROFinder, VocRehabPOC, VocRehabReportSetup,
  REPORTS;

var
    TVocRehab_AddExam : boolean;

//------------------------------------------------------------------------------
// CONSTRUCTOR
//------------------------------------------------------------------------------
constructor TVocRehab.Create;
begin
   inherited;
   TVocRehab_DateOptions := TStringList.Create;
   ErrorDate := EncodeDate(1900,1,1);
   CurrentDate := now;  // Creating a variable allows testing for other date options
end;
//------------------------------------------------------------------------------
// DESTRUCTOR
//------------------------------------------------------------------------------
destructor TVocRehab.Destroy;
begin
  FreeAndNil(TVocRehab_DateOptions);
  inherited;                                     
end;

//------------------------------------------------------------------------------
// PROPERTY -- AddExam (Boolean)
//------------------------------------------------------------------------------
class procedure TVocRehab.SetAddExam(const Value: Boolean);
begin
    TVocRehab_AddExam := Value;
end;

class function TVocRehab.GetAddExam: Boolean;
begin
  result := TVocRehab_AddExam;
end;

//------------------------------------------------------------------------------
// PROPERTY -- Date Option List (TStringList)
//------------------------------------------------------------------------------
// Sample Date Range below uses 2012 as current year.  The dates are calculated
// with a 30 month cap on historical dates.
//   Date Range (only option that allows user date input)
//   All Fiscal Years (10/1/2010-Today)
//   Fiscal Year 2012 (10/1/2011-Today)
//   Fiscal Year 2011 (10/1/2010-9/30/2011)
//   Fiscal Year 2010 (10/1/2009-9/30/2010) - only available in Oct - March
//   All Calendar Years (1/1/2010-Today)
//   Calendar Year 2012 (1/1/2012-Today)
//   Calendar Year 2011 (1/1/2011-12/31/2011)
//   Calendar Year 2010 (1/1/2010-12/31/2010) - only available January - June
//------------------------------------------------------------------------------
function TVocRehab.GetDateOptionList: TStringList;
begin
  TVocRehab_DateOptions.Clear;
  CalculateDates;
  TVocRehab_DateOptions.Add(DATE_RANGE);
  TVocRehab_DateOptions.Add(FiscalYearCurrentText);
  TVocRehab_DateOptions.Add(FiscalYearPrevious1Text);
  if FiscalYearPrevious2Text <> '' then
     TVocRehab_DateOptions.Add(FiscalYearPrevious2Text);
  TVocRehab_DateOptions.Add(FiscalYearAllText);
  TVocRehab_DateOptions.Add(CalendarYearCurrentText);
  TVocRehab_DateOptions.Add(CalendarYearPrevious1Text);   
  if CalendarYearPrevious2Text <> '' then
     TVocRehab_DateOptions.Add(CalendarYearPrevious2Text);
  TVocRehab_DateOptions.Add(CalendarYearAllText);
  Result := TVocRehab_DateOptions;
end;

//------------------------------------------------------------------------------
// PROPERTY -- 30 Month Cap (TDateTime)
//------------------------------------------------------------------------------
function TVocRehab.GetThirtyMonthCap: TDateTime;  
begin
   Result := IncMonth(now,-30);
end;

//------------------------------------------------------------------------------
// PROPERTY -- Get VA Claim Number
//             Patient file #2, field #.313
//------------------------------------------------------------------------------
function TVocRehab.GetVAClaimNumber: String;
begin
  frmVRMedicalRequest.FMGetsVRMedServices.FieldNumbers.Clear;
  frmVRMedicalRequest.FMGetsVRMedServices.FieldNumbers.Add('.313');
  frmVRMedicalRequest.FMGetsVRMedServices.FileNumber := '2';
  frmVRMedicalRequest.FMGetsVRMedServices.IENS := PatientIEN;
  frmVRMedicalRequest.FMGetsVRMedServices.GetData;
  if frmVRMedicalRequest.FMGetsVRMedServices.Results.Count < 1 then
     Result := ''
  else
     Result := frmVRMedicalRequest.FMGetsVRMedServices.GetField('.313').FMDBExternal;
end;

//------------------------------------------------------------------------------
// PUBLISHED -- Get Point of Contact Info
//------------------------------------------------------------------------------
// File Number:      200  (New Person File)
//    Field Number:  .01  (Point of Contact Name)
//    Field Number:  .151 (Point of Contact Email)
//    Field Number:  .132 (Point of Contact Phone)
// File Number:      396.9 (Form 8861 File)
//    Field Number:  3 (Point of Contact Regional Office
//------------------------------------------------------------------------------
// Input parameter:  Point of Contact IEN - either from database or logged in user
//------------------------------------------------------------------------------
function TVocRehab.PopulatePOCInfo(sPOC_IEN : string) : Boolean;
var
   i : Integer;
   VisitedFrom : string;
begin
   Result := TRUE;
   if sPOC_IEN = NO_RECORD_FOUND then  // it is possible to pass in a error IEN
   begin
      Result := FALSE;
      exit;
   end;
   // Get POC Name, Phone and Email address from New Person File
   frmMain.FMGetsVocRehab.FieldNumbers.Clear;
   frmMain.FMGetsVocRehab.FieldNumbers.Add('.01');
   frmMain.FMGetsVocRehab.FieldNumbers.Add('.151');
   frmMain.FMGetsVocRehab.FieldNumbers.Add('.132');
   frmMain.FMGetsVocRehab.IENS := sPOC_IEN;
   frmMain.FMGetsVocRehab.FileNumber := '200';
   frmMain.FMGetsVocRehab.GetData;
   if frmMain.FMGetsVocRehab.Results.Count > 0 then
   begin
      POC_Name := frmMain.FMGetsVocRehab.GetField('.01').FMDBExternal;
      POC_Email := lowercase(frmMain.FMGetsVocRehab.GetField('.151').FMDBExternal);
      POC_Phone := frmMain.FMGetsVocRehab.GetField('.132').FMDBExternal;  // Phone number for local users
   end;
   // Get data from Form 8861 File
   frmMain.FMGetsVocRehab.FieldNumbers.Clear;
   frmMain.FMGetsVocRehab.FieldNumbers.Add('3');
   frmMain.FMGetsVocRehab.IENS := frmVRMedicalRequest.SelectedMedicalRequestIEN;
   frmMain.FMGetsVocRehab.FileNumber := '396.9';
   frmMain.FMGetsVocRehab.GetData;
   if frmMain.FMGetsVocRehab.Results.Count > 0 then
   begin
      POC_LocationIEN := frmMain.FMGetsVocRehab.GetField('3').FMDBInternal;
      POC_Location := frmMain.FMGetsVocRehab.GetField('3').FMDBExternal;
   end;

   // Get phone number for POC logged in remotely 
   // Get VISITED FROM field from File 200 / Multiple field 8910
   frmVRPointOfContact.FMListerVR_POC.Screen := '';
   frmVRPointOfContact.FMListerVR_POC.FMIndex := '';
   frmVRPointOfContact.FMListerVR_POC.DisplayFields.Clear;
   frmVRPointOfContact.FMListerVR_POC.FieldNumbers.Clear;
   frmVRPointOfContact.FMListerVR_POC.IENS := ',' + sPOC_IEN;
   frmVRPointOfContact.FMListerVR_POC.FileNumber := '200.06';   
   frmVRPointOfContact.FMListerVR_POC.DisplayFields.Add('.01');
   frmVRPointOfContact.FMListerVR_POC.DisplayFields.Add('5');
   frmVRPointOfContact.FMListerVR_POC.FieldNumbers.Add('.01');
   frmVRPointOfContact.FMListerVR_POC.FieldNumbers.Add('5');
   frmVRPointOfContact.FMListerVR_POC.GetList(nil);
   
   // Loop through multiples and find the Visited From value that matches the Station Number
   for i := 0 to frmVRPointOfContact.FMListerVR_POC.Results.Count - 1 do
   begin
     VisitedFrom := frmVRPointOfContact.FMListerVR_POC.GetRecord(frmVRPointOfContact.FMListerVR_POC.Results[i]).GetField('.01').FMDBExternal;
     if VisitedFrom = UserDivisionNumber then
     begin
       POC_Phone := frmVRPointOfContact.FMListerVR_POC.GetRecord(frmVRPointOfContact.FMListerVR_POC.Results[i]).GetField('5').FMDBExternal;
       break;
     end;
   end;
end;

//------------------------------------------------------------------------------
// PUBLISHED -- Get Point Of Contact IEN
//------------------------------------------------------------------------------
// Input:  NewRecord field of type Boolean
// Return: Point of Contact IEN
//   If this is a new request for medical services (NewRecord = TRUE), 
//      return the logged in user IEN
//   If this is displaying an existing request (NewRecord = FALSE), 
//      return the IEN stored in File 396.9, Field 11
//------------------------------------------------------------------------------
function TVocRehab.GetPointOfContactIEN(NewRecord : Boolean) : string;
var
  errormsg : string;
begin
  if NewRecord then // get the current users IEN
  begin
    if not CallRPC(frmMain.RPCBroker1, 'XWB GET VARIABLE VALUE', ['DUZ'], nil, True) then
      exit;
    result := frmMain.RPCBroker1.Results[0];
  end
  else  // get the existing IEN from the database file # 396.9, field #11
  begin
    frmMain.FMGetsVocRehab.FieldNumbers.Clear;
    frmMain.FMGetsVocRehab.FieldNumbers.Add('11');
    frmMain.FMGetsVocRehab.FileNumber := '396.9';
    frmMain.FMGetsVocRehab.IENS := frmVRMedicalRequest.SelectedMedicalRequestIEN;
    frmMain.FMGetsVocRehab.GetData;
    if frmMain.FMGetsVocRehab.Results.Count < 1 then
    begin
       errormsg := NO_RECORD_FOUND;
       result := NO_RECORD_FOUND;
    end
    else
       result := frmMain.FMGetsVocRehab.GetField('11').FMDBInternal;
  end;
end;


//------------------------------------------------------------------------------
// PUBLISHED -- Run Form 8861 Status Report
//------------------------------------------------------------------------------
procedure TVocRehab.RunForm8861StatusReport;
var
  DelimitedLines: TStringList;
  AbbrevReportName, FileName, FilePath, Delimited: String;
  RPCSuccess: Boolean;
begin                        
   frmMain.RPCBroker1.Results.Clear;
   AbbrevReportName := 'Form8861StatusRpt';
   FileName := AbbrevReportName + '_'
                               + FormatDateTime('YYYYMMDD', dtReportStartDate)
                               + '_'
                               + FormatDateTime('YYYYMMDD', dtReportStopDate);
   DelimitedLines := TStringList.Create;
   if ReportIsDelimited then
     Delimited := '1'   // delimited text
   else
     Delimited := '0';  // plain text
   try
     frmReports.DisplayBusyAnimation(True);
     RPCSuccess := CallRPC(frmMain.RPCBroker1, 'DVBAB REPORTS',
                   ['15',
                   FMToDateConvert(FormatDateTime('mm/dd/yyyy',dtReportStartDate))
                   + '^'
                   + FMToDateConvert(FormatDateTime('mm/dd/yyyy',dtReportStopDate))
                   + '^'
                   + ReportRegionalOffice
                   + '^'
                   + ReportStatus       
                   + '^'
                   + Delimited], nil, False);
     if RPCSuccess then
     begin
       frmReports.ReportMemo.Visible := False;
       if frmMain.RPCBroker1.Results.Count > 0 then
         if NOT AnsiStartsText('NO DATA FOUND', frmMain.RPCBroker1.Results[0]) then
           if ReportIsDelimited then
             CAPRIQuickCopy(frmMain.RPCBroker1.Results, DelimitedLines)
           else
             CAPRIQuickCopy(frmMain.RPCBroker1.Results, frmReports.Reportmemo);
     end;
     frmReports.DisplayBusyAnimation(False);

     if (DelimitedLines.Count > 0) then
     begin
       FilePath := frmReports.SaveDelimitedReport(DelimitedLines, FileName);
       frmReports.ReportMemo.Lines.Insert(0, '');
       if (FilePath <> '') then
         frmReports.ReportMemo.Lines.Insert(0, IntToStr(DelimitedLines.Count)
                                    + ' Extract lines saved to: ' + #13#10
                                    + FilePath)
       else
         frmReports.ReportMemo.Lines.Insert(0, 'No delimited text lines were saved!');
     end;

     If (frmReports.ReportMemo.Lines.Count = 0) and (DelimitedLines.Count = 0) Then
       frmReports.ReportMemo.Lines.Add('No data found for parameters entered.');
     if ReportIsDelimited then
       frmReports.ReportMemo.Lines.Insert(0, ''); //only insert space when delimited to maintain existing text format
     frmReports.ReportMemo.Visible := True;
     frmReports.ReportMemo.SelStart := 0;
     frmReports.ReportMemo.SelLength := 0;
     frmMain.RPCBroker1.Results.Clear;
   finally
     FreeAndNil(DelimitedLines);
   end;
end;

//------------------------------------------------------------------------------
// PUBLISHED -- Send MailMan Message
//------------------------------------------------------------------------------
// MsgType:
//   NEW - sent when a new Request for Medical Services is sent
//   PENDING - sent when a VHA Coordinator assigns a consult to a medical request
//   CANCELLED - sent when a VRE Counselor cancels the request for services  
//------------------------------------------------------------------------------
procedure TVocRehab.SendMailManMessage(MsgType: string);
begin
  try
    if not CallRPC(frmMain.RPCBroker1, 'DVBAB 8861 NOTIFICATIONS', [frmVRMedicalRequest.SelectedMedicalRequestIEN, MsgType],nil, False, False, False) then
       ShowMessageCAPRI('Unable to send Mailman notifications of ' + MsgType + ' Medical Service Request!');
  except
    on e:exception do
       ShowMessageCAPRI('Unable to send Mailman notifications of ' + MsgType + ' Medical Service Request!');
  end;
end;

//------------------------------------------------------------------------------
// PUBLISHED -- Get Report Status List 
//------------------------------------------------------------------------------
// Instead of storing an IEN as part of the listbox object, we are going to store
// the ASCII value of the one letter code.  
// "A" = 65, "C" = 67, "N" = 78, "P" = 80, "X" = 87
//------------------------------------------------------------------------------
procedure TVocRehab.GetReportStatusList;   
var
  i : Integer;
  buffer, buffer2 : string;
  AsciiVal : SmallInt;
begin    
  frmVREReportSetup.lbVRStatus.Clear;                              
  if CallRPC(frmMain.RPCBroker1, 'DVBAB GET SET', ['396.9', '13'],nil, False, False, False) then
    if (frmMain.RPCBroker1.Results.Count > 0) then
    begin
      frmVREReportSetup.lbVRStatus.Items.Clear;
      for i := 0 to frmMain.RPCBroker1.Results.Count - 1 do
      begin
        buffer := Piece(frmMain.RPCBroker1.Results[i],'^',2); // Get Full Status Names
        buffer2 := Piece(frmMain.RPCBroker1.Results[i],'^',1); // Get one character Status codes (A,C,N,P,X)
        AsciiVal := Ord(buffer2[1]);  // Save the ASCII value of the one character code as part of the listbox Object 
        frmVREReportSetup.lbVRStatus.AddItem(buffer,TObject(AsciiVal));
      end;
      frmVREReportSetup.lbVRStatus.ItemIndex := 0;
    end
    else
       ShowMessageCAPRI('Unable to retrieve Report Status Options');
end;


//------------------------------------------------------------------------------
// PUBLISHED -- GetDateRanges
//------------------------------------------------------------------------------
procedure TVocRehab.GetDateRanges(strRange: String; out dtStart,
  dtEnd: TDateTime);
begin
   if strRange = DATE_RANGE then
   begin
     dtStart := Now;
     dtEnd := Now;
   end;
   if strRange = FiscalYearAllText then
   begin
     dtStart := FiscalYearAll_StartDate;
     dtEnd := FiscalYearAll_EndDate;
   end;
   if strRange = FiscalYearCurrentText then
   begin
     dtStart := FiscalYearCurrent_StartDate;
     dtEnd := FiscalYearCurrent_EndDate;
   end;
   if strRange = FiscalYearPrevious1Text then
   begin
     dtStart := FiscalYearPrevious1_StartDate;
     dtEnd := FiscalYearPrevious1_EndDate;
   end;
   if strRange = FiscalYearPrevious2Text then
   begin
     dtStart := FiscalYearPrevious2_StartDate;
     dtEnd := FiscalYearPrevious2_EndDate;
   end;
   if strRange = CalendarYearAllText then
   begin
     dtStart := CalendarYearAll_StartDate;
     dtEnd := CalendarYearAll_EndDate;
   end;
   if strRange = CalendarYearCurrentText then
   begin
     dtStart := CalendarYearCurrent_StartDate;
     dtEnd := CalendarYearCurrent_EndDate;
   end;
   if strRange = CalendarYearPrevious1Text then
   begin
     dtStart := CalendarYearPrevious1_StartDate;
     dtEnd := CalendarYearPrevious1_EndDate;
   end;
   if strRange = CalendarYearPrevious2Text then
   begin
     dtStart := CalendarYearPrevious2_StartDate;
     dtEnd := CalendarYearPrevious2_EndDate;
   end;
end;


//------------------------------------------------------------------------------
// PUBLISHED -- Enable Reason For Referral Edits
//------------------------------------------------------------------------------
procedure TVocRehab.EnableReasonForReferral;
begin
   frmVRMedicalRequest.lblServicesOther.Font.Color := clBlack;
   frmVRMedicalRequest.edtServicesOther.ReadOnly := FALSE;
   frmVRMedicalRequest.edtServicesOther.Text := '';
   frmVRMedicalRequest.edtServicesOther.Color := clWhite;
   frmVRMedicalRequest.edtServicesOther.Font.Color := clBlack;
end;

//------------------------------------------------------------------------------
// PUBLISHED -- Validate Date Range
//------------------------------------------------------------------------------
function TVocRehab.ValidateDateRange(StartDate, EndDate: TDateTime): Boolean;
var
   DateError : Boolean;
   Date30MonthsAgo : TDateTime;
   CompareValue : Integer;
begin
   DateError := FALSE;
   // 30 month cap for StartDate, Current Date for EndDate
   if CompareDate(EndDate,Now) = GreaterThanValue then
      DateError := TRUE;
   Date30MonthsAgo := IncMonth(Now, -30);
   CompareValue := CompareDate(Date30MonthsAgo,StartDate);
   if CompareValue = GreaterThanValue then
      DateError := TRUE;
   Result := DateError;   
end;

//------------------------------------------------------------------------------
// PUBLISHED -- Setup Medical Services List
//------------------------------------------------------------------------------
procedure TVocRehab.SetupMedicalServicesList;
Var
  x,i: integer;
  RequestDate, CancelDate, CompleteDate : String;
  buffer : string;
Begin
  // Get Medical Services List
  frmMain.lbVocRehab.Clear;
  // FMLister
  frmMain.FMListerVocRehab.DisplayFields.Clear;
  frmMain.FMListerVocRehab.FieldNumbers.Clear;
  frmMain.FMListerVocRehab.PartList.Clear;
  frmMain.FMListerVocRehab.RPCBroker := frmMain.RPCBroker1;
  frmMain.FMListerVocRehab.FileNumber := '396.9';
  frmMain.FMListerVocRehab.FMIndex := 'C';            
  frmMain.FMListerVocRehab.DisplayFields.Add('.01');  // Request Date
  frmMain.FMListerVocRehab.DisplayFields.Add('2');    // Completed Date
  frmMain.FMListerVocRehab.DisplayFields.Add('4');    // Patient IEN
  frmMain.FMListerVocRehab.DisplayFields.Add('15');   // Cancellation Date
  frmMain.FMListerVocRehab.FieldNumbers.Add('.01');
  frmMain.FMListerVocRehab.FieldNumbers.Add('2');
  frmMain.FMListerVocRehab.FieldNumbers.Add('4');
  frmMain.FMListerVocRehab.FieldNumbers.Add('15');
  frmMain.FMListerVocRehab.Screen := 'I $P(^(0),U,5)=' + PatientIEN; // filter for PatientIEN
  frmMain.FMListerVocRehab.GetList(nil);
  if frmMain.FMListerVocRehab.Results.Count < 1 then
     frmMain.lbVocRehab.Items.Add('No records found');
  // Get and parse the returned records  
  for x := 0 to frmMain.FMListerVocRehab.Results.Count - 1 do
  begin
    RequestDate := frmMain.FMListerVocRehab.GetRecord(frmMain.FMListerVocRehab.Results[x]).GetField('.01').FMDBExternal;
    CancelDate := frmMain.FMListerVocRehab.GetRecord(frmMain.FMListerVocRehab.Results[x]).GetField('15').FMDBExternal;
    CompleteDate := frmMain.FMListerVocRehab.GetRecord(frmMain.FMListerVocRehab.Results[x]).GetField('2').FMDBExternal;
    // format listbox output - listbox uses a fixed font so spaces work for formatting                
    buffer := RequestDate + ' |';                                                                                             
    for i := length(buffer) to FIELD2COL do
      buffer := buffer + ' ';
    if CompleteDate <> '' then
      buffer := buffer + CompleteDate;
    if (VAUtils.ScreenReaderActive = TRUE) AND (CompleteDate = '') then
      buffer := buffer + NOT_COMPLETED;
    for i := length(buffer) to FIELD3COL - 3 do
      buffer := buffer + ' ';
    buffer := buffer + ' | ';
    if CancelDate <> '' then
      buffer := buffer + CancelDate + '[Cancelled]';
    frmMain.lbVocRehab.AddItem(buffer,TObject(StrToInt(frmMain.FMListerVocRehab.Results[x]))); // add item stores the record IEN in the obejct
  end;
  //508 accessibility
  if VAUtils.ScreenReaderActive then
  begin
    if frmMain.lbVocRehab.visible then
      frmMain.lbVocRehab.SetFocus;
  end;
end; 

//------------------------------------------------------------------------------
// PRIVATE -- CALCULATE DATES
//------------------------------------------------------------------------------
procedure TVocRehab.CalculateDates;
var
  wMonth, wDay, wYear : Word;
begin
   DecodeDate(dtCurrentDate, wYear, wMonth, wDay);
   SetupCalendarYearCurrent(wYear);
   SetupCalendarYearPrevious1(wYear);
   SetupCalendarYearPrevious2(wMonth, wYear);
   SetupFiscalYearCurrent(wYear);
   SetupFiscalYearPrevious1(wYear);
   SetupFiscalYearPrevious2(wMonth, wYear);
end;

//------------------------------------------------------------------------------
// PRIVATE -- CALENDAR DATE ROUTINES
//------------------------------------------------------------------------------
procedure TVocRehab.SetupCalendarYearAll;
begin  
   // Calendar Year <CurrentYear-2>
   // Start Date:  1/1/<CurrentYear-2>
   // Stop Date:  12/31/<CurrentYear-2>
   if CalendarYearPrevious2_StartDate = ErrorDate then
      CalendarYearAll_StartDate := CalendarYearPrevious1_StartDate
   else
      CalendarYearAll_StartDate := CalendarYearPrevious2_StartDate;
   CalendarYearAll_EndDate := dtCurrentDate;
   CalendarYearAllText := 'All Calendar Years'; 
end; 

//-------------------------------------------//
procedure TVocRehab.SetupCalendarYearCurrent(wYear : Word);
begin
   // Calendar Year <CurrentYear>
   // Start Date:  1/1/<CurrentYear>
   // Stop Date:  Today
   CalendarYearCurrent_StartDate := EncodeDate(wYear, 1, 1);
   CalendarYearCurrent_EndDate := dtCurrentDate;
   CalendarYearCurrentText := 'Calendar Year ' + IntToStr(wYear);
end;

//-------------------------------------------//
procedure TVocRehab.SetupCalendarYearPrevious1(wYear : Word);
begin
   // Calendar Year <CurrentYear-1>
   // Start Date:  1/1/<CurrentYear-1>
   // Stop Date:  12/31/<CurrentYear-1>
   wYear := wYear - 1;  // previous year
   CalendarYearPrevious1_StartDate := EncodeDate(wYear, 1, 1);
   CalendarYearPrevious1_EndDate := EncodeDate(wYear, 12, 31);
   CalendarYearPrevious1Text := 'Calendar Year ' + IntToStr(wYear);
end;

//-------------------------------------------//
procedure TVocRehab.SetupCalendarYearPrevious2(wMonth, wYear : Word);
begin
   // Calendar Year <CurrentYear-2>
   // Start Date:  1/1/<CurrentYear-2>
   // Stop Date:  12/31/<CurrentYear-2>
   wYear := wYear - 2;  // Current year - 2
   CalendarYearPrevious2_StartDate := EncodeDate(wYear, 1, 1);
   CalendarYearPrevious2_EndDate := EncodeDate(wYear, 12, 31);
   CalendarYearPrevious2Text := 'Calendar Year ' + IntToStr(wYear);
   // calculate except if date exceeds the 30 month cap
   if wMonth > 6 then
   begin
      CalendarYearPrevious2_StartDate := ErrorDate;
      CalendarYearPrevious2_EndDate := ErrorDate;
      CalendarYearPrevious2Text := '';  // Months July-Dec of the current year will exceed the 30 month gap, so don't add text             
   end; 
   SetupCalendarYearAll; // Always must be called after above routine 
end;

//-------------------------------------------//
procedure TVocRehab.SetupFiscalYearAll;        
begin
   // Note:  Needs to be called after SetupFiscalYearPrevious2
   // Fiscal Year All
   // Start Date: 10/1/<CurrentYear-2>
   // Stop Date:  CurrentDate
   if FiscalYearPrevious2_StartDate = ErrorDate then
      FiscalYearAll_StartDate := FiscalYearPrevious1_StartDate
   else
      FiscalYearAll_StartDate := FiscalYearPrevious2_StartDate;
   FiscalYearAll_EndDate := dtCurrentDate;   
   FiscalYearAllText := 'All Fiscal Years';
end;

//-------------------------------------------//
procedure TVocRehab.SetupFiscalYearCurrent(wYear : Word);
begin
   // Fiscal Year <CurrentYear>
   // Start Date: 10/1/<CurrentYear-1>
   // Stop Date:  CurrentDate
   FiscalYearCurrent_StartDate := EncodeDate(wYear-1, 10, 1);
   FiscalYearCurrent_EndDate := dtCurrentDate;
   FiscalYearCurrentText := 'Fiscal Year ' + IntToStr(wYear);
end;

//-------------------------------------------//
procedure TVocRehab.SetupFiscalYearPrevious1(wYear : Word);
begin
   // Fiscal Year <CurrentYear-1>
   // Start Date: 10/1/<CurrentYear-2>
   // Stop Date:  9/30/<CurrentYear-1>
   FiscalYearPrevious1_StartDate := EncodeDate(wYear-2, 10, 1);
   FiscalYearPrevious1_EndDate := EncodeDate(wYear-1,9,30);
   FiscalYearPrevious1Text := 'Fiscal Year ' + IntToStr(wYear-1);
end;

//-------------------------------------------//
procedure TVocRehab.SetupFiscalYearPrevious2(wMonth, wYear : Word);
begin
   // Note:  30 month cap prevents this option from being always shown
   // Fiscal Year <CurrentYear-2>
   // Start Date: 10/1/<CurrentYear-3>
   // Stop Date:  9/30/<CurrentYear-2>
   FiscalYearPrevious2_StartDate := EncodeDate(wYear-3, 10, 1);
   FiscalYearPrevious2_EndDate := EncodeDate(wYear-2,9,30);
   FiscalYearPrevious2Text := 'Fiscal Year ' + IntToStr(wYear-2);
   // calculate except if date exceeds the 30 month cap
   if (wMonth >= 4) AND (wMonth <= 9) then  // April through September
   begin
      FiscalYearPrevious2_StartDate := ErrorDate;
      FiscalYearPrevious2_EndDate := ErrorDate;
      FiscalYearPrevious2Text := '';  // Months April-Sept of the current year will exceed the 30 month gap, so don't add text             
   end;   
   SetupFiscalYearAll;  // Always must be called after above routine
end;

end.
