unit XmlComponentBindings;

interface

uses Classes, Controls, StdCtrls, SysUtils, XsdComplexType, ExtCtrls,
     CPWMRadioGroup, CPWMObjects, CheckLst, XmlConfig, XMLIntf, XMLDoc;

type
  TXmlComponentBindings = class
    private
      XmlConfig: TXmlConfig;
      NsPrefix: String;
      ParentNode: IXMLNode;
      function getEditXml(complexType: TXsdComplexType; control: TEdit): IXMLNode;
      function getMemoXml(complexType: TXsdComplexType; control: TMemo): IXMLNode;
      function getCheckBoxXml(complexType: TXsdComplexType; control: TCheckBox): IXMLNode;
      function getComboBoxXml(complexType: TXsdComplexType; control: TComboBox): IXMLNode;
      function getListBoxXml(complexType: TXsdComplexType; control: TListBox): IXMLNode;
      function getRadioGroupXml(complexType: TXsdComplexType; control: TRadioGroup): IXMLNode;
      function getCPWMRadioGroupXml(complexType: TXsdComplexType; control: TCPWMRadioGroup): IXMLNode;
      function getRadioButtonXml(complexType: TXsdComplexType; control: TRadioButton): IXMLNode;
      function getCheckListBoxXml(complexType: TXsdComplexType; control: TCheckListBox): IXMLNode;
    public
      function AddXmlNode(component: TComponent; complexType: TXsdComplexType): IXMLNode;
    published
      constructor Create(theXmlConfig: TXmlConfig; theParentNode: IXMLNode);
  end;

implementation

constructor TXmlComponentBindings.Create(theXmlConfig: TXmlConfig; theParentNode: IXMLNode);
begin
  XmlConfig := theXmlConfig;
  NsPrefix := XmlConfig.GetValue(TXmlConfig.NAMESPACE_PREFIX);
  ParentNode := theParentNode;
end;

function TXmlComponentBindings.AddXmlNode(component: TComponent; complexType: TXsdComplexType): IXMLNode;
var
  control: TControl;
begin
  try
    control := (component as TControl);
    if control is TEdit then
      Result := getEditXml(complexType, control as TEdit)
    else if control is TMemo then
      Result := getMemoXml(complexType, control as TMemo)
    else if control is TCheckBox then
      Result := getCheckBoxXml(complexType, control as TCheckBox)
    else if control is TComboBox then
      Result := getComboBoxXml(complexType, control as TComboBox)
    else if control is TListBox then
      Result := getListBoxXml(complexType, control as TListBox)
    else if control is TRadioGroup then
      Result := getRadioGroupXml(complexType, control as TRadioGroup)
    else if control is TCPWMRadioGroup then
      Result := getCPWMRadioGroupXml(complexType, control as TCPWMRadioGroup)
    else if control is TRadioButton then
      Result := getRadioButtonXml(complexType, control as TRadioButton)
    else if control is TCheckListBox then
      Result := getCheckListBoxXml(complexType, control as TCheckListBox);
  except
    Exit;
  end;
end;

function TXmlComponentBindings.getEditXml(complexType: TXsdComplexType; control: TEdit): IXMLNode;
begin
  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);

  if (Length(control.Text) = 0) and (complexType.RequiredXml = False) then
    Result.AddChild(NsPrefix + ':Value').Attributes['xsi:nil'] := 'true'
  else
    Result.AddChild(NsPrefix + ':Value').NodeValue := control.Text;

  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;

function TXmlComponentBindings.getMemoXml(complexType: TXsdComplexType; control: TMemo): IXMLNode;  
begin
  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);

  if (Length(control.Text) = 0) and (complexType.RequiredXml = False) then
    Result.AddChild(NsPrefix + ':Value').Attributes['xsi:nil'] := 'true'
  else
    Result.AddChild(NsPrefix + ':Value').NodeValue := Trim(control.Text);

  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;

function TXmlComponentBindings.getCheckBoxXml(complexType: TXsdComplexType; control: TCheckBox): IXMLNode;
begin
  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);
  Result.AddChild(NsPrefix + ':Value').NodeValue := control.Checked;
  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;

function TXmlComponentBindings.getComboBoxXml(complexType: TXsdComplexType; control: TComboBox): IXMLNode;
begin
  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);

  if (Trim(control.Text) = '') and (complexType.RequiredXml = False) then
    Result.AddChild(NsPrefix + ':Value').Attributes['xsi:nil'] := 'true'
  else
    Result.AddChild(NsPrefix + ':Value').NodeValue := control.Text;
    
  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;

function TXmlComponentBindings.getListBoxXml(complexType: TXsdComplexType; control: TListBox): IXMLNode;
var
  i: Integer;
  valuesNode: IXMLNode;
begin
  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);
  valuesNode := Result.AddChild(NsPrefix + ':Values');

  with valuesNode do
  begin
    for i := 0 to control.Items.Count - 1 do
    begin
      if control.Selected[i] = True then
        AddChild('Selected').NodeValue := control.Items[i];
    end;

    if (valuesNode.ChildNodes.Count = 0) and (complexType.RequiredXml = False) then
      AddChild('Selected').Attributes['xsi:nil'] := 'true'

  end;

  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;

function TXmlComponentBindings.getRadioGroupXml(complexType: TXsdComplexType; control: TRadioGroup): IXMLNode;
var
  value: String;
begin
  if control.ItemIndex >= 0 then
    value := control.Items[control.ItemIndex]
  else
    value := '';

  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);

  if (complexType.RequiredXml = False) and (value = '') then
     Result.AddChild(NsPrefix + ':Value').Attributes['xsi:nil'] := 'true'
  else
    Result.AddChild(NsPrefix + ':Value').NodeValue := value;

  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;

function TXmlComponentBindings.getCPWMRadioGroupXml(complexType: TXsdComplexType; control: TCPWMRadioGroup): IXMLNode;
var
  value: String;
begin
  if control.ItemIndex >= 0 then
    value := control.Items[control.ItemIndex]
  else
    value := '';

  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);

  if (complexType.RequiredXml = False) and (value = '') then
     Result.AddChild(NsPrefix + ':Value').Attributes['xsi:nil'] := 'true'
  else
    Result.AddChild(NsPrefix + ':Value').NodeValue := value;

  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;

function TXmlComponentBindings.getRadioButtonXml(complexType: TXsdComplexType; control: TRadioButton): IXMLNode;
begin
  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);
  Result.AddChild(NsPrefix + ':Value').NodeValue := control.Checked;
  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;

function TXmlComponentBindings.getCheckListBoxXml(complexType: TXsdComplexType; control: TCheckListBox): IXMLNode;
var
  i: Integer;
  valuesNode: IXMLNode;
begin
  Result := ParentNode.AddChild(NsPrefix + ':' + complexType.XmlTag);
  valuesNode := Result.AddChild(NsPrefix + ':Values');

  with valuesNode do
  begin
    for i := 0 to control.Items.Count - 1 do
    begin
      if control.Checked[i] = True then
        AddChild('Selected').NodeValue := control.Items[i];
    end;

    if (valuesNode.ChildNodes.Count = 0) and (complexType.RequiredXml= False) then
      AddChild('Selected').Attributes['xsi:nil'] := 'true'
      
  end;

  with Result.AddChild(NsPrefix + ':Metadata') do
  begin
    AddChild(NsPrefix + ':code').NodeValue := complexType.CodeAttribute;
    AddChild(NsPrefix + ':codeSystem').NodeValue := complexType.CodeSystemAttribute;
    AddChild(NsPrefix + ':version').NodeValue := complexType.VersionAttribute;
    AddChild(NsPrefix + ':source').NodeValue := complexType.SourceAttribute;
  end;
end;


end.
