unit XmlEngine;

interface

uses Controls, Classes, ExtCtrls, CPWMObjects, XmlConfig, XsdEngineLite,
  XsdComplexType, VlerDasClaim, XmlComponentBindings, XMLIntf, XMLDoc, SysUtils,
  Main, Windows, Trpcb, MFunStr, Patient, Fmcmpnts;

type
  TApprover = record
    ApproverID : String;                      // CodeCR563 JRL 6/9/14
    ApproverAssigningAuthority : String;      // CodeCR563 JRL 6/9/14
    ApproverLastName : String;                // CodeCR563 JRL 6/9/14
    ApproverFirstName : String;               // CodeCR563 JRL 6/9/14
    ApproverMiddleName : String;              // CodeCR563 JRL 6/9/14
  end;
  TExamInfo = record
    PhysicalSource : String;                  // CodeCR563 JRL 6/9/14
    ExamID : String;                          // CodeCR563 JRL 6/9/14
    ExaminerType : String;                    // CodeCR563 JRL 6/9/14
    ContractType : String;                    // CodeCR563 JRL 6/9/14
    StatusDate : String;                      // CodeCR563 JRL 6/9/14
    Comments : String;                        // CodeCR563 JRL 6/9/14
    StatusText : String;                      // CodeCR563 JRL 6/10/14
  end;


  TXmlEngine = class
    private
      Exam: TCPWMExam;
      XmlConfig: TXmlConfig;
      XsdEngine: TXsdEngineLite;

      VlerDasClaim: TVlerDasClaim;
      AttachmentsList: TList; //of TVlerDasAttachment
      DocumentIdentificationID: String;
      Patient: TPatient;
      //function GetStringGUID: string;                 // DONE: LMS moved generic function
      //function GetXusUserInfo: TStrings;              // DONE: LMS moved generic function
    public
      ExamInfo : TExamInfo;
      Approver : TApprover;
      function RenderToXml(rootPanel: TPanel; includeAttachments: Boolean = true): IXMLDocument;
      procedure AddClaimAttachment(attachment: TVlerDasAttachment);
      procedure SetPatient(thePatient: TPatient);
      function GetDocumentIdentificationID: String;
      function GetDocumentTitleText: String;
      function GetUniqueFilename(fileName: String; fileExtension: String): String;
      function GetW3cDateTime: String;
      // function GetTempDirectory: String;   // CodeCR496 - LMS - 2014-07-14 refactor GetTempDir to one method call
      procedure PopulateApproverInfo(AppID, AppAssnAuth, AppLast, AppFirst, AppMid : String);
      procedure PopulateExamInfo(ExamPhysSrc, ExamID, ExamType, ConType, StatDt, Comments, StatusTxt : String);
    published
      constructor Create(theExam: TCPWMExam);
  end;

implementation

uses untConstVals
     , untMiscMthds
     , CAPRISupport
     ;

constructor TXmlEngine.Create(theExam: TCPWMExam);
begin
  Exam := theExam;
  XmlConfig := theExam.GetXmlConfig;
  XsdEngine := theExam.GetXsdEngine;

  VlerDasClaim := TVlerDasClaim.Create(XsdEngine, XmlConfig);
end;

function TXmlEngine.RenderToXml(rootPanel: TPanel; includeAttachments: Boolean = true): IXMLDocument;
var
  XusGetUserInfoResult: TStrings;
begin
  VlerDasClaim.SetRootPanel(rootPanel);

  VlerDasClaim.SetXmlNamespace(XmlConfig.GetValue(XmlConfig.NAMESPACE_PREFIX),
    XmlConfig.GetValue(XmlConfig.NAMESPACE));

  if includeAttachments = True then
    VlerDasClaim.SetAttachments(AttachmentsList)
  else
    VlerDasClaim.SetAttachments(Nil);

  VlerDasClaim.DocumentTypeVersion := XmlConfig.GetValue(TXmlConfig.VERSION);
  VlerDasClaim.DocumentVersion := XmlConfig.GetValue(TXmlConfig.VERSION);
  VlerDasClaim.DocumentTitleText := XmlConfig.GetValue(TXmlConfig.DOCUMENT_TITLE_TEXT);

  DocumentIdentificationID := GetStringGUID; //* web service invocation also needs DocumentIdentificationID */
  VlerDasClaim.DocumentIdentificationID := DocumentIdentificationID;
  VlerDasClaim.DocumentCreationDate := GetW3cDateTime;

  VlerDasClaim.ClaimID := Patient.SocialSecurityNumber;

  if Length(Patient.IntegrationControlNumber) > 0 then
  begin
    VlerDasClaim.ClientIdentificationID := Patient.IntegrationControlNumber;
    VlerDasClaim.ClientAssigningAuthority := CLIENT_ASSIGNING_AUTHORITY_VA;
  end
  else
  begin
    VlerDasClaim.ClientIdentificationID := Patient.SocialSecurityNumber;
    VlerDasClaim.ClientAssigningAuthority := CLIENT_ASSIGNING_AUTHORITY_SSA;
  end;

  XusGetUserInfoResult := GetXusUserInfo; //* calls RPC: XUS GET USER INFO */

  VlerDasClaim.ServiceProviderIdentificationID := XusGetUserInfoResult[0];
  VlerDasClaim.ServiceProviderAssigningAuthority := CLIENT_ASSIGNING_AUTHORITY_VA;
  VlerDasClaim.ServiceProviderRoleTitle := XusGetUserInfoResult[4];

  VlerDasClaim.PatientFirstName := Patient.FirstName;
  VlerDasClaim.PatientMiddleName := Patient.MiddleName;
  VlerDasClaim.PatientLastName := Patient.LastName;
  VlerDasClaim.PatientGender := Patient.Gender;
  VlerDasClaim.PatientSSN := Patient.SocialSecurityNumber;
  VlerDasClaim.PatientDateOfBirth := Patient.Birthdate;

//  VlerDasClaim.ProviderFirstName := Piece(XusGetUserInfoResult[1], ',', 2);
  VlerDasClaim.ProviderFirstName := Piece(Piece(XusGetUserInfoResult[1], ',', 2), ' ', 1);  // JRL 8/4/14
  VlerDasClaim.ProviderMiddleName := Piece(Piece(XusGetUserInfoResult[1], ',', 2), ' ', 2);
  VlerDasClaim.ProviderLastName := Piece(XusGetUserInfoResult[1], ',');

  VlerDasClaim.FacilityIdentificationID := Piece(XusGetUserInfoResult[3], '^', 3);
  VlerDasClaim.FacilityIdentificationText := Uppercase(frmMain.RPCBroker1.ANUStrServer); //Piece(XusGetUserInfoResult[3], '^', 2);
  VlerDasClaim.FacilityName := Uppercase(frmMain.RPCBroker1.ANUStrServer); //Piece(XusGetUserInfoResult[3], '^', 2);

  VlerDasClaim.ApproverID := Approver.ApproverID;                             // CodeCR563 JRL 6/10/14
  VlerDasClaim.ApproverAssigningAuthority := Approver.ApproverAssigningAuthority; // CodeCR563 JRL 6/10/14
  VlerDasClaim.ApproverLastName := Approver.ApproverLastName;                 // CodeCR563 JRL 6/10/14
  VlerDasClaim.ApproverFirstName := Approver.ApproverFirstName;               // CodeCR563 JRL 6/10/14
  VlerDasClaim.ApproverMiddleName := Approver.ApproverMiddleName;             // CodeCR563 JRL 6/10/14

  VlerDasClaim.ExamID := ExamInfo.ExamID;                                     // CodeCR563 JRL 6/10/14
  VlerDasClaim.PhysicalSource := ExamInfo.PhysicalSource;                     // CodeCR563 JRL 6/10/14
  VlerDasClaim.ExaminerType := ExamInfo.ExaminerType;                         // CodeCR563 JRL 6/10/14
  VlerDasClaim.ContractType := ExamInfo.ContractType;                         // CodeCR563 JRL 6/10/14
  VlerDasClaim.StatusDate := GetW3cDateTime;                                  // CodeCR563 JRL 6/10/14
  VlerDasClaim.Comments := ExamInfo.Comments;                                 // CodeCR563 JRL 6/10/14
  VlerDasClaim.StatusText := ExamInfo.StatusText;                             // CodeCR563 JRL 6/10/14

  Result := VlerDasClaim.ToXml;
end;

procedure TXmlEngine.SetPatient(thePatient: TPatient);
begin
  Patient := thePatient;
end;

function TXmlEngine.GetDocumentIdentificationID: String;
begin
  Result := DocumentIdentificationID;
end;

function TXmlEngine.GetDocumentTitleText: String;
begin
  Result := XmlConfig.GetValue(TXmlConfig.DOCUMENT_TITLE_TEXT);
end;

{
  Return a unique filename using given fileExtension (example - 'txt' or 'pdf' not '.txt' or '.pdf')
}
function TXmlEngine.GetUniqueFilename(fileName: String; fileExtension: String): String;
var
  i: Integer;
  intChar: Integer;
begin
  Result := fileName;

  for i := 1 to Length(Result) do
  begin
    intChar := Ord(Result[i]);
    if ((intChar >= 48) and (intChar <= 57)) or
       ((intChar >= 65) and (intChar <= 90)) or
       ((intChar >= 97) and (intChar <= 122)) then
    begin
       {then the char is [0-9][a-z][A-Z]}
    end
    else
      Result[i] := '_';
  end;

  Result := Result + IntToStr(Random(High(Integer))) + '.' + fileExtension;
end;

(*  // CodeCR496 - LMS - 2014-07-14 consolidate/refactor all GetTempDir into one for Win 7 specific handling
  function TXmlEngine.GetTempDirectory: String;
  var
    tempFolder: array[0..MAX_PATH] of Char;
  begin
    GetTempPath(MAX_PATH, @tempFolder);
    result := StrPas(tempFolder);
  end;
*)

function TXmlEngine.GetW3cDateTime: String;
var
  sysTime: TSystemTime;
  month: String;
  day: String;
  hour: String;
  minute: String;
  second: String;
begin
  Windows.GetSystemTime(sysTime);

  month := IntToStr(sysTime.wMonth);
  if Length(month) = 1 then
    month := '0' + month;

  day := IntToStr(sysTime.wDay);
  if Length(day) = 1 then
    day := '0' + day;

  hour := IntToStr(sysTime.wHour);
  if Length(hour) = 1 then
    hour := '0' + hour;

  minute := IntToStr(sysTime.wMinute);
  if Length(minute) = 1 then
    minute := '0' + minute;

  second := IntToStr(sysTime.wSecond);
  if Length(second) = 1 then
    second := '0' + second;

  Result := IntToStr(sysTime.wYear)+'-'+month+'-'+day+'T'+hour+':'+
              minute+':'+second+'Z';
end;

procedure TXmlEngine.AddClaimAttachment(attachment: TVlerDasAttachment);
begin
  if AttachmentsList = Nil then
    AttachmentsList := TList.Create;

  AttachmentsList.Add(attachment);
end;

//  procedure TXmlEngine.SetClaimAttachments(listOfAttachments: TList); //of TVlerDasAttachment
//  begin
//    AttachmentsList := listOfAttachments;
//  end;

(*  // DONE : LMS Unnecessary duplicate code
function TXmlEngine.GetStringGUID: string;
var
  lGUID:            TGUID;
  lGUIDStr:         string;
begin
  CreateGUID(lGUID);
  lGUIDStr := GUIDToString(lGUID);
  lGUIDStr := StringReplace(lGUIDStr, '{', '', [rfReplaceAll, rfIgnoreCase]);
  lGUIDStr := StringReplace(lGUIDStr, '}', '', [rfReplaceAll, rfIgnoreCase]);
  result := lGUIDStr;
end;
*)


{-------------------------------------------------------------------------------
  Added routine to allow population of these fields from TIUSign
  CodeCR563 JRL 6/13/14
-------------------------------------------------------------------------------}
procedure TXmlEngine.PopulateApproverInfo(AppID, AppAssnAuth, AppLast, AppFirst, AppMid : String);
begin
  Approver.ApproverID := AppID;
  Approver.ApproverAssigningAuthority := AppAssnAuth;
  Approver.ApproverLastName := AppLast;
  Approver.ApproverFirstName := AppFirst;
  Approver.ApproverMiddleName := AppMid;
end;


{-------------------------------------------------------------------------------
  Added routine to allow population of these fields from TIUSign
  CodeCR563 JRL 6/13/14
-------------------------------------------------------------------------------}
procedure TXmlEngine.PopulateExamInfo(ExamPhysSrc, ExamID, ExamType, ConType, StatDt, Comments, StatusTxt : String);
begin
  ExamInfo.PhysicalSource := ExamPhysSrc;
  ExamInfo.ExamID := ExamID;
  ExamInfo.ExaminerType := ExamType;
  ExamInfo.ContractType := ConType;
  ExamInfo.StatusDate := StatDt;
  ExamInfo.Comments := Comments;
  ExamInfo.StatusText := StatusTxt;
end;

end.
