unit clsVendConn;

interface

uses Classes, Forms
     , CCOWRPCBrokerCAPRI
     , Fmcmpnts
     ;

type TVendorConnect = class(TComponent)
  private
    FRemoteVendor: string;
    FVLEREnabled: String;
    FHaveVendorInfo: Boolean;
    FVendorName: string;
    procedure SetRemoteVendor(const Value: string);
    function GetVLEREnabled: String;
    procedure SetVLEREnabled(const Value: String);
    procedure SetHaveVendorInfo(const Value: Boolean);
  { Private Declarations }
  protected
    FCCOWBrkr: TCCOWRPCBrokerCAPRI;
    FVndrDfltPth: String;
    FDwnLdFiles: Boolean;
    FRemoteIEN: String;
    fmInfoGets: TFMGets;
    FUserName: String;
    FPassword: String;
    FAddress: String;
    FPort: Integer;
    function GetVndrConnInfo(ARemoteVendor: String): Boolean; virtual;
  public
  { Public Declarations }
    (*constructor Create(AOwner: TComponent); overload; override;*)
    constructor Create(RmtVndrIEN: String; DownLoadFiles: Boolean; AOwnerBroker: TCCOWRPCBrokerCAPRI); reintroduce; overload; virtual;
    destructor Destroy; override;
    function GetFileDirContents: TList; virtual;
    function DownloadFiles(VndrFles: TStrings): Boolean; virtual;
    procedure ArchiveFiles(VndrFles: TStrings); virtual;
    procedure UploadFile(UpldExmFle: String); virtual;
    property RemoteVendor : string read FRemoteVendor write SetRemoteVendor;
    property CCOWRPCBroker: TCCOWRPCBrokerCAPRI read FCCOWBrkr;
    property RemoteIEN: String read FRemoteIEN;
    property DwnLdFiles: Boolean read FDwnLdFiles;
    property VLEREnabled: String read GetVLEREnabled write SetVLEREnabled;
    property HaveVendorInfo: Boolean read FHaveVendorInfo write SetHaveVendorInfo;
    property VendorName: string read FVendorName;
  end;


implementation

uses sysutils, strutils, main
     , untBrkrMthds
     , Trpcb
     , MFunStr
     ;

(*
{ TVendorConnect }
type
  TVendInfo = record
    CCOWBrkr: TCCOWRPCBrokerCAPRI;
    RemoteIEN: String;
    UserName: String;
    Password: String;
    Address: String;
    Port: Integer;
    VLEREnabled: String;
  end;

var
  VendInfo : TVendInfo;
  //VendInfoArr : Array of VendInfo;  // Future use
*)

procedure TVendorConnect.ArchiveFiles(VndrFles: TStrings);
begin
  EnsureBrokerConnected(CCOWRPCBroker);      // SFTP operations are prone to time outs. VLER could too.. This is preventative.
end;

constructor TVendorConnect.Create(RmtVndrIEN: String; DownLoadFiles: Boolean; AOwnerBroker: TCCOWRPCBrokerCAPRI);
begin
  inherited Create(AOwnerBroker);
  FCCOWBrkr := AOwnerBroker;
  RemoteVendor := RmtVndrIEN;
  FHaveVendorInfo := GetVndrConnInfo(RmtVndrIEN);
end;

(*
constructor TVendorConnect.Create(AOwner: TComponent);
begin
  inherited;

end;
*)

destructor TVendorConnect.Destroy;
begin
  inherited;
end;

function TVendorConnect.DownloadFiles(VndrFles: TStrings): Boolean;
begin
  Result := EnsureBrokerConnected(CCOWRPCBroker);      // SFTP operations are prone to time outs. VLER could too.. This is preventative.
end;

function TVendorConnect.GetFileDirContents: TList;
begin
  EnsureBrokerConnected(CCOWRPCBroker);      // SFTP operations are prone to time outs. VLER could too.. This is preventative.
  Result := nil;
end;

function TVendorConnect.GetVLEREnabled: String;
begin
  if not FHaveVendorInfo then
    if FRemoteVendor <> '' then
      FHaveVendorInfo := GetVndrConnInfo(FRemoteVendor);

  {
  if FHaveVendorInfo then
    Result := FVLEREnabled
  else
    Result := '';
  }
  if FVLEREnabled = '' then
    FVLEREnabled := 'N';
  Result := FVLEREnabled;
end;

function TVendorConnect.GetVndrConnInfo(ARemoteVendor: String): Boolean;
var
  i: Integer;
  ClmsSysCCOWBrkr: TCCOWRPCBrokerCAPRI;
begin
  Result := False;
  ClmsSysCCOWBrkr := TCCOWRPCBrokerCAPRI.Create(Self {may be behind AV/s CCOWRPCBroker});  { Remote Broker Object }
  try
    { Use BSE to Connect to CLAIMS/TESTCLAIMS to retrieve Vendor Info }
    if RemoteClaimsLogin(CCOWRPCBroker, ClmsSysCCOWBrkr) then
    begin
      { Retrieve Specific Vendor SFTP Info from remote system }
      fmInfoGets := TFMGets.Create(ClmsSysCCOWBrkr);
      fmInfoGets.RPCBroker := ClmsSysCCOWBrkr;
      fmInfoGets.IENS := ARemoteVendor;
      fmInfoGets.FieldNumbers.Clear;
      { Add SFTP Fields to Retrieve from 2507 EXAM CONTRACTORS File #396.45 }

      for i := 100 to 105 do           // LMS - CodeCR??? - Add support for VLER
          // done: lms remove Workaround the VLEREnabled field not existing in TESTCLAIMS.
      //for i := 100 to 104 do           // LMS - CodeCR??? - Add support for VLER
      begin
        fmInfoGets.FieldNumbers.Add(IntToStr(i));
      end;
      fmInfoGets.FieldNumbers.Add('.01');  // LMS - Also using Name for Physical Source

      fmInfoGets.FileNumber := '396.45';
      try
        try
          fmInfoGets.GetData;
          if fmInfoGets.ErrorList.Count > 0 then begin
            result := false;
            exit;
          end;

          //if (fmInfoGets.GetField('105').FMDBInternal = 'Y') then   //
          //  exit;  // It's a VLER Enabled vendor.

          FVendorName := piece(fmInfoGets.GetField('.01').FMDBInternal, ' ');  // remove any test claims or other string after the name initials
          FUserName := fmInfoGets.GetField('100').FMDBInternal;
          FPassword := fmInfoGets.GetField('101').FMDBInternal;
          FAddress := fmInfoGets.GetField('102').FMDBInternal;
          FPort := StrToIntDef(fmInfoGets.GetField('103').FMDBInternal, 0);
          FVndrDfltPth := fmInfoGets.GetField('104').FMDBInternal;

          FVLEREnabled := fmInfoGets.GetField('105').FMDBInternal;

          // done: lms remove Workaround the VLEREnabled field not existing in TESTCLAIMS.
          {if (ARemoteVendor = '3')    // LHI   MSLA  VES
           or (ARemoteVendor = '4')
           or (ARemoteVendor = '1')
            then begin
            FVLEREnabled := 'Y';
          end;}

          if (FVLEREnabled = 'N') or (FVLEREnabled = '') then begin  //
            { UserName/Password are encrypted during entry onto the CLAIMS systems,
              therefore they must be decrypted before they can be used. }
            ClmsSysCCOWBrkr.Results.Clear;
            ClmsSysCCOWBrkr.RemoteProcedure := 'DVBAD CONTRACTED EXAM CRYPTO';
            ClmsSysCCOWBrkr.Param[0].Value := '2';  { Decryption Algorithm }
            ClmsSysCCOWBrkr.Param[0].PType := literal;
            ClmsSysCCOWBrkr.Param[1].Value := FUsername + '^' + FPassword;  { Values to Decrypt }
            ClmsSysCCOWBrkr.Param[1].PType := literal;
            ClmsSysCCOWBrkr.Call;  { Execute RPC - Results returned in same location as sent }
            FUserName := Piece(ClmsSysCCOWBrkr.Results[0], '^', 1);
            FPassword := Piece(ClmsSysCCOWBrkr.Results[0], '^', 2);
          end;
          Result := True;
        except  { Do Nothing on Exception }
        end;  { End Inner Try/Except Block }
      finally
        fmInfoGets.Free;
      end;  { End Inner Try/Finally Block }
    end;  { End If Successfully connected to Remote CLAIMS system }
  finally
    ClmsSysCCOWBrkr.Connected := False;
    ClmsSysCCOWBrkr.Free;
  end;  { End Outer Try/Finally Block }
end;

procedure TVendorConnect.SetHaveVendorInfo(const Value: Boolean);
begin
  FHaveVendorInfo := Value;
end;

procedure TVendorConnect.SetRemoteVendor(const Value: string);
begin
  if (FRemoteVendor <> Value) then begin
    FRemoteVendor := Value;
    FHaveVendorInfo := false;
    FVendorName := '';
  end;
end;

procedure TVendorConnect.SetVLEREnabled(const Value: String);
begin
  FVLEREnabled := Value;
end;

procedure TVendorConnect.UploadFile(UpldExmFle: String);
begin
  EnsureBrokerConnected(CCOWRPCBroker);      // SFTP operations are prone to time outs. VLER could too.. This is preventative.
end;

end.
