Unit CPWMObjects;

Interface
Uses
  Windows,
  buttons,
  CAPRISupport,
  Classes,
  ColorCapriBtn,
  comctrls,
  Controls,
  CPWMRadioGroup,
  Dialogs,
  DialogsCAPRI,
  extctrls,
  forms,
  Graphics,
  HyperlinkLabel,
  Menus,
  ORCtrls,
  PNCSCodeRun,
  StdCtrls,
  SysUtils,
  tabnotbk,
  XmlConfig,
  XsdEngineLite;

Type
  TCPWMForm = Class; // forwared to the definition further below
  TCPWMExam = Class; // forwared to the TCPWMExam definition further below
  TCPWMObject = Class(TObject)
    ObjectText: String;
    Lines: TStringList;
    Children: TList;
    screenObject: TControl;
  Private
    fBeenRendered  : boolean;
    fCaption       : String;
    fControlType   : String; // e.g. TTabPage, TButton ....
    fDeleted       : boolean;
    fDepth         : integer;
    fExamObject    : TCPWMExam;
    fFormObject    : TCPWMForm;
    fHintValue     : String;
    fInserted      : boolean;
    fIsTabScrollBox: boolean;
    fIsUserObject  : boolean;
    fLeft          : integer;
    fName          : String;
    fOnTabName     : String;
    fParent        : TCPWMObject;
    fPassCount     : integer;
    fProcessed     : boolean;
    fTabOrder      : integer;
    fTop           : integer;
    Function FindUserObject(IDNum: String; childen: TList): TCPWMObject;
    Function GetCaption: String;
    Function GetControlHint: String;
    Function GetName: String;
    Function GetObjectTree(aObject: TCPWMObject): TCPWMObject;
    Function GetParentName: String;
    Function GetUserObject(childen: TList): TCPWMObject;
    Function IsHasChildren: boolean;
    Function ParseObjectProperty(propToParse: String): String;
    Function StringToComponent(Value: String; ParentControl: TControl): TComponent;
    Procedure CountUserObjects(aTag: LongInt; Var userObjectList: TList);
    Procedure GetScreenLocation;
    Procedure SetName(aName: String);
    Procedure SetOnTabName(tabPageName: String; childen: TList);
    Procedure SetParentage(baseParent: TWinControl);
    Procedure SetTabOrder;
    Procedure RenderToPrinter(baseForm: TComponent; UserObjectList,
      ScreenObjectList: TList; OnClick, OnChange: TNotifyEvent; aOwner,
      aParent: TWinControl; OnHint: TMouseMoveEvent; PopupEditMenu: TPopupMenu; PopupButtonMenu: TPopupMenu);
  Public
    OnPopUp: boolean;
    PopUp  : boolean;
    Tag    : Longint; // For Misc use
    Constructor Create;
    Destructor Destroy; Override;
    Function CountScreenObjects: integer;
    Procedure ExamineObject;
    Property BeenRendered: boolean Read fBeenRendered Write fBeenRendered;
    Property Caption: String Read GetCaption Write fCaption;
    Property ControlType: String Read fControlType Write fControlType; // e.g. TTabPage, TButton ....
    Property Deleted: boolean Read fDeleted Write fDeleted;
    Property Depth: integer Read fDepth Write fDepth;
    Property Exam: TCPWMExam Read fExamObject Write fExamObject;
    Property Form: TCPWMForm Read fFormObject Write fFormObject;
    Property HasChildren: boolean Read IsHasChildren;
    Property HintValue: String Read fHintValue Write fHintValue;
    Property Inserted: boolean Read fInserted Write fInserted;
    Property IsTabScrollBox: boolean Read fIsTabScrollBox Write fIsTabScrollBox;
    Property IsUserObject: boolean Read fIsUserObject Write fIsUserObject;
    Property Left: integer Read fLeft Write fLeft;
    Property Name: String Read GetName Write SetName;
    Property OnTabName: String Read fOnTabName Write fOnTabName;
    Property Parent: TCPWMObject Read fParent Write fParent;
    Property ParentName: String Read GetParentName;
    Property PassCount: integer Read fPassCount Write fPassCount;
    Property Processed: boolean Read fProcessed Write fProcessed;
    Property TabOrder: integer Read fTabOrder Write fTabOrder;
    Property Top: integer Read fTop Write fTop;
    Procedure Render(baseForm: TComponent; UserObjectList: TList; ScreenObjectList: TList; OnClick: TNotifyEvent; OnChange: TNotifyEvent; aOwner: TWinControl; aParent: TWinControl; OnHint:
      TMouseMoveEvent; PopupEditMenu: TPopupMenu; PopupButtonMenu: TPopupMenu);
  End;

  TTypeOfForm = (Popup, Notebook, TypeOfForm, FormID);

  TCPWMForm = Class(TObject) // a Popup or (the) TTabbed Notebook
    Name: String; // right of the base 'object'
    Caption: String; //
    Children: TList; // the VCLControls that make up this form. Usually a panel
    Tabs: TList; // tabs on this form ( a notebook )
  Private
    fExamObject: TCPWMExam;
    BeenRendered: boolean;
    Procedure SetTabs(childen: TList);
    Procedure CountUserObjects(aTag: LongInt; Var userObjectList: TList);
    Procedure SetTabOrder;
  Protected
    Procedure Reposition(baseControl: TWinControl; buttonBarBottom: integer);
  Public
    FormType: TTypeOfForm;
    Property Exam: TCPWMExam Read fExamObject Write fExamObject;
    Constructor Create;
    Destructor Destroy; Override;
    Procedure SendToBack;
    Procedure Render(baseForm: TComponent; UserObjectList: TList; ScreenObjectList: TList; OnClick: TNotifyEvent; OnChange: TNotifyEvent; baseControl: TWinControl; buttonBarBottom: integer; OnHint:
      TMouseMoveEvent; PopupEditMenu: TPopupMenu; PopupButtonMenu: TPopupMenu);
    Procedure RenderToPrinter(baseForm: TComponent; UserObjectList: TList; ScreenObjectList: TList; OnClick: TNotifyEvent; OnChange: TNotifyEvent; baseControl: TWinControl; buttonBarBottom: integer;
      OnHint: TMouseMoveEvent; PopupEditMenu: TPopupMenu; PopupButtonMenu: TPopupMenu);
    Function CountScreenObjects: integer;
  End;

  TCPWMCode = Class; // forward declaration

  TCPWMExam = Class(TObject) // An entire exam
    Name: String; // right of the base 'object'
    ButtonText: String; // Use for the button caption when rendering on button bars
    FWidth : Integer;   // Dimensions defined for template form                 //CodeCR458 JRL 5/23/13
    FHeight : Integer;                                                          //CodeCR458 JRL 5/23/13
    Caption: String; //
    Tabs: TList; // SuperSet of all the form tabs for the whole exam
    UserObjectList: TList;
    ScreenObjectList: TList;
    Properties: TStringList;
  Private
    XmlConfig: TXmlConfig;
    XsdEngine: TXsdEngineLite;
    tabNames: TStringList;
    Function ParseForm(objectTextToAdd: String): TCPWMForm;
    Function ParseObjects(lines: TStringList;
      parent: TCPWMObject): TCPWMObject;
    Function GetEndDepth(aLine: String): integer;
    Function GetObjectDepth(aLine: String): integer;
    Function IsEndOfDefinition(aLine: String): boolean;
    Function IsObjectDefinition(aLine: String): boolean;
    Function DetermineExamName: String;
    Function GetNotebook: TCPWMForm;
    Procedure ScriptMerge(aExam: TCPWMExam; currentButtonName: String);
    Procedure OrderTabNames;
    Function ReadProperties: String;
    Function GetProperty(propertyName: String): String;
    Function NumericString(line: String): boolean;
    Procedure SetTabOrder;
  Public
    Forms: TList; // of TCPWMForm // the set of all forms ( TTypeofForm )
    Code: TCPWMCode;
    aMetaNotebook: TTabbedNotebook;
    Procedure ExamNotebookTabChange(Sender: TObject; NewTab: Integer;
      Var AllowChange: Boolean);
    Procedure Merge(aExam: TCPWMExam);
    Procedure ParseExamFile(fileLines: TStrings);
    Procedure ParseExamCode(script: TStrings);
    Function GetXmlConfig: TXmlConfig;
    Function GetXsdEngine: TXsdEngineLite;
    Function IsXmlBasedExam: Boolean;
    Function GetButtonBarCaption: String;
    Constructor Create(theXmlMapping: TStringList; theXmlConfig: TStringList);
    Destructor Destroy; Override;
    Procedure Render(AOwner: TComponent; OnClick: TNotifyEvent; OnChange: TNotifyEvent; baseControl: TWinControl; OnHint: TMouseMoveEvent; ExamList: TList; PopupEditMenu: TPopupMenu; PopupButtonMenu:
      TPopupMenu);
    Procedure PostProcessScript(ExamList: TList);
{$IFDEF STANDALONE}
    Procedure PrintPopups(AOwner: TComponent; OnClick: TNotifyEvent; OnChange: TNotifyEvent; baseControl: TWinControl; ToPrinter: boolean; OnHint: TMouseMoveEvent);
{$ENDIF}
    Function CountScreenObjects: integer;
    Procedure CountUserObjects(aTag: LongInt; Var userObjectList: TList);
    Procedure ExamButtonClick(Sender: TObject);
  End;

  TCPWMCode = Class(TObject) // The exam script
    FirstRun: TStringList;
    MainRun: TStringList;
    LastRun: TStringList;
    Script: TStringList;
  Private
    ExamName: String;
  Public
    Procedure Parse;
    Constructor Create;
    Procedure Merge(mergingCode: TCPWMCode; currentButtonName: String);
    Destructor Destroy; Override;
  End;

Function LinesAsStrings(aString: String; Var aReturnStringList: TStringList): TStringList;
Function ParseObjectCaption(aObjectStringList: TStringList): String;
Function ParseObjectName(aString: String): String;
Function ParseObjectType(aString: String): String;
Function RemoveChr(aChr: char; aString: String): String;
Function StringAsWords(aString: String; Var aStringList: TStringList): TStringList;
Function TrimComment(scriptLine: String): String;
Function TrimCRLF(aDirtyString: String): String;
Function ValidateName(proposedName: String): String;
Function WordCat(aString: String; delimChar: char): String;

Var
  currentTabbedNotebook, firstTabbedNotebook: TTabbedNotebook;
Const
  CR       : Char   = Chr(13);
  FIRST_RUN: String = 'FirstRun=1';
  LAST_RUN : String = 'LastRun=1';
  LF       : Char   = Chr(10);
  NULL     : Char   = Char(0);
Implementation

{$IFDEF STANDALONE}
Uses
  ExamBase,
  UnitMemoText,
  UnitPrintPopupForm;
{$ENDIF}

Procedure TCPWMForm.SendToBack;
Var
  aCPWMObject: TCPWMObject;
Begin
  aCPWMObject := Children[0]; // the first object in a 'Form' would be the parent
  aCPWMObject.screenObject.SendToBack;
End;

Procedure TCPWMForm.Reposition(baseControl: TWinControl; buttonBarBottom: integer);
Var
  aCPWMObject: TCPWMObject;
  aPopup: TPanel;
Begin
  aCPWMObject := Children[0]; // the first object in a 'Form' would be the parent
  aPopup := aCPWMObject.screenObject As TPanel;
  If (aPopup.Height + aPopup.Top + 21) <= baseControl.Height Then
  Begin
    aPopup.Top := buttonBarBottom + 2;
    aPopup.Left := aPopup.Left + 2;
  End;
End;

Procedure TCPWMForm.Render(baseForm: TComponent; UserObjectList: TList; ScreenObjectList: TList; OnClick: TNotifyEvent; OnChange: TNotifyEvent; baseControl: TWinControl; buttonBarBottom: integer;
  OnHint: TMouseMoveEvent; PopupEditMenu: TPopupMenu; PopupButtonMenu: TPopupMenu);
Var
  aCPWMObject: TCPWMObject;
  i: integer;
Begin

  For i := 0 To Children.Count - 1 Do
  Begin
    aCPWMObject := Children[i];
    aCPWMObject.Render(baseForm, UserObjectList, ScreenObjectList, OnClick, OnChange, baseControl, baseControl, OnHint, PopupEditMenu, PopupButtonMenu);
  End;
  BeenRendered := True;

  If FormType = Popup Then
  Begin // Initialize all popups to send to back
    SendToBack;
  End;

End;

Procedure TCPWMForm.RenderToPrinter(baseForm: TComponent; UserObjectList: TList; ScreenObjectList: TList; OnClick: TNotifyEvent; OnChange: TNotifyEvent; baseControl: TWinControl; buttonBarBottom:
  integer; OnHint: TMouseMoveEvent; PopupEditMenu: TPopupMenu; PopupButtonMenu: TPopupMenu);
Var
  aCPWMObject: TCPWMObject;
  i: integer;
Begin
  For i := 0 To Children.Count - 1 Do
  Begin
    aCPWMObject := Children[i];
    aCPWMObject.RenderToPrinter(baseForm, UserObjectList, ScreenObjectList, OnClick, OnChange, baseControl, baseControl, OnHint, PopupEditMenu, PopupButtonMenu);
  End;
  BeenRendered := True;

  If FormType = Popup Then // Initialize all popups to send to back
    SendToBack;

End;

Function TrimComment(scriptLine: String): String;
Var
  tempStr: String;
  cmtPos: integer;
Begin
  cmtPos := Pos('//', scriptLine);
  If cmtPos > 0 Then
    tempStr := Copy(scriptLine, 1, cmtPos - 1)
  Else
    tempStr := scriptLine;
  Result := tempStr;
End;

Function TCPWMObject.StringToComponent(Value: String; ParentControl: TControl): TComponent;
Var
  StrStream: TStringStream;
  BinStream: TMemoryStream;
Begin
  StrStream := TStringStream.Create(Value);
  Try
    BinStream := TMemoryStream.Create;
    Try
      ObjectTextToBinary(StrStream, BinStream);
      BinStream.Seek(0, soFromBeginning);
      Result := BinStream.ReadComponent(ParentControl);
    Finally
      BinStream.Free;
    End;
  Finally
    StrStream.Free;
  End;
End;

Procedure TCPWMObject.Render(baseForm: TComponent; UserObjectList: TList; ScreenObjectList: TList; OnClick: TNotifyEvent; OnChange: TNotifyEvent; aOwner: TWinControl; aParent: TWinControl; OnHint:
  TMouseMoveEvent; PopupEditMenu: TPopupMenu; PopupButtonMenu: TPopupMenu);
Var
  Msg: String;
  genericControl: TControl;
  deserializedComponet: TComponent;
  aCPWMChild: TCPWMObject;
  i: integer;
Begin
  genericControl := Nil;

  Try
    If ControlType = 'TMemo' Then
    Begin
      genericControl := TMemo.Create(aOwner);
      (genericControl As TMemo).OnChange := OnChange;
      (genericControl As TMemo).OnMouseMove := OnHint;
      (genericControl As TMemo).OnClick := OnClick;
      (genericControl As TMemo).PopUpMenu := PopupEditMenu;
    End
    Else
      If ControlType = 'TLabel' Then
      Begin
        genericControl := TLabel.Create(aOwner);
      End
      Else
        If ControlType = 'TTabbedNotebook' Then
        Begin
          genericControl := TTabbedNotebook.Create(aOwner);
        End
        Else
          If ControlType = 'TPageControl' Then
          Begin
            genericControl := TPageControl.Create(aOwner);
            (genericControl As TPageControl).OnChange := OnChange;
          End
          Else
            If ControlType = 'TTabPage' Then
            Begin
              genericControl := TTabPage.Create(aOwner);
            End
            Else
              If ControlType = 'TTabSheet' Then
              Begin
                genericControl := TTabSheet.Create(aOwner);
              End
              Else
                If ControlType = 'TScrollBox' Then
                Begin
                  genericControl := TScrollBox.Create(aOwner);
                End
                Else
                  If ControlType = 'TComboBox' Then
                  Begin
                    genericControl := TComboBox.Create(aOwner);
                    (genericControl As TComboBox).OnChange := Nil;
                    (genericControl As TComboBox).OnChange := OnChange;
                  End
                  Else
                    If ControlType = 'TPanel' Then
                    Begin
                      genericControl := TPanel.Create(aOwner);
                    End
                    Else
                      If ControlType = 'TListBox' Then
                      Begin
                        genericControl := TListBox.Create(aOwner);
                        (genericControl As TListBox).OnClick := OnClick;
                      End
                      Else
                        If ControlType = 'TORListBox' Then
                        Begin
                          genericControl := TORListBox.Create(aOwner);
                          (genericControl As TORListBox).OnClick := OnClick;
                          (genericControl As TORListBox).OnChange := OnChange;
                        End
                        Else
                          If ControlType = 'TEdit' Then
                          Begin
                            genericControl := TEdit.Create(aOwner);
                            (genericControl As TEdit).OnChange := OnChange;
                            (genericControl As TEdit).PopUpMenu := PopupEditMenu;
                          End
                          Else
                            If ControlType = 'TRadioGroup' Then
                            Begin
                              genericControl := TRadioGroup.Create(aOwner);
                              (genericControl As TRadioGroup).OnClick := OnClick;
                            End
                            Else
                              If ControlType = 'TCPWMRadioGroup' Then
                              Begin
                                genericControl := TCPWMRadioGroup.Create(aOwner);
                                (genericControl As TCPWMRadioGroup).OnClick := OnClick;
                              End
                              Else
                                If ControlType = 'TImage' Then
                                Begin
                                  genericControl := TImage.Create(aOwner);
                                  (genericControl As TImage).OnClick := OnClick;
                                End
                                Else
                                  If ControlType = 'TButton' Then
                                    If (Pos('SUBMIT', Uppercase(Trim(LowerCase(ParseObjectProperty('Caption'))))) > 0) Then
                                    Begin // Use a Color button for 'Submit' buttons
                                      ControlType := 'TColorCAPRIBtn';
                                      self.ObjectText := StringReplace(self.ObjectText, 'TButton', 'TColorCAPRIBtn', [rfIgnoreCase]);
                                      Insert('  ButtonColor = $0080FFFF ' + #$D#$A, self.ObjectText, Pos(' end', self.ObjectText));
                                      genericControl := TColorCAPRIBtn.Create(aOwner);
                                      (genericControl As TColorCAPRIBtn).OnClick := OnClick;
                                    End
                                    Else
                                    Begin // else do the normal TButton
                                      genericControl := TColorCAPRIBtn.Create(aOwner);
                                      (genericControl As TColorCAPRIBtn).OnClick := OnClick; //baseForm.xButtonGenericClick;
                                      If (genericControl As TColorCAPRIBtn).buttoncolor <> $0080FFFF Then
                                        (genericControl As TColorCAPRIBtn).PopUpMenu := PopupButtonMenu;
                                    End
                                  Else
                                    If ControlType = 'TGroupBox' Then
                                    Begin
                                      genericControl := TGroupBox.Create(aOwner);
                                    End
                                    Else
                                      If ControlType = 'TShape' Then
                                      Begin
                                        genericControl := TShape.Create(aOwner);
                                      End
                                      Else
                                        If ControlType = 'THyperlinkLabel' Then
                                        Begin
                                          genericControl := THyperLinkLabel.Create(aOwner);
                                        End
                                        Else
                                          If ControlType = 'TCheckBox' Then
                                          Begin
                                            genericControl := TCheckBox.Create(aOwner);
                                            (genericControl As TCheckBox).OnClick := OnClick;
                                          End;

    If genericControl = Nil Then
      Raise Exception.Create('Could not create a ' + ControlType + ' control: ' + Name);

    genericControl.Parent := aParent;

    deserializedComponet := StringToComponent(ObjectText, genericControl);

    If deserializedComponet = Nil Then
      Raise Exception.Create('Could not deserialize a ' + ControlType + ' control: ' + Name);

    screenObject := genericControl;

    BeenRendered := True;

    SetParentage(aParent); // the current CPWMObject

    If IsUserObject Then
      UserObjectList.Add(self); // list of just the user objects

    ScreenObjectList.Add(self); // everything list

    // Now handle any children this object may have
    If HasChildren Then
      For i := 0 To Children.Count - 1 Do
      Begin
        aCPWMChild := Children[i];
        aCPWMChild.Render(baseForm, UserObjectList, ScreenObjectList, OnClick, OnChange, aOwner, genericControl As TWinControl, OnHint, PopupEditMenu, PopupButtonMenu);
      End;

    // @@@@@@@@@ Special handling @@@@@@@@@@@
    // Special handling if a scroll box
    If (ScreenObject Is TScrollBox) Then
    Begin
      // Force to position 0,0
      (ScreenObject As TScrollBox).HorzScrollBar.Position := 0; // force position
      (ScreenObject As TScrollBox).VertScrollBar.Position := 0;
      // force autoscroll if this is a ScrollBox so children expand
      If ((ScreenObject As TScrollBox).AutoScroll = false) Then
        (ScreenObject As TScrollBox).AutoScroll := true;
    End;

    // if this is a TabbedNotebook (the main interface) then try to set the
    // first tab selected to make the interface look right.
    If ScreenObject = firstTabbedNotebook Then
    Try
      (ScreenObject As TTabbedNotebook).PageIndex := 0; // try to set the tabbednotebook to the first page
    Except
    End;
    // @@@@@@@@@ End of Special handling @@@@@@@@@@@

  Except
    On E: Exception Do
    Begin
      Msg := 'CPWMObject.Render failed. Message: ' + CR + LF;
      Msg := Msg + E.Message;
      Application.MessageBox(PChar(Msg), 'Error Rendering controls', MB_OK);
      Raise;
    End;
  End;

End;

Procedure TCPWMObject.RenderToPrinter(baseForm: TComponent; UserObjectList: TList; ScreenObjectList: TList; OnClick: TNotifyEvent; OnChange: TNotifyEvent; aOwner: TWinControl; aParent: TWinControl;
  OnHint: TMouseMoveEvent; PopupEditMenu: TPopupMenu; PopupButtonMenu: TPopupMenu);
Var
  Msg: String;
  genericControl: TControl;
  deserializedComponet: TComponent;
  aCPWMChild: TCPWMObject;
  i: integer;
  hintCaption: TLabel;
Begin

  genericControl := Nil;

  Try
    If ControlType = 'TMemo' Then
    Begin
      genericControl := TMemo.Create(aOwner);
      (genericControl As TMemo).OnChange := OnChange;
      (genericControl As TMemo).OnMouseMove := OnHint;
      (genericControl As TMemo).OnClick := OnClick;
      (genericControl As TMemo).PopUpMenu := PopupEditMenu;
    End
    Else
      If ControlType = 'TLabel' Then
      Begin
        genericControl := TLabel.Create(aOwner);
      End
      Else
        If ControlType = 'TTabbedNotebook' Then
        Begin
          genericControl := TTabbedNotebook.Create(aOwner);
        End
        Else
          If ControlType = 'TPageControl' Then
          Begin
            genericControl := TPageControl.Create(aOwner);
            (genericControl As TPageControl).OnChange := OnChange;
          End
          Else
            If ControlType = 'TTabPage' Then
            Begin
              genericControl := TTabPage.Create(aOwner);
            End
            Else
              If ControlType = 'TTabSheet' Then
              Begin
                genericControl := TTabSheet.Create(aOwner);
              End
              Else
                If ControlType = 'TScrollBox' Then
                Begin
                  genericControl := TScrollBox.Create(aOwner);
                End
                Else
                  If ControlType = 'TComboBox' Then
                  Begin
                    genericControl := TComboBox.Create(aOwner);
                    (genericControl As TComboBox).OnChange := Nil;
                    (genericControl As TComboBox).OnChange := OnChange;
                  End
                  Else
                    If ControlType = 'TPanel' Then
                    Begin
                      genericControl := TPanel.Create(aOwner);
                    End
                    Else
                      If ControlType = 'TListBox' Then
                      Begin
                        genericControl := TListBox.Create(aOwner);
                        (genericControl As TListBox).OnClick := OnClick;
                      End
                      Else
                        If ControlType = 'TORListBox' Then
                        Begin
                          genericControl := TORListBox.Create(aOwner);
                          (genericControl As TORListBox).OnClick := OnClick;
                          (genericControl As TORListBox).OnChange := OnChange;
                        End
                        Else
                          If ControlType = 'TEdit' Then
                          Begin
                            genericControl := TEdit.Create(aOwner);
                            (genericControl As TEdit).OnChange := OnChange;
                            (genericControl As TEdit).PopUpMenu := PopupEditMenu;
                          End
                          Else
                            If ControlType = 'TRadioGroup' Then
                            Begin
                              genericControl := TRadioGroup.Create(aOwner);
                              (genericControl As TRadioGroup).OnClick := OnClick;
                            End
                            Else
                              If ControlType = 'TCPWMRadioGroup' Then
                              Begin
                                genericControl := TCPWMRadioGroup.Create(aOwner);
                                (genericControl As TCPWMRadioGroup).OnClick := OnClick;
                              End
                              Else
                                If ControlType = 'TImage' Then
                                Begin
                                  genericControl := TImage.Create(aOwner);
                                  (genericControl As TImage).OnClick := OnClick;
                                End
                                Else
                                  If ControlType = 'TButton' Then
                                  Begin
                                    genericControl := TButton.Create(aOwner);
                                    (genericControl As TButton).OnClick := OnClick; //baseForm.xButtonGenericClick;
                                  End
                                  Else
                                    If ControlType = 'TGroupBox' Then
                                    Begin
                                      genericControl := TGroupBox.Create(aOwner);
                                    End
                                    Else
                                      If ControlType = 'TShape' Then
                                      Begin
                                        genericControl := TShape.Create(aOwner);
                                      End
                                      Else
                                        If ControlType = 'TCheckBox' Then
                                        Begin
                                          genericControl := TCheckBox.Create(aOwner);
                                          (genericControl As TCheckBox).OnClick := OnClick;
                                        End;

    If genericControl = Nil Then
      Raise Exception.Create('Could not create a ' + ControlType + ' control: ' + Name);

    genericControl.Parent := aParent;

    deserializedComponet := StringToComponent(ObjectText, genericControl);

    If deserializedComponet = Nil Then
      Raise Exception.Create('Could not deserialize a ' + ControlType + ' control: ' + Name);

    screenObject := genericControl;

    BeenRendered := True;

    // Now handle any children this object may have
    If HasChildren Then
      For i := 0 To Children.Count - 1 Do
      Begin
        aCPWMChild := Children[i];
        aCPWMChild.RenderToPrinter(baseForm, UserObjectList, ScreenObjectList, OnClick, OnChange, aOwner, genericControl As TWinControl, OnHint, PopupEditMenu, PopupButtonmenu);
      End;

    // @@@@@@@@@ Special handling @@@@@@@@@@@
    // Special handling if a scroll box
    If (ScreenObject Is TScrollBox) Then
    Begin
      // Force to position 0,0
      (ScreenObject As TScrollBox).HorzScrollBar.Position := 0; // force position
      (ScreenObject As TScrollBox).VertScrollBar.Position := 0;
      // force autoscroll if this is a ScrollBox so children expand
      If ((ScreenObject As TScrollBox).AutoScroll = false) Then
        (ScreenObject As TScrollBox).AutoScroll := true;
    End;

    // if this is a TabbedNotebook (the main interface) then try to set the
    // first tab selected to make the interface look right.
    If ScreenObject = firstTabbedNotebook Then
    Try
      (ScreenObject As TTabbedNotebook).PageIndex := 0; // try to set the tabbednotebook to the first page
    Except
    End;
    // @@@@@@@@@ End of Special handling @@@@@@@@@@@

    If IsUserObject Then
    Begin
      Try
        hintCaption := TLabel.Create(baseForm);
        hintCaption.Caption := (ScreenObject As TControl).Hint;
        hintCaption.Font.Color := clRed;

        hintCaption.Parent := (ScreenObject As TWinControl); //aParent;
        If (ScreenObject Is TMemo) Then
          hintCaption.Left := (ScreenObject As TWinControl).Width - hintCaption.Canvas.TextWidth(hintCaption.Caption) - 22
        Else
          hintCaption.Left := (ScreenObject As TWinControl).Width - hintCaption.Canvas.TextWidth(hintCaption.Caption) - 7;
        hintCaption.BringToFront;

      Except
      End;
    End;

  Except
    On E: Exception Do
    Begin
      Msg := 'CPWMObject.RenderToPrinter could not render ' + CR + LF;
      Msg := Msg + 'Message: ' + CR + LF + E.Message;
      Application.MessageBox(PChar(Msg), 'Error Rendering controls', MB_OK);
      Raise;
    End;
  End;

End;

Function WordCat(aString: String; delimChar: char): String;
Var
  words: TStringList;
  i: integer;
  outString: String;
Begin
  words := TStringList.Create;
  words := StringAsWords(aString, words);
  For i := 0 To words.Count - 1 Do
  Begin
    outString := outString + words[i];
    If i < words.Count - 1 Then
      outString := outString + delimChar;
  End;
  Result := outString;
  words.free; // SPH 5-4-2006
End;

Procedure TCPWMObject.SetParentage(baseParent: TWinControl); // the current CPWMObject
Var
  parentControl: TWinControl;

Begin

  If Not Assigned(self.Parent) Then
    parentControl := baseParent
  Else
    parentControl := (self.Parent.screenObject As TWinControl);

  If (firstTabbedNotebook = Nil) And (self.screenObject Is TTabbedNotebook) Then
  begin                                                                         //CodeCR458 JRL 5/23/13
    firstTabbedNotebook := self.screenObject As TTabbedNotebook;
    firstTabbedNotebook.Anchors := [akLeft, akRight, akTop, akBottom];          //CodeCR458 JRL 5/23/13
  end;                                                                          //CodeCR458 JRL 5/23/13
  (self.screenObject As TControl).Parent := (parentControl As TWinControl);

End;

Function TCPWMObject.IsHasChildren: boolean;
Begin
  Result := Children.Count > 0;
End;

Function TCPWMObject.GetParentName: String;
Begin
  If Assigned(Parent) Then
    Result := Parent.Name
  Else
    Result := '[NO PARENT]';
End;

Function TCPWMObject.GetName: String;
Var
  caption: String;
Begin
  // Just return the name if we have it
  If Length(Trim(fName)) > 0 Then
  Begin
    Result := fName;
    self.SetName(Result);
    exit;
  End;
  If LowerCase(ParseObjectType(self.Lines.Strings[0])) = 'ttabpage' Then
  Begin
    caption := ParseObjectCaption(self.Lines);
    If (length(trim(caption))) > 0 Then
      Result := caption;
    self.SetName(Result);
    exit;
  End;

  Result := ParseObjectName(self.Lines.Strings[0]);
  self.SetName(Result);

End;

Function TCPWMObject.GetCaption: String;
Begin
  If Length(Trim(fCaption)) <= 0 Then
  Begin
    fCaption := ParseObjectCaption(self.Lines);
  End;
  Result := fCaption
End;

Procedure TCPWMObject.SetName(aName: String);
Begin
  fName := aName;
End;

Constructor TCPWMObject.Create;
Begin
  Lines := TStringList.Create;
  Children := TList.Create;
  fBeenRendered := False;
  fIsTabScrollBox := False;
End;

Constructor TCPWMForm.Create;
Begin
  Children := TList.Create;
  Tabs := TList.Create;
End;

Function RemoveChr(aChr: char; aString: String): String;
Var
  outString: String;
  i: integer;
Begin
  For i := 1 To Length(aString) Do
  Begin
    If Char(aString[i]) <> aChr Then
      outString := outString + aString[i];
  End;
  Result := outString;
End;

Function ParseObjectCaption(aObjectStringList: TStringList): String;
Var
  hintString, str: String;
  i, j: integer;
Begin
  Result := '';
  Try
    hintString := '';
    For i := 0 To aObjectStringList.Count - 1 Do
    Begin
      hintString := trim(LowerCase(aObjectStringList[i])); // note the string is already lowercase for below
      If Pos(LowerCase('Caption'), hintString) > 0 Then
      Begin
        If hintString[Length(hintString)] = '=' Then
        Begin
          j := i;
          hintString := '';
          While (True) Do
          Begin
            inc(j);
            str := RemoveChr('''', trim(aObjectStringList[j]));
            If (str[Length(str)] = '+') Then
              hintString := hintString + Copy(str, 1, Length(str) - 2)
            Else
            Begin
              hintString := hintString + str;
              break;
            End;
          End;
          Result := hintString;
        End
        Else
          Result := RemoveChr('''', Copy(aObjectStringList[i], pos('=', aObjectStringList[i]) + 2, Length(aObjectStringList[i])));
        break;
      End;
    End;
  Except
    On E: Exception Do
      ShowMessageCAPRI('[ParseObjectCaption] ' + E.Message);
  End;
End;

Function ParseObjectName(aString: String): String;
Var
  words: TStringList;
Begin
  words := TStringList.Create;
  Try
    If pos('object', aString) < 0 Then
      Result := ''; // not found
    words := StringAsWords(aString, words);
    If words.Count = 3 Then
      Result := Copy(words[1], 1, Pos(':', words[1]) - 1)
    Else
      Result := '';
  Finally
    FreeAndNil(words);
  End;
End;

Function StringAsWords(aString: String; Var aStringList: TStringList): TStringList;
Var
  i: integer;
  stringToAdd: String;
  tempString: String;
Begin
  tempString := '';
  For i := 1 To Length(aString) Do
  Begin

    If Not (aString[i] In [' ', CR, LF, Char(0)]) Then
    Begin
      tempString := tempString + aString[i];
    End;

    If aString[i] In [' ', CR, LF] Then
    Begin
      stringToAdd := Trim(tempString);
      If Length(stringToAdd) > 0 Then
        aStringList.Add(stringToAdd);
      tempString := '';
    End;

  End; // rof

  stringToAdd := Trim(tempString);
  If Length(stringToAdd) > 0 Then
    aStringList.Add(stringToAdd);

  Result := aStringList;
End;

Function ParseObjectType(aString: String): String;
Var
  words: TStringList;
Begin
  words := TStringList.Create;
  Try
    If pos('object', aString) < 0 Then
      Result := ''; // not found
    words := StringAsWords(aString, words);
    If words.Count = 3 Then
      Result := words[2]
    Else
      If words.Count = 2 Then
        Result := words[1]
      Else
        Result := '';
  Finally
    FreeAndNil(words);
  End;
End;

Function TrimCRLF(aDirtyString: String): String;
Var
  cleanString: String;
  j: integer;
Begin
  For j := 1 To Length(aDirtyString) Do
    If Not (aDirtyString[j] In [CR, LF, NULL]) Then
      cleanString := cleanString + aDirtyString[j];
  Result := cleanString;
End;

Function ValidateName(proposedName: String): String;
Var
  tempString: String;
Begin
  tempString := StringReplace(proposedName, ' ', '_', [rfReplaceAll]);
  tempString := StringReplace(tempString, '/', '_', [rfReplaceAll]);
  Result := tempString;
End;

Function LinesAsStrings(aString: String; Var aReturnStringList: TStringList): TStringList;
Var
  i: integer;
  previousLineStart: integer;
  tempString: String;
Begin

  previousLineStart := 1; // start at beginning of string

  For i := 1 To Length(aString) Do
  Begin
    If aString[i] = LF Then
    Begin
      previousLineStart := i + 1;
      continue; // ignore line feed but go to next character
    End;

    If aString[i] = CR Then
    Begin
      tempString := Copy(aString, previousLineStart, i - previousLineStart);
      previousLineStart := i + 1;
      If Length(TrimCRLF(tempString)) > 0 Then
        aReturnStringList.Add(TrimCRLF(tempString));
    End;
  End; // rof
  tempString := Copy(aString, previousLineStart, Length(aString));
  If Length(TrimCRLF(tempString)) > 0 Then
    aReturnStringList.Add(TrimCRLF(tempString));

  Result := aReturnStringList;
End;

Function TCPWMObject.GetControlHint: String;
Var
  hintString: TStringList;
  i, j: integer;
Begin
  hintString := TStringList.Create;
  Try
    Result := '';
    For i := 0 To Lines.Count - 1 Do
    Begin
      { Note: we need to look for just hint (not ShowHint or ParentShowHint)
        therefore need to be sure word being parsed has a leading space.
        This should be okay since any property text would be indented at
        least two spaces in the .dfm type text. }
      If Pos(' hint', LowerCase(Lines[i])) > 0 Then {// look for the 'hint' property}
      Begin
        hintString := StringAsWords(Lines[i], hintString);
        For j := 0 To hintString.Count - 1 Do
        Begin
          If LowerCase(hintString[j]) = 'hint' Then
            continue;
          If hintString[j] = '=' Then
            continue;
          HintValue := RemoveChr('''', hintString[j]); // anything not either 'hint' or '=' is the value
          Result := HintValue;
        End;
      End;
    End;
  Finally
    FreeAndNil(hintString);
  End;
End;

Procedure TCPWMObject.GetScreenLocation;
Var
  leftStr, TopStr: String;
Begin
  LeftStr := ParseObjectProperty('Left');
  Try
    Left := StrToInt(LeftStr);
  Except
  End;
  TopStr := ParseObjectProperty('Top');
  Try
    Top := StrToInt(TopStr);
  Except
  End;
End;

Function TCPWMObject.ParseObjectProperty(propToParse: String): String;
Var
  propString: String;
  i: integer;
  value: String;
Begin
  Result := '';
  Try
    propString := '';
    For i := 0 To Lines.Count - 1 Do
    Begin
      propString := LowerCase(Lines[i]); // note the string is already lowercase for below
      If Pos(LowerCase(propToParse), propString) > 0 Then
      Begin
        value := RemoveChr('''', Copy(Lines[i], pos('=', Lines[i]) + 2, Length(Lines[i])));
        Result := value;
        break;
      End;
    End;
  Except
    On E: Exception Do
      ShowMessageCAPRI('[ParseObjectProperty] ' + E.Message);
  End;
End;

Procedure TCPWMObject.ExamineObject;
Var
  i: integer;
  child: TCPWMObject;
  tabCaption: String;
  numericHint: integer;
Begin
  Try
    For i := self.Lines.Count - 1 Downto 0 Do
    Begin
      If Pos('onclick', LowerCase(self.Lines[i])) > 0 Then
        self.Lines.Delete(i);
    End;
    ObjectText := Lines.Text; // after being cleaned of onclick
    GetName; // side effect also sets the name
    fDepth := Pos('object ', LowerCase(Lines[0])) Div 2;
    ControlType := ParseObjectType(self.Lines[0]); // object type in first line of object string array

    { since we recurse depth first during object parsing, we need to determine
      if an object is a Tab's scrollbox when we examine the tab, which happens
      AFTER we examined it's children. }
    If (ControlType = 'TTabPage') And HasChildren Then
    Begin
      child := Children[0];
      If child.ControlType = 'TScrollBox' Then
        child.IsTabScrollBox := True;
    End;

    If (ControlType = 'TTabPage') And HasChildren Then
    Begin
      tabCaption := ParseObjectCaption(self.Lines);
      tabCaption := WordCat(tabCaption, '_');
      SetOnTabName('TTabPage_' + tabCaption, Children);
    End;

    GetControlHint; // side effect also sets the hint

    If HintValue = 'POPUP' Then
      PopUp := True
    Else
      PopUp := False;

    If Not Popup Then
    Begin // no need to check hint any further

      //jcs - modified logic, removed try/except
      //numericHint := StrToInt(HintValue);
      numericHint := StrToIntDef(HintValue, 0);
      IsUserObject := numericHint > 0;
    End;

    GetScreenLocation;

    //jcs - modified logic, removed try/except
    //TabOrder := StrToInt(ParseObjectProperty('TabOrder'));
    TabOrder := StrToIntDef(ParseObjectProperty('TabOrder'),0);

  Except
    On E: Exception Do
      ShowMessageCAPRI('Error during ExamineObject ' + e.message);
  End;
End;

Procedure TCPWMObject.SetOnTabName(tabPageName: String; childen: TList);
Var
  i: integer;
  child: TCPWMObject;
Begin
  For i := 0 To childen.Count - 1 Do
  Begin
    child := childen.Items[i];
    child.OnTabName := tabPageName;
    If child.HasChildren Then
      child.SetOnTabName(tabPageName, child.Children); // recurse
  End;
End;

Constructor TCPWMExam.Create(theXmlMapping: TStringList; theXmlConfig: TStringList);
Begin
  Forms := TList.Create;
  Code := TCPWMCode.Create;
  Tabs := TList.Create;
  UserObjectList := TList.Create;
  ScreenObjectList := TList.Create;
  Properties := TStringList.Create;
  tabNames := TStringList.Create;

  if theXmlMapping <> Nil then
    XsdEngine := TXsdEngineLite.Create(theXmlMapping);

  if theXmlConfig <> Nil then
    XmlConfig := TXmlConfig.Create(theXmlConfig);
End;

Function TCPWMExam.NumericString(line: String): boolean;
Var
  i: integer;
  currChar: char;
Begin
  Result := true; // default
  For i := 1 To Length(line) Do
  Begin
    currChar := line[i];
    If ord(currChar) = 39 Then
      continue;
    If Not (// 1 through 0 or A through F     Hex characters
      (ord(currChar) >= 48) And (ord(currChar) <= 57) Or
      (ord(currChar) >= 65) And (ord(currChar) <= 70)
      ) Then
    Begin
      Result := false;
      exit;
    End;
  End;
End;

Procedure TCPWMExam.ParseExamFile(fileLines: TStrings);
Var
  objectText, line: String;
  bulkLines: TStringList;
  lineCount, i: integer;
  aForm: TCPWMForm;
Begin
  Try
    bulkLines := TStringList.Create;

    objectText := '';
    lineCount := 0;

    For i := 0 To fileLines.Count - 1 Do
    Begin
      line := fileLines[i];

      { Special handling for 'picture' data}
      If NumericString(Trim(line)) Then
      Begin
        bulkLines.Add(line);
        lineCount := lineCount + 1;
        continue; // to next line
      End;

      If bulkLines.count > 0 Then
      Begin
        objectText := objectText + #13#10 + bulkLines.Text;
        objectText := objectText + #13#10 + line;
        bulkLines.Clear;
        continue
      End;
      { End of 'picture' data handling }

      If ((Pos('object ', line) = 1) And (lineCount = 0) And (Pos(': T', line) > 0)) Or
        ((Pos('object T', line) = 1)) Then
      Begin
        objectText := line;
        continue; // at least one more line follows
      End;

      If (Pos('end', line) <> 1) Then
      Begin
        objectText := objectText + #13#10 + line;
        lineCount := lineCount + 1;
        continue; // to next line
      End;

      // now at 'end' in first position
      objectText := objectText + #13#10 + line; // should be 'end' at 0th position of line

      Try
        aForm := ParseForm(AdjustLineBreaks(objectText)); // Build object
        Forms.Add(aForm); // Add to Forms List
      Finally
      End;

      // Reset text and count for next object-graph in file
      objectText := '';
      lineCount := 0;

    End; // rof
  Finally
    FreeAndNil(bulkLines);
  End;

  Name := DetermineExamName;
  Code.ExamName := Name;
  ReadProperties;
End;

Function TCPWMExam.GetXmlConfig: TXmlConfig;
begin
  Result := XmlConfig;
end;

Function TCPWMExam.GetXsdEngine: TXsdEngineLite;
begin
  Result := XsdEngine;
end;

Function TCPWMExam.IsXmlBasedExam: Boolean;
begin
  if (XsdEngine <> Nil) and (XmlConfig <> Nil) then
    Result := True
  else
    Result := False;
end;

Function TCPWMExam.GetButtonBarCaption: String;
begin
  Result := ButtonText;
end;

Procedure TCPWMExam.ParseExamCode(script: TStrings);
Begin
  self.Code.Script.Clear;
  self.Code.Script.AddStrings(script);
  self.Code.Parse;
End;

Function TCPWMExam.DetermineExamName: String;
Var
  i: integer;
  aForm: TCPWMForm;
  aObject: TCPWMObject;
Begin

  For i := 0 To Forms.Count - 1 Do
  Begin
    aForm := Forms.items[i];
    If aForm.FormType = FormID Then
    Begin
      aObject := aForm.Children[0];
      Result := aObject.Caption;
      break;
    End;
  End;
End;

Function TCPWMExam.ReadProperties: String;
Var
  i, j, k: integer;
  aForm: TCPWMForm;
  aObject: TCPWMObject;
  child: TCPWMObject;
  propObject: TCPWMObject;
Begin
  For i := 0 To Forms.Count - 1 Do
  Begin
    aForm := Forms.items[i];
    If aForm.FormType = Popup Then
    Begin
      aObject := aForm.Children[0];
      For j := 0 To aObject.Children.Count - 1 Do
      Begin
        child := aObject.Children[j];
        If UpperCase(Trim(child.Caption)) = 'PROPERTIES' Then
        Begin
          Properties.Clear;
          For k := 0 To aObject.Children.Count - 1 Do
          Begin
            propObject := aObject.Children[k];
            If (propObject.ControlType = 'TLabel') And (UpperCase(Trim(propObject.Caption)) <> 'PROPERTIES') Then
            Begin
              Properties.Add(propObject.Caption);
            End;
          End;
          exit; // Only one such Popup to be done on each exam
        End;
      End;
    End;
  End;

End;

Function TCPWMExam.GetProperty(propertyName: String): String;
Var
  i: integer;
  currProp: String;
Begin
  Result := ''; // default to empty result
  With Properties Do
    For i := 0 To Count - 1 Do
    Begin
      currProp := Copy(Properties[i], 0, LastDelimiter('=', Properties[i]) - 1);
      If UpperCase(Trim(currProp)) = UpperCase(Trim(propertyName)) Then
      Begin
        Result := Copy(Properties[i], LastDelimiter('=', Properties[i]) + 1, Length(Properties[i]));
        break;
      End;
    End;
End;

Function TCPWMExam.ParseForm(objectTextToAdd: String): TCPWMForm;
Var
  aForm: TCPWMForm;
  objectStringList: TStringList;
  aCPWMObject: TCPWMObject;
Begin
  aForm := TCPWMForm.Create;

  objectStringList := TStringList.Create;
  objectStringList := LinesAsStrings(objectTextToAdd, objectStringList);

  aForm.Children.Add(ParseObjects(objectStringList, Nil));

  aCPWMObject := aForm.Children[0];

  If aCPWMObject.PopUp Then
    aForm.FormType := Popup;
  If aCPWMObject.Name = 'LabelTypeOfForm' Then
    aForm.FormType := TypeOfForm;
  If aCPWMObject.Name = 'LabelFormID' Then
    aForm.FormType := FormID;
  If aCPWMObject.ControlType = 'TTabbedNotebook' Then
    aForm.FormType := Notebook;

  aForm.Name := 'Form [' + aCPWMObject.Name + ']';
  aForm.Caption := objectTextToAdd;
  aForm.Exam := self;

  objectStringList.Free; // SPH 5-4-2006
  Result := aForm;

End;

Function TCPWMExam.ParseObjects(lines: TStringList; parent: TCPWMObject): TCPWMObject;
Var
  line: String;
  i: integer;
  obDepth: integer;
  aCPWMObject: TCPWMObject;
  childLines: TStringList;
  childDepth: integer;
Begin
  i := 0;
  line := lines.Strings[i];
  obDepth := GetObjectDepth(line);
  aCPWMObject := TCPWMObject.Create;
  aCPWMObject.Lines.Add(line); // the 'object' line
  aCPWMObject.Name := aCPWMObject.GetName; // This will set the name except for TTabPages
  aCPWMObject.Parent := parent; // have to pass in the parent TCPWMObject object

  Inc(i); // goto next line
  Repeat
    line := lines.Strings[i];
    If IsObjectDefinition(line) And (GetObjectDepth(line) > obDepth) Then
    Begin
      childLines := TStringList.Create;
      Try
        childDepth := GetObjectDepth(line);
        Repeat
          line := lines.Strings[i];
          childLines.Add(lines.Strings[i]);
          Inc(i);
        Until (IsEndOfDefinition(line) And (GetEndDepth(line) <= childDepth));
        Dec(i); // back up the counter since the embedded end was
        aCPWMObject.Children.Add(ParseObjects(childLines, aCPWMObject)); // recurse
      Finally
        childLines.Free;
      End;
    End;
    aCPWMObject.Lines.Add(line);
    Inc(i);
  Until (IsEndOfDefinition(line) And (GetEndDepth(line) <= obDepth)); //or (i >= Lines.count - 1);
  aCPWMObject.ExamineObject; // set all the object properties

  aCPWMObject.Exam := self;

  Result := aCPWMObject;
End;

Function TCPWMExam.IsObjectDefinition(aLine: String): boolean;
Begin
  If ((Pos('object ', aLine) > 0) And (Pos(': T', aLine) > 0)) Or
    ((Pos('object T', aLine) > 0)) Then
    Result := True
  Else
    Result := False;
End;

Function TCPWMExam.IsEndOfDefinition(aLine: String): boolean;
Begin
  If (Pos('end', aLine) > 0) Then
    Result := True
  Else
    Result := False;
End;

Function TCPWMExam.GetEndDepth(aLine: String): integer;
Begin
  Result := Pos('end', LowerCase(aLine)) Div 2;
End;

Function TCPWMExam.GetObjectDepth(aLine: String): integer;
Begin
  Result := Pos('object ', LowerCase(aLine)) Div 2
End;

Destructor TCPWMExam.Destroy;
Var
  i: integer;
  currentForm: TCPWMForm;
Begin
  currentForm := Nil;
  { Remove the Forms }
  For i := Forms.Count - 1 Downto 0 Do
  Begin
    Try
      currentForm := Forms[i];
      If Assigned(currentForm) Then
      Begin
        currentForm.free;
        currentForm := Nil;
      End;
    Except
      ShowMessageCAPRI('Could not free form ' + currentForm.Name + '.');
    End;
  End;

  Try
    Forms.Free;
    Tabs.Free;
    tabNames.Free;
    UserObjectList.Free;
    ScreenObjectList.Free;
    Code.Free;
  Except
  End;
  Inherited;
End;

Destructor TCPWMObject.Destroy;
Var
  i: integer;
  currentObject: TCPWMObject;
Begin
  For i := 0 To Children.Count - 1 Do
  Begin
    currentObject := Children[i];
    If Assigned(currentObject) Then
      currentObject.Free;
  End;

  Try
    screenObject := Nil; // SPH ZZZZZZZZZZZ

    Children.Free;
    Lines.Free;
  Except
  End;

  Inherited;
End;

Destructor TCPWMForm.Destroy;
Var
  i: integer;
  currentObject: TCPWMObject;
Begin
  For i := 0 To Children.Count - 1 Do
  Begin
    currentObject := Children[i];
    currentObject.Free;
  End;
  Try
    Children.Free;
    Tabs.Free;
  Except
  End;
  Inherited;
End;

Procedure TCPWMExam.ExamButtonClick(Sender: TObject);
Var
  currentButton: integer;
  currentButtonName: String;
  j, k: integer;
  currentPageIndex: integer;
  currentPage: TTabPage;
  bbScroller: TScrollBox;  //CodeCR300 - rpm 3/6/12
  buttonBar: TToolBar;
  isAllowChange: Boolean;  //CodeCR190 - rpm 4/20/11
  senderButton: TSpeedButton;
Begin
  TControl((aMetaNotebook.parent).findcomponent('xCPWMxPanel_' + inttostr(TComponent(Sender).Tag))).bringToFront;
  if(Sender is TSpeedButton) then
    senderButton := Sender As TSpeedButton;
  currentButtonName := Piece((Sender As TSpeedButton).Hint, '^', 3);
  Try
  ChangeRunningExam(currentButtonName)Except
  End; // From PNCSCodeRun

  For j := tabNames.Count - 1 Downto 0 Do
  Begin
    currentPageIndex := aMetaNotebook.Pages.IndexOf(tabNames[j]);
    currentPage := TTabPage(aMetaNotebook.Pages.Objects[currentPageIndex]);
    bbScroller := TScrollBox(currentPage.Controls[0]);  //CodeCR300 - rpm 3/6/12
    buttonBar := TToolBar(bbScroller.Controls[0]);
    currentButton := 0;
    For k := 0 To buttonBar.ControlCount - 1 Do
    Begin
      If piece(buttonBar.controls[k].Name, '_', 2) = currentButtonName Then
        currentButton := k;
    End;
    Try
    (buttonBar.controls[currentbutton] As tSpeedButton).down := true Except
      Showmessage('Change button ' + inttostr(currentbutton) + ' failed');
    End;
    {CodeCR190 - rpm 4/19/11 - Execute the ExamNoteBookTabChange handler
     whenever the Exam button is clicked to update the 'CLICKED' status
     for the current active tab.}
    isAllowChange := True;
    ExamNoteBookTabChange(aMetaNotebook, aMetaNotebook.GetIndexForPage(aMetaNotebook.ActivePage), isAllowChange);
    PNCSEvent := True;
  End;
End;

Procedure TCPWMExam.ExamNotebookTabChange(Sender: TObject; NewTab: Integer; Var AllowChange: Boolean);
Var
  k: integer;
  currentPage: TTabPage;
  bbScroller: TScrollBox;  //CodeCR300 - rpm 3/6/12
  buttonBar: TToolBar;
  button: TSpeedButton;
  contentFoundFlag: boolean;
Begin
  contentFoundFlag := False;
  currentPage := TTabPage(aMetaNotebook.Pages.Objects[NewTab]);
  bbScroller := TScrollBox(currentPage.Controls[0]);  //CodeCR300 - rpm 3/6/12
  buttonBar := TToolBar(bbScroller.Controls[0]);
  For k := 0 To buttonBar.ControlCount - 1 Do
  Begin
    button := buttonBar.Controls[k] As TSpeedButton;
    If ((button.Down) And (button.Enabled)) Then
    Begin
      TControl((aMetaNotebook.parent).findcomponent('xCPWMxPanel_' + inttostr(button.Tag))).bringToFront;
      contentFoundFlag := True;
      If piece(button.Hint, '^', 4) <> 'CLICKED' Then
        button.Hint := button.Hint + '^CLICKED';

      // Inform script that the tab has changed
      PNCSEvent := True;
      FormIDChanged(Piece(button.Hint, '^', 3));
      ChangeRunningExam(Piece(button.Hint, '^', 3));

      break;
    End
  End;

  If Not contentFoundFlag Then
    For k := 0 To buttonBar.ControlCount - 1 Do
    Begin
      button := buttonBar.Controls[k] As TSpeedButton;
      If (button.Enabled) Then
      Begin
        TControl((aMetaNotebook.parent).findcomponent('xCPWMxPanel_' + inttostr(button.Tag))).bringToFront;
        button.Down := True;
        If piece(button.Hint, '^', 4) <> 'CLICKED' Then
          button.Hint := button.Hint + '^CLICKED';

        PNCSEvent := True;
        FormIDChanged(Piece(button.Hint, '^', 3));
        ChangeRunningExam(Piece(button.Hint, '^', 3));

        break;
      End
    End;

End;

Procedure TCPWMExam.PostProcessScript(ExamList: TList);
Var
  j: integer;
  aExam: TCPWMExam;

Begin
  For j := 0 To examlist.Count - 1 Do
  Begin
    aExam := examlist[j];
    self.Code.MainRun.Insert(0, 'EndIf');
    self.Code.MainRun.Insert(0, 'Goto ' + aExam.Name + 'AUTOLABELBEGINNINGOFEXAMSCRIPT');
    self.Code.MainRun.Insert(0, 'If xCurrentExamNumber="' + aExam.name + '"');
  End;

  self.Code.LastRun.Insert(0, 'Label ENDOFMERGEDTEMPLATESSECTION');

  self.Code.Script.Clear;
  self.Code.Script.AddStrings(self.Code.FirstRun);
  self.Code.Script.AddStrings(self.Code.MainRun);
  self.Code.Script.AddStrings(self.Code.LastRun);

End;

Procedure TCPWMExam.Render(AOwner: TComponent; OnClick: TNotifyEvent; OnChange: TNotifyEvent; baseControl: TWinControl; OnHint: TMouseMoveEvent; ExamList: TList; PopupEditMenu: TPopupMenu;
  PopupButtonMenu: TPopupMenu);
Var
  Msg: String;
  i, j, k: integer;
  tmpWdth, tmpHght : Integer; //CodeCR458 JRL 5/23/13
  aForm: TCPWMForm;
  currentPageIndex: integer;
  currentPage: TTabPage;
  bbScroller: TScrollBox;  //CodeCR300 - rpm 3/6/12
  buttonBar: TToolBar;
  button: TSpeedButton;
  aTabObject: TCPWMObject;
  child: TCPWMObject;
  contentPanel: TPanel;
  aExam: TCPWMExam;
  examIndex: integer;
  tempBoolean: boolean;
  buttonBarBottom: integer;
Begin
  tmpWdth := 780; //Minimum default allowable form size                         //CodeCR458 JRL 5/23/13
  tmpHght := 580;                                                               //CodeCR458 JRL 5/23/13
  buttonBar := Nil;
  child := Nil;
  currentPage := Nil;
  UserObjectList.Clear; // Clear any residual from the list of all user objects
  aMetaNotebook := TTabbedNotebook.Create(AOwner); // Build the metaTab notebook
  aMetaNotebook.Name := 'xMetaNotebook';
  aMetaNotebook.Parent := baseControl;
  aMetaNotebook.Pages.Clear;

  Try
    For j := 0 To tabs.count - 1 Do
    Begin
      aTabObject := tabs[j];
      If examList.IndexOf(aTabObject.Exam) < 0 Then
        examList.Insert(0, aTabObject.Exam);
    End;

    For i := 0 To tabNames.Count - 1 Do
    Begin
      currentPageIndex := aMetaNotebook.Pages.Add(tabNames[i]);
      currentPage := TTabPage(aMetaNotebook.Pages.Objects[currentPageIndex]);
      currentPage.Parent := aMetaNotebook;

      bbScroller := TScrollBox.Create(AOwner);  //CodeCR300 - rpm 3/6/12
      bbScroller.Name := 'xBBScroller' + IntToStr(i);
      bbScroller.Parent := currentPage; // the tabpage
      bbScroller.Align := alTop;
      bbScroller.Color := clSilver;

      buttonBar := TToolBar.Create(AOwner);
      buttonBar.Name := 'xButtonBar' + IntToStr(i);
      buttonBar.Align := alNone;
      buttonBar.AutoSize := false;
      buttonBar.Height := 21;
      buttonBar.EdgeBorders := [];
      buttonBar.Parent := bbScroller;  //CodeCR300 - rpm 3/6/12
      bbScroller.Height := (buttonBar.Height * 2);  //CodeCR300 - rpm 3/6/12

      For j := 0 To examList.count - 1 Do
      Begin
        aExam := examList[j];
        button := TSpeedButton.Create(AOwner);
        button.Name := ValidateName('xButton_' + aExam.Name + '_' + tabNames[i] + IntToStr(j));

        aExam.ButtonText := aExam.GetProperty('ButtonText'); // Get property if any
        aExam.FHeight := StrToIntDef(aExam.GetProperty('FormHeight'), 0); // Get property if any //CodeCR458 JRL 5/23/13
        aExam.FWidth := StrToIntDef(aExam.GetProperty('FormWidth'), 0); // Get property if any   // CodeCR458 JRL 5/23/13
        { Determine Largest Dimensions for form from each assigned template }   //CodeCR458 JRL 5/23/13
        if (aExam.FWidth > tmpWdth) then                                        //CodeCR458 JRL 5/23/13
          tmpWdth := aExam.FWidth;                                              //CodeCR458 JRL 5/23/13
        if (aExam.FHeight > tmpHght) then                                       //CodeCR458 JRL 5/23/13
          tmpHght := aExam.FHeight;                                             //CodeCR458 JRL 5/23/13
        If Length(Trim(aExam.ButtonText)) > 0 Then
          button.Caption := aExam.ButtonText
        Else
          button.Caption := aExam.Name; // Default back to the Exam Name if no property found

        button.Width := buttonBar.Canvas.TextWidth(button.Caption) + 15; // Need sizing for caption text
        button.Parent := buttonBar;
        button.GroupIndex := 1; // So one will stay down, and all others pop up
        buttonBar.ButtonHeight := 17;
        buttonBar.Width := buttonBar.Width + button.Width;
        button.Tag := j; // We will need this index to hook up to the button when the content panel is created below
        button.Hint := IntToStr(j) + '^' + tabNames[i] + ' - ' + aExam.Name + '^' + aExam.Name;
        button.OnClick := ExamButtonClick;
        button.Enabled := false; // default disabled till below (if we have content for this exam)
      End;

    End;

    { Set Default Form Width/Height in 1st Exam }                               //CodeCR458 JRL 5/23/13 
    TCPWMExam(examList[0]).FWidth := tmpWdth;                                   //CodeCR458 JRL 5/23/13
    TCPWMExam(examList[0]).FHeight := tmpHght;                                  //CodeCR458 JRL 5/23/13
    aMetaNotebook.PageIndex := 0; // force to first tab
    aMetaNotebook.Parent := baseControl;
    aMetaNotebook.Align := alClient; // was alNone
    aMetaNotebook.Left := 2;
    aMetaNotebook.Top := 2;
    aMetaNotebook.Height := aMetaNotebook.Parent.Height - 4;
    aMetaNotebook.Width := aMetaNotebook.Parent.Width - 4;
    aMetaNotebook.Anchors := [akLeft, akTop, akRight, akBottom]; // added akRight  //CodeCR458 JRL 5/23/13

    buttonBarBottom := currentPage.Top + buttonBar.Height;
    For i := 0 To Forms.Count - 1 Do {// for each form on the exam}
    Begin
      aForm := Forms.items[i];
      If aForm.FormType = Popup Then
      Begin
        aForm.Render(AOwner, UserObjectList, ScreenObjectList, OnClick, OnChange, baseControl, buttonBarBottom, OnHint, PopupEditMenu, PopupButtonMenu);
        aForm.SetTabOrder;
      End;
    End;

    { Build and populate the 'Meta' tabs }
    For i := 0 To tabs.count - 1 Do
    Begin
      aTabObject := tabs[i];
      For j := 0 To tabNames.Count - 1 Do
      Begin
        currentPageIndex := aMetaNotebook.Pages.IndexOf(tabNames[j]);
        currentPage := TTabPage(aMetaNotebook.Pages.Objects[currentPageIndex]);

        If aTabObject.Name = tabNames[j] Then
        Begin
          bbScroller := TScrollBox(currentPage.Controls[0]);  //CodeCR300 - rpm 3/6/12
          buttonBar := TToolBar(bbScroller.Controls[0]);

          contentPanel := TPanel.Create(AOwner);
          contentPanel.Parent := currentPage; // Set the Content container's Parent to our 'Meta'Tab
          contentPanel.Align := alClient;
          contentPanel.Name := 'xCPWMxPanel_' + Inttostr(LongInt(contentpanel));
          contentPanel.BevelInner := bvNone;
          contentPanel.BevelOuter := bvNone;
          contentPanel.Caption := '';

          {  Pass in the button's tag so we can display when button is clicked }
          examIndex := examList.IndexOf(aTabObject.Exam); // Find the index of the Exam in the ExamList
          button := buttonBar.Controls[examIndex] As TSpeedButton;
          button.Tag := LongInt(contentPanel);
          button.Enabled := True;

          If aTabObject.HasChildren Then
          Begin // Must have at least the ScrollBox
            Try
              child := aTabObject.Children[0]; // get the ScrollBox
              Assert(child.screenObject Is TScrollBox, child.Name + ' is not a scrollbox object.');

              child.Parent.screenObject := contentPanel; // and reset the old parent to the new content pane
              child.Render(AOwner, UserObjectList, ScreenObjectList, OnClick, OnChange, baseControl, contentPanel, OnHint, PopupEditMenu, PopupButtonMenu); // Now render the scrollbox on the content
              child.SetTabOrder;
            Except
              On E: Exception Do
              Begin
                Msg := 'CPWMExam.Render could not render ' + CR + LF;
                Msg := Msg + child.Name + '.' + CR + LF + 'Message: ' + CR + LF + E.Message;
                Application.MessageBox(PChar(Msg), 'Error Rendering controls', MB_OK);
                Raise;
              End;
            End;
          End;
          currentPage.Visible := True;
          currentPage.Enabled := True;
        End;
      End;
    End;

    SetTabOrder; //Ensure TabOrder is properly set over all

    { Finally, insure the first enabled button on each tab is depressed by default}
    For j := tabNames.Count - 1 Downto 0 Do
    Begin
      currentPageIndex := aMetaNotebook.Pages.IndexOf(tabNames[j]);
      currentPage := TTabPage(aMetaNotebook.Pages.Objects[currentPageIndex]);
      bbScroller := TScrollBox(currentPage.Controls[0]);  //CodeCR300 - rpm 3/6/12
      buttonBar := TToolBar(bbScroller.Controls[0]);
      For k := 0 To buttonBar.ControlCount - 1 Do
      Begin
        button := buttonBar.Controls[k] As TSpeedButton;
        If button.Enabled Then
        Begin
          button.Down := true;
          TControl((aMetaNotebook.parent).findcomponent('xCPWMxPanel_' + inttostr(button.Tag))).bringToFront;
          If j = 0 Then
            ChangeRunningExam(Piece(button.Hint, '^', 3)); // From PNCSCodeRun
          continue; // on to the next 'Meta' tab
        End;
      End;

    End;

    aMetaNotebook.OnChange := ExamNotebookTabChange;
    tempBoolean := true;
    ExamNotebookTabChange(aMetaNotebook, 0, tempBoolean);

  Finally
  End;
End;

Procedure TCPWMExam.ScriptMerge(aExam: TCPWMExam; currentButtonName: String);
Begin
  Code.Merge(aExam.Code, currentButtonName);
End;

Procedure TCPWMExam.Merge(aExam: TCPWMExam);
Var
  i: integer;
  mergingTab: TCPWMObject;
  baseTab: TCPWMObject;
  aForm: TCPWMForm;
  baseNotebook: TCPWMForm;
  mergeNotebook: TCPWMForm;
  currentButtonname: String;
Begin
  mergeNotebook := Nil;
  { Add all popups from incoming to 'Base' }
  If aExam <> Nil Then
  Begin // Except when no merginging exam
    For i := 0 To aExam.Forms.Count - 1 Do
    Begin // incoming Exam
      aForm := aExam.Forms[i];
      If aForm.FormType = Popup Then
        self.Forms.Add(aForm);
    End;

    currentButtonname := aExam.Name;

    { Merge Script }
    ScriptMerge(aExam, currentButtonName);
  End
  Else
  Begin
    // Will add a double END command without the next line
  End;

  { Get the base notebook }
  baseNotebook := self.GetNotebook;
  If Not Assigned(baseNotebook) Then
    Raise Exception.Create('Merge: Could not identify the base notebook');

  { Get the merge notebook }
  If aExam <> Nil Then
  Begin // Except when no merginging exam
    mergeNotebook := aExam.GetNotebook;
    If Not Assigned(mergeNotebook) Then
      Raise Exception.Create('Merge: Could not identify the merge notebook');
  End;

  baseNotebook.Tabs.Clear;
  baseNotebook.SetTabs(baseNotebook.Children); // list of tabs on base

  If aExam <> Nil Then
  Begin // Except when no merginging exam
    mergeNotebook.Tabs.Clear;
    mergeNotebook.SetTabs(mergeNotebook.Children); // list of tabs on merge
  End;

  If tabs.Count <= 0 Then // only get the base tabs once
    For i := 0 To baseNotebook.Tabs.Count - 1 Do
      tabs.Add(baseNotebook.Tabs[i]);

  If aExam <> Nil Then
  Begin // Except when no merginging exam
    For i := 0 To mergeNotebook.Tabs.Count - 1 Do
      tabs.Add(mergeNotebook.Tabs[i]);
  End;

  For i := 0 To baseNotebook.Tabs.Count - 1 Do
  Begin
    baseTab := baseNotebook.Tabs[i];
    If tabNames.IndexOf(baseTab.Name) = -1 Then
      tabNames.Add(baseTab.Name);
  End;

  If aExam <> Nil Then
  Begin // Except when no merginging exam
    For i := 0 To mergeNotebook.Tabs.Count - 1 Do
    Begin
      mergingTab := mergeNotebook.Tabs[i];
      If tabNames.IndexOf(mergingTab.Name) = -1 Then
        tabNames.Add(mergingTab.Name);
    End;
  End;

  OrderTabNames;

End;

Procedure TCPWMExam.OrderTabNames;
Var
  i, index, breakOutCount: integer;
  ValidNames: TStringList;
  currName: String;
  retry: boolean;
Begin
  ValidNames := TStringList.Create;
  Try
    ValidNames.Add('History'); //0
    ValidNames.Add('ROS'); //1
    ValidNames.Add('Physical Exam'); //2
    ValidNames.Add('Psych Exam'); //3
    ValidNames.Add('Tests'); //4
    ValidNames.Add('Diagnosis'); //5
    ValidNames.Add('Medical Opinion'); //6
    ValidNames.Add('Competency'); //7
    ValidNames.Add('Psych Summary'); //8
    ValidNames.Add('Misc'); //9

    breakOutCount := 0;
    Repeat
      Begin
        retry := false;
        Inc(breakOutCount);
        For i := 0 To tabNames.Count - 1 Do
        Begin
          currName := tabNames[i];
          index := ValidNames.IndexOf(currName);
          If (index = -1) Then
          Begin
            If (i < tabNames.Count - 1) Then
              tabNames.Exchange(i, i + 1);
            If (i + 1 <= tabNames.Count - 1) Then
              If Not ((ValidNames.IndexOf(tabNames[i]) = -1) And (ValidNames.IndexOf(tabNames[i + 1]) = -1)) Then
                retry := True;
          End
          Else
            If (-1 < index) And (index < i) Then
            Begin
              If (i < tabNames.Count - 1) Then
                tabNames.Exchange(i, i + 1);
              If (i + 1 <= tabNames.Count - 1) Then
                If Not ((ValidNames.IndexOf(tabNames[i]) = -1) And (ValidNames.IndexOf(tabNames[i + 1]) = -1)) Then
                  retry := True;
            End;
        End;
        { Special so we don't get stuck in a loop. If the message below is encountered, then the
          conditions above have to be re-examined. }
        If breakOutCount > 15 Then
        Begin
          break;
        End;
    End Until Not retry;
  Finally
    ValidNames.Free;
  End;
End;

Function TCPWMObject.GetObjectTree(aObject: TCPWMObject): TCPWMObject;
Var
  parent: TCPWMObject;
Begin
  parent := aObject.Parent;
  If Not parent.IsTabScrollBox Then
    Result := GetObjectTree(parent)
  Else
    Result := aObject;
End;

Function TCPWMObject.GetUserObject(childen: TList): TCPWMObject;
Var
  i: integer;
  child: TCPWMObject;
Begin
  Result := Nil;
  For i := 0 To childen.Count - 1 Do
  Begin
    Result := Nil;
    child := childen.Items[i];

    If (child.IsUserObject And (Not child.Processed)) Then
    Begin
      child.Processed := true;
      Result := child;
    End;

    If child.HasChildren Then
      Result := self.GetUserObject(child.Children);

    If Assigned(Result) Then
      break;
  End;
End;

Function TCPWMObject.FindUserObject(IDNum: String; childen: TList): TCPWMObject;
Var
  i: integer;
  child: TCPWMObject;
Begin
  Result := Nil;
  For i := 0 To childen.Count - 1 Do
  Begin
    Result := Nil;
    child := childen.Items[i];

    If (child.IsUserObject And (child.HintValue = IDNum)) Then
    Begin
      Result := child;
    End;

    If child.HasChildren Then
      Result := self.FindUserObject(IDNum, child.Children); // recurse

    If Assigned(Result) Then
      break;
  End;
End;

Procedure TCPWMForm.SetTabs(childen: TList);
Var
  i: integer;
  child: TCPWMObject;
Begin
  For i := 0 To childen.Count - 1 Do
  Begin
    child := childen.Items[i];
    // The check of the parens'e ravent allowse us to ignore imbedded tabbed Notebooks.
    If (child.ControlType = 'TTabPage') And (child.Parent.Parent = Nil) Then
      Tabs.Add(child);
    If child.HasChildren Then
      self.SetTabs(child.Children); // recurse
  End;
End;

Function TCPWMExam.GetNotebook: TCPWMForm;
Var
  i: integer;
  aForm: TCPWMForm;
Begin
  Result := Nil;
  For i := 0 To self.Forms.Count - 1 Do
  Begin // incoming Exam
    aForm := self.Forms[i];
    If aForm.FormType = Notebook Then
    Begin
      Result := aForm;
      exit;
    End;
  End;
End;

Constructor TCPWMCode.Create;
Begin
  FirstRun := TStringList.Create;
  MainRun := TStringList.Create;
  LastRun := TStringList.Create;
  Script := TStringList.Create;
End;

Destructor TCPWMCode.Destroy;
Begin
  FirstRun.Free;
  MainRun.Free;
  LastRun.Free;
  Script.Free;

  Inherited;
End;

Procedure TCPWMCode.Merge(mergingCode: TCPWMCode; currentButtonName: String);
Var
  i: integer;
  workStringList: TStringList;
  outFirstRun, outMiddle, outLastRun: TStringList;
Begin
  outFirstRun := FirstRun; // base FirstRun Script
  Try
    workStringList := mergingCode.FirstRun;
    If workStringList.Count > 0 Then
      If outFirstRun.Count > 0 Then
      Try
        outFirstRun.Delete(outFirstRun.Count - 1); // remove last end if
      Except
      End;
    For i := 0 To workStringList.Count - 1 Do
    Begin
      If outFirstRun.Count > 0 Then
        If Pos(LowerCase(FIRST_RUN), LowerCase(workStringList[i])) > 0 Then
          continue; // ignore the merging if FirstRun line
      outFirstRun.Add(workStringList[i]);
    End;
  Except
    On E: Exception Do
      ShowMessageCAPRI('PerformScriptMerge: Failed FirstRun merge ;' + E.message);
  End;

  outMiddle := MainRun; // base middle Script
  Try
    outMiddle.Add(' ');
    workStringList := mergingCode.MainRun;
    For i := 0 To workStringList.Count - 1 Do
      outMiddle.Add(workStringList[i]);
  Except
    On E: Exception Do
      ShowMessageCAPRI('PerformScriptMerge: Failed MiddleRun merge ;' + E.message);
  End;
  LastRun.Insert(0, 'Set ReallyOKToExit=1');
  LastRun.Insert(0, 'DefineNumeric ReallyOKToExit');
  outLastRun := LastRun; // base LastRun Script
  Try
    workStringList := mergingCode.LastRun;
    If workStringList.Count > 0 Then
      If outLastRun.Count > 0 Then
      Try
        outLastRun.Delete(outLastRun.Count - 1); // remove last end if
      Except
      End;
    outLastRun.Add('If ExitOK=0');
    outLastRun.Add('  Set ReallyOKToExit=0');
    outLastRun.Add('EndIf');
    For i := 0 To workStringList.Count - 1 Do
    Begin
      If outLastRun.Count > 0 Then
        If Pos(LowerCase(LAST_RUN), LowerCase(workStringList[i])) > 0 Then
          continue; // ignore the merging if LastRun line
      outLastRun.Add(workStringList[i]);
    End;
  Except
    On E: Exception Do
      ShowMessageCAPRI('PerformScriptMerge: Failed LastRun merge ;' + E.message);
  End;

  Try
    outLastRun.Delete(outLastRun.Count - 1); // remove last end if
  Except
  End;
  outLastRun.Add('  If ExitOK=0');
  outLastRun.Add('    Set ReallyOKToExit=0');
  outLastRun.Add('  EndIf');
  outLastRun.Add('  Set ExitOK=ReallyOKToExit');
  outLastRun.Add('EndIf');
  { Now rebuild this Code object's script from these components }
  Script.Clear;
  Script.AddStrings(FirstRun);
  Script.AddStrings(MainRun);
  Script.AddStrings(LastRun);

End;

Procedure TCPWMCode.Parse;
Const
  first = 1;
  middle = 2;
  last = 3;
  none = 0;
Var
  i: integer;
  ifCount, endIfCount: integer;
  section: integer; // 0 = no section, 1 = First, 2 = Middle, 3 = Last
  scriptLine: String;
  x: integer;
Begin

  If Script.Count > 0 Then
  Begin
    x := Script.Count - 1;
    If trim(Script[x]) = '' Then
      Repeat
        Script.Delete(x);
        dec(x);
      Until (x = 0) Or (trim(Script[x]) <> '');
  End;

  ifCount := 0;
  endIfCount := 0;

  section := middle; // default to the middle section
  mainrun.Add('Label ' + self.examname + 'AUTOLABELBEGINNINGOFEXAMSCRIPT');
  For i := 0 To Script.Count - 1 Do
  Begin
    scriptLine := Script[i];
    If Pos(LowerCase(FIRST_RUN), LowerCase(TrimComment(scriptLine))) > 0 Then
    Begin
      section := first;
      ifCount := 0;
      endIfCount := 0;
    End
    Else
      If Pos(LowerCase(LAST_RUN), LowerCase(TrimComment(scriptLine))) > 0 Then
      Begin
        section := last;
        ifCount := 0;
        endIfCount := 0;
      End;

    If Pos(LowerCase('endif'), LowerCase(trim(TrimComment(scriptLine)))) = 1 Then
      Inc(endIfCount)
    Else
      If Pos(LowerCase('if '), LowerCase(trim(TrimComment(scriptLine)))) = 1 Then
        Inc(ifCount);

    If ifCount > endIfCount Then
      Case section Of
        first:
          FirstRun.Add(scriptLine);
        middle:
          MainRun.Add(scriptLine);
        last:
          LastRun.Add(scriptLine);
        none: ;
      End; // esac

    If ifCount = endIfCount Then // close the section
      Case section Of
        first:
          Begin
            FirstRun.Add(scriptLine); // the final endif
            section := middle;
          End;
        middle:
          Begin
            MainRun.Add(scriptLine); // the final endif
          End;
        last:
          Begin
            LastRun.Add(scriptLine); // the final endif
          End;
        none: ;
      End; // esac
  End; //rof
  mainrun.Add('If xCurrentExamNumber="' + self.examname + '"'); // Jump to end to speed up script
  mainrun.Add('Goto ENDOFMERGEDTEMPLATESSECTION');
  mainrun.Add('EndIf');
End;

Function TCPWMExam.CountScreenObjects: integer;
Var
  i: integer;
  childForm: TCPWMForm;
  screenObjectCount: integer;
Begin
  screenObjectCount := 0;

  { Count TCPWMObjects on the Forms }
  For i := 0 To Forms.Count - 1 Do
  Begin
    childForm := Forms[i];
    screenObjectCount := screenObjectCount + childForm.CountScreenObjects;
  End;

  Result := screenObjectCount;
End;

Procedure TCPWMExam.CountUserObjects(aTag: LongInt; Var userObjectList: TList);
Var
  i: integer;
  childForm: TCPWMForm;
Begin
  { Count User Objects on the Forms }
  For i := 0 To Forms.Count - 1 Do
  Begin
    childForm := Forms[i];
    childForm.CountUserObjects(aTag, userObjectList);
  End;
End;

Procedure TCPWMExam.SetTabOrder;
Var
  i: integer;
  childForm: TCPWMForm;
Begin
  For i := 0 To Forms.Count - 1 Do
  Begin
    childForm := Forms[i];
    childForm.SetTabOrder;
  End;
End;

Function TCPWMForm.CountScreenObjects: integer;
Var
  i: integer;
  child: TCPWMObject;
  screenObjectCount: integer;
Begin
  screenObjectCount := 0;

  { Count objects on the form }
  For i := 0 To Children.Count - 1 Do
  Begin
    child := Children[i];
    screenObjectCount := screenObjectCount + child.CountScreenObjects;
  End;

  Result := screenObjectCount;
End;

Procedure TCPWMForm.CountUserObjects(aTag: LongInt; Var userObjectList: TList);
Var
  i: integer;
  child: TCPWMObject;
Begin
  { Count objects on the form }
  For i := 0 To Children.Count - 1 Do
  Begin
    child := Children[i];
    child.CountUserObjects(aTag, userObjectList);
  End;
End;

Procedure TCPWMForm.SetTabOrder;
Var
  i: integer;
  child: TCPWMObject;
Begin
  For i := 0 To Children.Count - 1 Do
  Begin
    child := Children[i];
    child.SetTabOrder;
  End;
End;

Function TCPWMObject.CountScreenObjects: integer;
Var
  i: integer;
  child: TCPWMObject;
  screenObjectCount: integer;
Begin
  screenObjectCount := 1; // Count this self

  For i := 0 To children.Count - 1 Do
  Begin
    child := children.items[i];
    screenObjectCount := screenObjectCount + child.CountScreenObjects;
  End;
  Result := screenObjectCount;
End;

Procedure TCPWMObject.CountUserObjects(aTag: LongInt; Var userObjectList: TList);
Var
  i: integer;
  child: TCPWMObject;
Begin

  If IsUserObject And (self.Tag <> aTag) Then
  Begin
    userObjectList.Add(self);
    self.Tag := aTag;
  End;
  For i := 0 To Children.Count - 1 Do
  Begin
    child := Children.items[i];
    child.CountUserObjects(aTag, userObjectList);
  End;
End;

Function TOCompare(item1, item2: Pointer): integer;
Var
  obj1, obj2: TCPWMObject;
Begin
  obj1 := TCPWMObject(item1);
  obj2 := TCPWMObject(item2);
  If obj1.TabOrder < obj2.TabOrder Then
    Result := -1
  Else
    If obj1.TabOrder = obj2.TabOrder Then
      Result := 0
    Else
      Result := 1;
End;

Procedure TCPWMObject.SetTabOrder;
Var
  i: integer;
  child: TCPWMObject;
Begin

  If (ScreenObject Is TWinControl) Then
  Begin
    (screenObject As TWinControl).TabOrder := self.TabOrder;
  End;

  Children.Sort(TOCompare);

  For i := 0 To Children.Count - 1 Do
  Begin
    child := Children.items[i];
    child.SetTabOrder;
  End;

End;

{$IFDEF STANDALONE}
Procedure TCPWMExam.PrintPopups(AOwner: TComponent; OnClick,
  OnChange: TNotifyEvent; baseControl: TWinControl; ToPrinter: boolean; OnHint: TMouseMoveEvent);
Var
  i, j: integer;
  aForm: TCPWMForm;
Begin
  For i := 0 To Forms.Count - 1 Do {// for each form on the exam}
  Begin
    aForm := Forms.items[i];
    If aForm.FormType = Popup Then
    Begin
      For j := baseControl.ComponentCount - 1 Downto 0 Do
        baseControl.Components[j].Free;
      aForm.RenderToPrinter(AOwner, UserObjectList, ScreenObjectList, OnClick, OnChange, baseControl, 0, Nil);
      (AOwner As TFormPopupPrint).xStatusBar.SimpleText := aForm.Name;
      If ToPrinter Then
        (AOwner As TFormPopupPrint).Print // Print to Printer
      Else
        (AOwner As TFormPopupPrint).ShowModal; // Show on screen
    End;
  End;
End;

{$ENDIF}

Initialization
  RegisterClass(TColorCAPRIBtn);

End.
