unit dlgFeeInfoEdtr;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, MFunStr, Registry, untMiscMthds, untConstVals,
  dialogsCAPRI;

type
  TfrmFeeEdtrDlg = class(TForm)
    Panel1: TPanel;
    btnSave: TButton;
    btnDelete: TButton;
    Label1: TLabel;
    txtbxFeeEmlAddrss: TEdit;
    btnCncl: TButton;
    tmrWndwClose: TTimer;
    chkBxDsbleFee: TCheckBox;
    Panel2: TPanel;
    procedure txtbxFeeEmlAddrssKeyPress(Sender: TObject; var Key: Char);
    procedure txtbxFeeEmlAddrssChange(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
    procedure btnCnclClick(Sender: TObject);
    procedure chkBxDsbleFeeClick(Sender: TObject);
  private
    { Private declarations }
    function EnbleSaveBtn(): Boolean;
    function SaveFeeInfo(): Boolean;
    procedure ResetTmr;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent; TmrVal: Cardinal); reintroduce;
    destructor Destroy; override;
  end;

var
  frmFeeEdtrDlg: TfrmFeeEdtrDlg;

implementation

{$R *.dfm}

{ Public }

constructor TfrmFeeEdtrDlg.Create(AOwner: TComponent; TmrVal: Cardinal);
begin
  inherited Create(AOwner);
  tmrWndwClose.Interval := TmrVal;
end;

destructor TfrmFeeEdtrDlg.Destroy;
begin
  inherited;
end;

procedure TfrmFeeEdtrDlg.FormShow(Sender: TObject);
var
  stsVal: String;
begin
  txtbxFeeEmlAddrss.Text := GetRgstryVal(FEEREGKEY, FEEREGKEYADDRSSVAL);
  stsVal := GetRgstryVal(FEEREGKEY, FEEREGKEYSTSVAL);
  btnDelete.Enabled := (Not SameText(txtbxFeeEmlAddrss.Text, '')) or
       (Not SameText(stsVal, ''));
  chkBxDsbleFee.Checked := StrToBoolDef(stsVal, False);
end;

procedure TfrmFeeEdtrDlg.btnSaveClick(Sender: TObject);
begin
  { Button is ModalOK so form will close after event }
  if SaveFeeInfo then
    ShowMessageCAPRI('Fee Contact Information has been successfully saved.')
  else
    MessageDlgCAPRI('Unable to save Fee Contact Information to system registry.',
         mtError, [mbOK], 0);
end;

procedure TfrmFeeEdtrDlg.btnCnclClick(Sender: TObject);
begin
  if (Not btnDelete.Enabled) then  { Registry Entry does not currently exist }
    MessageDlgCAPRI('Remember to enter the Fee email contact information ' + #13#10 +
         'on any contracted exam email before selecting Send' + #13#10 +
         'if the request is not for an insufficient exam(s).', mtWarning, [mbOK], 0);
end;

procedure TfrmFeeEdtrDlg.btnDeleteClick(Sender: TObject);
begin
  { Button is ModalOK so form will close after event }
  if DeleteRstryVal(FEEREGKEY) then
    ShowMessageCAPRI('Fee Contact Information has been deleted.')
  else
    MessageDlgCAPRI('Error occurred deleting Fee Contact Information.', mtError, [mbOK], 0);
end;

{ Private }

procedure TfrmFeeEdtrDlg.txtbxFeeEmlAddrssChange(Sender: TObject);
var
  modTxtVal: String;
begin
  { Ensure's valid pasted text content }
  modTxtVal := Trim(txtbxFeeEmlAddrss.Text);  { Remove Leading and Trailing spaces }
  if (Not SameText(txtbxFeeEmlAddrss.Text, modTxtVal)) then
    txtbxFeeEmlAddrss.Text := modTxtVal;  { Update only if value changed }
  btnSave.Enabled := EnbleSaveBtn;
end;

procedure TfrmFeeEdtrDlg.txtbxFeeEmlAddrssKeyPress(Sender: TObject;
  var Key: Char);
begin
  if (((Length(txtbxFeeEmlAddrss.Text) = 0) or (txtbxFeeEmlAddrss.SelStart = 0))
       and (Key = #32)) then
    Key := #0;  { Prevent space in first position of text }
end;

procedure TfrmFeeEdtrDlg.chkBxDsbleFeeClick(Sender: TObject);
begin
  btnSave.Enabled := EnbleSaveBtn;
end;

function TfrmFeeEdtrDlg.EnbleSaveBtn(): Boolean;
var
  addrss: String;
begin
  ResetTmr;
  Result := True;
  addrss := txtbxFeeEmlAddrss.Text;
  { Ensure Valid Formatted Email Address if text entered }
  if ((SameText(Piece(addrss, '@'), '')) or
       (SameText(Piece(Piece(addrss, '@', 2), '.'), '')) or
       (SameText(Piece(Piece(addrss, '@', 2), '.', 2), ''))) then
    Result := False;
  { Allow saving just disable flag to registry }
  if SameText(addrss, '') and chkBxDsbleFee.Checked then
    Result := True;
end;

procedure TfrmFeeEdtrDlg.ResetTmr;
begin
  tmrWndwClose.Enabled := False;
  tmrWndwClose.Enabled := True;
end;

function TfrmFeeEdtrDlg.SaveFeeInfo(): Boolean;
var
  status: String;
begin
  status := '';  { False }
  if chkBxDsbleFee.Checked then
    status := '1';  { True }
  { Save Fee Email Address and then Status Field if successful }
  Result := SetRgstryVal(FEEREGKEY, FEEREGKEYADDRSSVAL, txtbxFeeEmlAddrss.Text);
  if Result then
    Result := SetRgstryVal(FEEREGKEY, FEEREGKEYSTSVAL, status);
end;

end.
