unit frmContractedBase;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, CheckLst, Fmcntrls, Fmcmpnts, clsPatientInfo,
  CCOWRPCBrokerCAPRI,     // DONE: CodeCR457 - LMS Changed from CCOWRPCBroker
  cls2507Info, ExtCtrls, untBrkrMthds, Diaccess, untMiscMthds, ComCtrls;

type
  { Contracted Exam Exception }
  ECntrctdExmError = class(Exception)
    Action: string;
    Code: integer;
    Mnemonic: string;
  end;
  TCntrctdBaseForm = class(TForm)
    grpbx2507Rqsts: TGroupBox;
    lblExamList: TLabel;
    btnRfrshRqsts: TButton;
    chkdLstBxExams: TCheckListBox;
    chkBxCntrctAllExams: TCheckBox;
    btnAction: TButton;
    btnCntrctdFrmCancel: TButton;
    tmrWndwClose: TTimer;
    tmrEnsrBrkrCnnctd: TTimer;
    fmLstrCntrctrs: TFMLister;
    Panel6: TPanel;
    Label4: TLabel;
    fmCmbBx2507Rqsts: TFMComboBox;
    cmbx2507Rqsts: TComboBox;
    procedure btnCntrctdFrmCancelClick(Sender: TObject);
    procedure btnRfrshRqstsClick(Sender: TObject);
    procedure fmCmbBx2507RqstsKeyPress(Sender: TObject; var Key: Char);
    procedure fmCmbBx2507RqstsChange(Sender: TObject);
    procedure btnActionClick(Sender: TObject);
    procedure chkBxCntrctAllExamsClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure chkdLstBxExamsClickCheck(Sender: TObject);
    procedure tmrWndwCloseTimer(Sender: TObject);
    procedure tmrEnsrBrkrCnnctdTimer(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure cmbx2507RqstsChange(Sender: TObject);
  protected
    { Protected declarations }
    _PtntInfo: TPatientInfo;
    _2507Info: T2507Info;
    IntCCOWBrkrRef: TCCOWRPCBrokerCAPRI;
    _fmTmpLstrExams: TFMLister;
    _RqstCntrctrs: TStringList;  { List of Exam Contractors for Request }
    procedure PopExamList(rqstIEN: String); virtual; abstract;
    procedure ExecuteBtnAction; virtual; abstract;
    procedure AllowAction;
    procedure GetContractors;
    procedure ResetWndwTmr;
  private
    { Private declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent; PtntInfo: TPatientInfo;
         CCOWBrkr: TCCOWRPCBrokerCAPRI; TmrVal: Cardinal); reintroduce;
    destructor Destroy; override;
  end;

var
  CntrctdBaseForm: TCntrctdBaseForm;

implementation

{$R *.dfm}

constructor TCntrctdBaseForm.Create(AOwner: TComponent; PtntInfo: TPatientInfo;
     CCOWBrkr: TCCOWRPCBrokerCAPRI; TmrVal: Cardinal);
begin
  inherited Create(AOwner);
  _PtntInfo := PtntInfo;
  IntCCOWBrkrRef := CCOWBrkr;
  _2507Info := T2507Info.Create(_PtntInfo, IntCCOWBrkrRef);
  _fmTmpLstrExams := TFMLister.Create(AOwner);
  _RqstCntrctrs := TStringList.Create;
  fmLstrCntrctrs.RPCBroker := IntCCOWBrkrRef;
  tmrWndwClose.Interval := TmrVal;
end;

destructor TCntrctdBaseForm.Destroy;
begin
  _2507Info.Free;
  _RqstCntrctrs.Free;
  inherited;
end;

procedure TCntrctdBaseForm.btnActionClick(Sender: TObject);
begin
  ResetWndwTmr;
  ExecuteBtnAction;
end;

procedure TCntrctdBaseForm.btnCntrctdFrmCancelClick(Sender: TObject);
begin
  ResetWndwTmr;
  Close;
end;

procedure TCntrctdBaseForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  tmrWndwClose.Enabled := False;
  tmrEnsrBrkrCnnctd.Enabled := False;
end;

procedure TCntrctdBaseForm.FormShow(Sender: TObject);
begin
  ResetWndwTmr;
end;

procedure TCntrctdBaseForm.tmrEnsrBrkrCnnctdTimer(Sender: TObject);
begin
  { Close Window if Broker cannot be reconnected }
  if Not EnsureBrokerConnected(IntCCOWBrkrRef) then
    Close;
end;

procedure TCntrctdBaseForm.tmrWndwCloseTimer(Sender: TObject);
begin
  { Shutdown Contracted window to allow app to close - tmr is set just short of
    Main Application's Shutdown Timer value }
  Close;
end;

{ Request / Exam Info }

procedure TCntrctdBaseForm.cmbx2507RqstsChange(Sender: TObject);
begin
  { Keep Comboboxes in Sync - for 508 COmptability }
  fmCmbBx2507Rqsts.ItemIndex := cmbx2507Rqsts.ItemIndex;
  fmCmbBx2507RqstsChange(Sender);
end;
  // gbh today
procedure TCntrctdBaseForm.fmCmbBx2507RqstsChange(Sender: TObject);
var
  rqstIEN: String;
begin
  ResetWndwTmr;
  chkBxCntrctAllExams.Checked := False;
  rqstIEN := fmCmbBx2507Rqsts.GetSelectedRecord.IEN;
  { Get Exams belonging to Request }
  _fmTmpLstrExams := _2507Info.Get2507Exams(rqstIEN);
  _RqstCntrctrs.Clear;  { Reset Contractor List }
  chkdLstBxExams.Items.Clear;  { Clear Exam List }
  chkBxCntrctAllExams.Enabled := False;
  { Populated Exam List Box for selected Request }
  PopExamList(rqstIEN);
end;

procedure TCntrctdBaseForm.fmCmbBx2507RqstsKeyPress(Sender: TObject;
  var Key: Char);
begin
  ResetWndwTmr;
  Key := #0;  { Prevent Key Events, as style cannot be changed to dropdownlist }
end;

procedure TCntrctdBaseForm.btnRfrshRqstsClick(Sender: TObject);
var
  itmsExist: Boolean;
begin
  ResetWndwTmr;
  fmCmbBx2507Rqsts.Text := '';
  fmCmbBx2507Rqsts.Items.Clear;  { Clear / Reset Contents }
  chkdLstBxExams.Items.Clear;
  chkBxCntrctAllExams.Checked := False;
  _2507Info.GetPatient2507Rqsts(fmCmbBx2507Rqsts);  { Add Requests to Combo-Box }
  { Add Delphi combobox over top FM combobox for 508 readability }
  cmbx2507Rqsts.Items.Clear();
  cmbx2507Rqsts.Items.AddStrings(fmCmbBx2507Rqsts.Items);
  itmsExist := (fmCmbBx2507Rqsts.Items.Count > 0);
  cmbx2507Rqsts.Enabled := itmsExist;
  fmCmbBx2507Rqsts.Enabled := itmsExist;
  chkdLstBxExams.Enabled := itmsExist;
  chkBxCntrctAllExams.Enabled := itmsExist;
  if itmsExist then
  begin
    cmbx2507Rqsts.ItemIndex := 0;
    fmCmbBx2507Rqsts.ItemIndex := 0;
    fmCmbBx2507RqstsChange(self);
  end;  { End If Request Items Exist in List }
  AllowAction;  { Ensures Button is Disabled on refresh - nothing checked }
end;

procedure TCntrctdBaseForm.chkBxCntrctAllExamsClick(Sender: TObject);
begin
  ResetWndwTmr;
  { Allow selection of specific exam(s) if all exam(s) NOT checked }
  chkdLstBxExams.Enabled := (Not chkBxCntrctAllExams.Checked);
end;

procedure TCntrctdBaseForm.chkdLstBxExamsClickCheck(Sender: TObject);
begin
  ResetWndwTmr;
end;

{ PROTECTED }

procedure TCntrctdBaseForm.ResetWndwTmr;
begin
  tmrWndwClose.Enabled := False;
  tmrEnsrBrkrCnnctd.Enabled := False;
  tmrWndwClose.Enabled := True;
  tmrEnsrBrkrCnnctd.Enabled := True;
end;

procedure TCntrctdBaseForm.GetContractors;
begin
  { Retrieve all contractors as contractors still assigned exams may have
    been subsequently inactivated. }
  fmLstrCntrctrs.Results.Clear;
  fmLstrCntrctrs.Screen := '';
  fmLstrCntrctrs.GetList(nil);
end;

procedure TCntrctdBaseForm.AllowAction;
begin
  { Allow Check-Out/In/Resend action if Exam(s) have been selected by the user. }
  btnAction.Enabled := (((GetCheckedCount(chkdLstBxExams) > 0) or
        (chkBxCntrctAllExams.Checked)));
end;

end.

