unit hintdisplay;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  StdCtrls,
  ComCtrls,
  ExtCtrls,
  PNCSMain,
  Main,
  Menus,
  ORFn,
  ClipBrd, Buttons;

type
  TformCustomHint = class(TForm)
    TimerShowHint: TTimer;
    RichEditHint: TMemo;
    MainMenu1: TMainMenu;
    Edit1: TMenuItem;
    Cut1: TMenuItem;
    Copy1: TMenuItem;
    Paste1: TMenuItem;
    Delete1: TMenuItem;
    SelectAll1: TMenuItem;
    N1: TMenuItem;
    Checkspelling1: TMenuItem;
    Panel1: TPanel;
    btnClose: TBitBtn;
    procedure TimerShowHintTimer(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure Edit1Click(Sender: TObject);
    procedure SelectAll1Click(Sender: TObject);
    procedure Delete1Click(Sender: TObject);
    procedure Paste1Click(Sender: TObject);
    procedure Copy1Click(Sender: TObject);
    procedure Cut1Click(Sender: TObject);
    procedure Checkspelling1Click(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure RichEditHintChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    FSourceMemo: TMemo;
    { Private declarations }
    FLoading: boolean;
    FTemplateMemoName: string;
    procedure SetTemplateMemoName(const Value: string);
  public
    property TemplateMemoName: string read FTemplateMemoName write
      SetTemplateMemoName;
    { Public declarations }
  end;

var
  formCustomHint: TformCustomHint;

implementation

uses connected;

{$R *.DFM}

procedure TformCustomHint.TimerShowHintTimer(Sender: TObject);
begin
  RichEditHint.SetFocus;
end;

procedure TformCustomHint.FormDeactivate(Sender: TObject);
begin
  if assigned(PNCSForm) then
    PNCSForm.xbuttondetails.visible := false;
end;

procedure TformCustomHint.Edit1Click(Sender: TObject);
begin
  if RichEditHint.SelText = '' then begin
    Cut1.Enabled := False;
    Copy1.Enabled := False;
    Delete1.Enabled := False;
  end
  else begin
    Cut1.Enabled := True;
    Copy1.Enabled := True;
    Delete1.Enabled := True;
  end;
  if ClipBoard.asText <> '' then
    Paste1.Enabled := True;

end;

procedure TformCustomHint.SelectAll1Click(Sender: TObject);
begin
  RichEditHint.SelectAll;
end;

procedure TformCustomHint.Delete1Click(Sender: TObject);
begin
  RichEditHint.SelText := '';
end;

procedure TformCustomHint.Paste1Click(Sender: TObject);
begin
  RichEditHint.PasteFromClipboard;
end;

procedure TformCustomHint.Copy1Click(Sender: TObject);
begin
  Clipboard.AsText := RichEditHint.SelText;
end;

procedure TformCustomHint.Cut1Click(Sender: TObject);
begin
  Clipboard.AsText := RichEditHint.SelText;
  RichEditHint.SelText := '';
end;

procedure TformCustomHint.Checkspelling1Click(Sender: TObject);
var
  Msg: TMsg;
  Handled: Boolean;
begin
  RichEditHint.SetFocus;
  msg.message := WM_KEYUP;
  Msg.wParam := VK_F7;
  handled := false;
  frmMain.AppMessage(msg, handled);
end;

procedure TformCustomHint.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TformCustomHint.RichEditHintChange(Sender: TObject);
begin
  if RichEditHint.ReadOnly then
    { turn off OnChange event assignment }
    RichEditHint.OnChange := nil;

  if FLoading then
    Exit;

  FSourceMemo.Text := RichEditHint.Text;
end;

procedure TformCustomHint.FormShow(Sender: TObject);
begin
  FLoading := True;
  FSourceMemo := pncsForm.xpanelBaseControl.FindComponent(FTemplateMemoName) as tMemo;
  richEditHint.Text := FSourceMemo.Text;
  richEditHint.SelStart := FSourceMemo.SelStart;
  FLoading := False;
end;

procedure TformCustomHint.FormCreate(Sender: TObject);
begin
  FLoading := True;
end;

procedure TformCustomHint.SetTemplateMemoName(const Value: string);
begin
  FTemplateMemoName := Value;
end;

procedure TformCustomHint.FormActivate(Sender: TObject);
begin
  //scroll caret into view (wParam and lParam parameters are ignored)
  SendMessage(richEditHint.Handle, EM_SCROLLCARET, 0, 0);
  richEditHint.SetFocus;
end;

end.

