(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"g); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TurboPower Internet Professional
 *
 * The Initial Developer of the Original Code is
 * TurboPower Software
 *
 * Portions created by the Initial Developer are Copyright (C) 2000-2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** *)

unit telnet;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  DialogsCAPRI,
  IpTerm,
  IpUtils,
  IpSock,
  StdCtrls,
  main,
  ExtCtrls,
  TRPCB,
  Buttons,
  jpeg,
  Menus,
  ComCtrls,
  Clipbrd,
  Dialogs,
  Fmcntrls,
  Diaccess,
  Fmcmpnts;

type
  TfrmTelnet = class(TForm)
    ServerEdit: TEdit;
    DisconnectBtn: TButton;
    TelnetPort: TCheckBox;
    PortEdit: TEdit;
    ExitBtn: TButton;
    IpClient1: TIpClient;
    IpVT100Emulator1: TIpVT100Emulator;
    PanelTerminal: TPanel;
    IpTerminal1: TIpTerminal;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    PanelTools: TPanel;
    BitBtnCopy: TBitBtn;
    StatusBar1: TStatusBar;
    Quit1: TMenuItem;
    Edit1: TMenuItem;
    Copy1: TMenuItem;
    Paste1: TMenuItem;
    N1: TMenuItem;
    ClearDisplay1: TMenuItem;
    ClearHistory1: TMenuItem;
    Tools1: TMenuItem;
    HistoryBuffer1: TMenuItem;
    Capture1: TMenuItem;
    BitBtnPaste: TBitBtn;
    BitBtnCapture: TBitBtn;
    Connect1: TMenuItem;
    Disconnect1: TMenuItem;
    Close1: TMenuItem;
    ConnectBtn: TButton;
    BitBtnConnect: TBitBtn;
    BitBtnDisconnect: TBitBtn;
    OpenDialog1: TOpenDialog;
    BitBtnCaptureOff: TBitBtn;
    ImageDisk: TImage;
    FMListerOption: TFMLister;
    FMListBoxOption: TFMListBox;
    SwitchSites1: TMenuItem;
    N3: TMenuItem;
    BitBtn1: TBitBtn;
    BitBtn2: TBitBtn;
    FontDialog1: TFontDialog;
    BitBtn4: TBitBtn;
    BitBtn5: TBitBtn;
    N2: TMenuItem;
    Print1: TMenuItem;
    PrintSelected1: TMenuItem;
    Options1: TMenuItem;
    IncreaseFontSize1: TMenuItem;
    DecreaseFontSize1: TMenuItem;
    SelectFont1: TMenuItem;
    Colors1: TMenuItem;
    FMListerTerminalType: TFMLister;
    FMGetsNewPersonFile: TFMGets;
    procedure ConnectBtnClick(Sender: TObject);
    procedure IpClient1Status(Sender: TObject; Socket: Cardinal;
      Event: TIpStatusType; const Connection: TIpConnRec;
      const StatRec: TIpSockStatRec);
    procedure DisconnectBtnClick(Sender: TObject);
    procedure TelnetPortClick(Sender: TObject);
    procedure IpClient1Error(Sender: TObject; Socket: Cardinal;
      ErrCode: Integer; const ErrStr: string);
    procedure ExitBtnClick(Sender: TObject);
    procedure IpClient1ReceivedBlob(const Line: string);
    procedure Disconnect1Click(Sender: TObject);
    procedure Connect1Click(Sender: TObject);
    procedure Close1Click(Sender: TObject);
    procedure BitBtnConnectClick(Sender: TObject);
    procedure BitBtnDisconnectClick(Sender: TObject);
    procedure ClearDisplay1Click(Sender: TObject);
    procedure ClearHistory1Click(Sender: TObject);
    procedure HistoryBuffer1Click(Sender: TObject);
    procedure Edit1Click(Sender: TObject);
    procedure IpTerminal1Click(Sender: TObject);
    procedure IpTerminal1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure Copy1Click(Sender: TObject);
    procedure BitBtnCopyClick(Sender: TObject);
    procedure Paste1Click(Sender: TObject);
    procedure BitBtnPasteClick(Sender: TObject);
    procedure BitBtnCaptureClick(Sender: TObject);
    procedure BitBtnCaptureOffClick(Sender: TObject);
    procedure SwitchSites1Click(Sender: TObject);
    procedure File1Click(Sender: TObject);
    procedure BitBtn2Click(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
    procedure BitBtn4Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure BitBtn5Click(Sender: TObject);
    procedure PrintSelected1Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure SaveSettings1Click(Sender: TObject);
    procedure Colors1Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    ServerStr: string;
  public
    { Public declarations }
    FMListBoxTerminalType: TFMListBox;
  end;

var
  frmTelnet: TfrmTelnet;
  LoggedInUsername: Boolean;
  LoggedInAccessCode: Boolean;
  LoggedInVTModeSent: Boolean;
  RepaintCount: Integer;

implementation

uses printsupport,
  telnetcolors;

{$R *.DFM}
// {$zR terminalfont.RES}

procedure TfrmTelnet.ConnectBtnClick(Sender: TObject);
var
  Temp: string;
begin
  ServerEdit.Text := ANUStrServer;

  IPTerminal1.ScrollBackRows := 0;
  Connect1.Enabled := False;
  Disconnect1.Enabled := False;

  IpTerminal1.ClearAll;
  ConnectBtn.Enabled := False;
  LoggedInUsername := False;
  LoggedInAccessCode := False;
  LoggedInVTModeSent := False;

  if (ServerEdit.Text = '') then begin
    ShowMessageCAPRI('No server specified.');
    ConnectBtn.Enabled := True;
    Connect1.Enabled := True;
    Disconnect1.Enabled := False;
    bitbtnDisconnect.Visible := False;
    bitbtnConnect.Visible := True;
  end
  else begin
    if TelnetPort.Checked then begin
      Temp := ServerEdit.Text + ':23';
    end
    else begin
      Temp := ServerEdit.Text + ':' + PortEdit.Text;
    end;
  end;
  StatusBar1.Simpletext := 'Trying to connect to ' + ServerEdit.Text + '...';
  IpClient1.ConnectSocket(Temp);
  ServerStr := Temp;
  if IPClient1.Connected = True then
    StatusBar1.Simpletext := 'Connected to ' + ServerStr;
  try
  IPTerminal1.SetFocus except
  end;
  IPTerminal1.ScrollBackRows := 1500;
end;

procedure TfrmTelnet.IpClient1Status(Sender: TObject; Socket: Cardinal;
  Event: TIpStatusType; const Connection: TIpConnRec;
  const StatRec: TIpSockStatRec);
begin
  case Event of
    stConnect: begin
        ConnectBtn.Enabled := False;
        DisconnectBtn.Enabled := True;
        if IPClient1.Connected = True then
          StatusBar1.Simpletext := 'Connected to ' + ServerStr;
        Connect1.Enabled := False;
        Disconnect1.Enabled := True;
        bitbtnDisconnect.Visible := True;
        bitbtnConnect.Visible := False;
        // If capture is on, then turn it off
        if bitbtnCaptureOff.Visible = True then
          bitbtnCaptureOffClick(Application);
        //frmTelnet.Caption := 'ExTelnet : Connected to ' + ServerStr;
        //try IpTerminal1.SetFocus except end;
      end;

    stDisconnect: begin
        ConnectBtn.Enabled := True;
        DisconnectBtn.Enabled := False;
        //frmTelnet.Caption := 'ExTelnet';
        StatusBar1.Simpletext := 'Disconnected';
        Connect1.Enabled := True;
        Disconnect1.Enabled := False;
        bitbtnDisconnect.Visible := False;
        bitbtnConnect.Visible := True;
        IPTerminal1.Clear;
        Copy1.Enabled := False;
        BitBtnCopy.Enabled := False;
        Paste1.Enabled := False;
        BitBtnPaste.Enabled := False;
        // If capture is on, then turn it off
        if bitbtnCaptureOff.Visible = True then
          bitbtnCaptureOffClick(Application);
      end;
  end;

  exit;
  // Set Copy/Paste Buttons
  Edit1Click(Application);
end;

procedure TfrmTelnet.DisconnectBtnClick(Sender: TObject);
begin
  IpClient1.CloseSocket;
  IPTerminal1.ClearAll;
  try
  IPTerminal1.SetFocus except
  end;
  IPTerminal1.ScrollBackRows := 0;
  Copy1.Enabled := False;
  BitBtnCopy.Enabled := False;
  Paste1.Enabled := False;
  BitBtnPaste.Enabled := False;
end;

procedure TfrmTelnet.TelnetPortClick(Sender: TObject);
begin
  PortEdit.Enabled := not (TelnetPort.Checked);
end;

procedure TfrmTelnet.IpClient1Error(Sender: TObject; Socket: Cardinal;
  ErrCode: Integer; const ErrStr: string);
begin
  ShowMessageCAPRI(ErrStr);
  ConnectBtn.Enabled := True;
  DisconnectBtn.Enabled := False;
  bitbtnDisconnect.Visible := False;
  bitbtnConnect.Visible := True;
  Connect1.Enabled := True;
  Disconnect1.Enabled := False;
end;

procedure TfrmTelnet.ExitBtnClick(Sender: TObject);
begin
  IpClient1.CloseSocket;
  Close;
end;

procedure TfrmTelnet.IpClient1ReceivedBlob(const Line: string);
var
  RPCBrokerTokenValue, RPCBrokerTokenValue2: string;
  x: integer;
  procedure sendString(text: string);
  var
    x: integer;
  begin
    ;
    if length(text) > 0 then
      for x := 1 to length(text) do
        IpVT100Emulator1.KeyPress(text[x]);
  end;
begin
  if LoggedInUsername = False then
    // Username: is used by VMS
    if (Length(line) - Pos(#$D + 'Username: ', line) = 10) or
      (Length(line) - Pos(#$A + 'Username: ', line) = 10)
      then begin
      ;
      LoggedInUsername := True;
      Sleep(500); // Wait 250msr
      if ANUTelnetLoginName <> '' then
        SendString(ANUTelnetLoginName + char(13))
      else
        SendString('vista' + char(13));
      exit;
    end;
  // login: is used by Linux
  if line = 'ogin: ' then begin
    ;
    LoggedInUsername := True;
    Sleep(250); // Wait 250msr
    if ANUTelnetLoginName <> '' then
      SendString(ANUTelnetLoginName + char(13))
    else
      SendString('vista' + char(13));
    exit;
  end;

  if Length(line) - Pos(#$A + 'DEVICE: HOME// ', line) = 15 then begin
    ;
    SendString('HOME;C-VT100;99999' + char(13));
    exit;
  end;
  if Length(line) - Pos(#$A + 'Select TERMINAL TYPE NAME: C-WYSE 75// ', line) = 39 then begin
    ;
    SendString('C-VT100' + char(13));
    exit;
  end;
  if Length(line) - Pos(#$A + 'Select TERMINAL TYPE NAME: C-VT100// ', line) = 37 then begin
    ;
    SendString('' + char(13));
    exit;
  end;

  if LoggedInAccessCode = False then
    if Length(line) - Pos(#$A + 'ACCESS CODE: ', line) = 13 then begin
      ;
      LoggedInUserName := True; // Just in case username login wasn't needed
      LoggedInAccessCode := True;
      Sleep(250); // Wait 250ms
      RPCBrokerTokenValue := GetAppHandle(frmMain.RPCBroker1); // Get a_ token for telnet login.
      (*
   frmMain.RPCBroker1.Results.Clear;
   frmmain.RPCBroker1.RemoteProcedure := 'XUS GET CCOW TOKEN';
   frmmain.RPCBrokerCall; try frmmain.RPCBroker1.Call;
     except
   On EBrokerError do begin;  ANURemoteProcedureCallInProgress:=False;
     frmMain.AnimateLogo(False);
     frmMain.StatusBarLoadPt.Caption:='XUS GET CCOW TOKEN could not be accessed!'; frmMain.StatusBarLoadPt.Repaint; Application.Processmessages;
     ShowMessageCAPRI('XUS GET CCOW TOKEN could not be accessed!');
   end; end;
   If frmmain.RPCBroker1.Results.Count>0 then
     RPCBrokerTokenValue2:=frmmain.RPCBroker1.Results[0]
   else RPCBrokerTokenValue2:='';
        *)

   //ShowmessageCAPRI(RPCBrokerTokenValue+char(13)+RPCBrokerTokenValue2);

      if RPCBrokerTokenValue <> '' then
        RPCBrokerTokenValue := '~~TOK~~' + RPCBrokerTokenValue + char(13);

      //clipboard.SetTextBuf(pchar(rpcbrokertokenvalue));
      sendstring(RPCBrokerTokenValue);

      (*
      if RPCBrokerTokenValue <> '' then
        for x:=1 to length(RPCBrokerTokenValue) do begin;
          SendString(RPCBrokerTokenValue[x]);
          Sleep(50);
          Application.ProcessMessages;
          Application.ProcessMessages;
          Application.ProcessMessages;
        end;
      *)
    end;
end;

procedure TfrmTelnet.Disconnect1Click(Sender: TObject);
begin
  DisconnectBtnClick(Application);
end;

procedure TfrmTelnet.Connect1Click(Sender: TObject);
begin
  ConnectBtnClick(Application);
end;

procedure TfrmTelnet.Close1Click(Sender: TObject);
begin
  frmMain.SaveCAPRISettings;
  DisconnectBtnClick(Application);
  FrmTelnet.Close;
end;

procedure TfrmTelnet.BitBtnConnectClick(Sender: TObject);
begin
  ConnectBtnClick(Application);
end;

procedure TfrmTelnet.BitBtnDisconnectClick(Sender: TObject);
begin
  DisconnectBtnClick(Application);
end;

procedure TfrmTelnet.ClearDisplay1Click(Sender: TObject);
begin
  IPTerminal1.Clear;
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.ClearHistory1Click(Sender: TObject);
begin
  IPTerminal1.ClearAll;
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.HistoryBuffer1Click(Sender: TObject);
begin
  if HistoryBuffer1.Checked = True then begin
    HistoryBuffer1.Checked := False;
    IPTerminal1.ScrollBack := False;
    IPTerminal1.ScrollBackRows := 0;
  end
  else begin
    HistoryBuffer1.Checked := True;
    IPTerminal1.ScrollBack := True;
    IPTerminal1.ScrollBackRows := 10000;
  end;
end;

procedure TfrmTelnet.Edit1Click(Sender: TObject);
begin
  // The extra if's here prevent button flickering
  if clipboard.AsText <> '' then begin
    if Paste1.Enabled = False then
      Paste1.Enabled := True;
    if BitBtnPaste.Enabled = False then
      BitBtnPaste.Enabled := True;
  end
  else begin
    if Paste1.Enabled = True then
      Paste1.Enabled := False;
    if BitBtnPaste.Enabled = True then
      BitBtnPaste.Enabled := False;
  end;
  if IPTerminal1.HaveSelection = True then begin
    if IPTerminal1.GetSelectedText <> '' then begin
      if Copy1.Enabled = False then
        Copy1.Enabled := True;
      if BitBtnCopy.Enabled = False then
        BitBtnCopy.Enabled := True;
    end;
  end
  else begin
    if Copy1.Enabled = True then
      Copy1.Enabled := False;
    if BitBtnCopy.Enabled = True then
      BitBtnCopy.Enabled := False;
  end;

  if bitbtnConnect.Visible = True then begin
    Copy1.Enabled := False;
    if BitBtnCopy.Enabled = True then
      BitBtnCopy.Enabled := False;
    Paste1.Enabled := False;
    if BitBtnPaste.Enabled = True then
      BitBtnPaste.Enabled := False;
  end;
end;

procedure TfrmTelnet.IpTerminal1Click(Sender: TObject);
begin
  // Set Paste Enabled/Disabled
  Edit1Click(Application);
end;

procedure TfrmTelnet.IpTerminal1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  // Set Paste Enabled/Disabled
  Edit1Click(Application);
end;

procedure TfrmTelnet.Copy1Click(Sender: TObject);
begin
  IPTerminal1.CopyToClipboard;
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.BitBtnCopyClick(Sender: TObject);
begin
  Copy1Click(Application);
end;

procedure TfrmTelnet.Paste1Click(Sender: TObject);
  procedure sendString(text: string);
  var
    x: integer;
  begin
    if length(text) > 0 then
      for x := 1 to length(text) do
        IpVT100Emulator1.KeyPress(text[x]);
  end;
begin
  try
  IPTerminal1.SetFocus except
  end;
  SendString(ClipBoard.AsText);
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.BitBtnPasteClick(Sender: TObject);
begin
  Paste1Click(Application);
end;

procedure TfrmTelnet.BitBtnCaptureClick(Sender: TObject);
var
  f: textfile;
begin
  if TelnetCaptureFile = '' then begin
    OpenDialog1.Title := 'Select a capture file.';
    if openDialog1.Execute then begin
      if OpenDialog1.FileName <> '' then begin
        TelnetCaptureFile := OpenDialog1.Filename;
        if FileExists(telnetcapturefile) then begin
          // Ask if overwrite
          if Application.MessageBox(Pchar('Do you really want to overwrite file ' + TelnetCaptureFile + '?'), 'Overwrite File?', 4) = 6 then begin
            // Yes Overwrite, so delete and recreate file here
            try
            DeleteFile(TelnetCaptureFile)except begin
                TelnetCaptureFile := '';
                ShowmessageCAPRI('Could not delete file.  Please choose a different filename.');
                exit;
              end;
            end;
            AssignFile(F, TelnetCaptureFile);
            Rewrite(F);
            CloseFile(F);
          end
          else begin
            TelnetCaptureFile := '';
            exit;
          end;
        end
        else begin
          // Create File here
          AssignFile(F, TelnetCaptureFile);
          Rewrite(F);
          CloseFile(F);
        end;
      end;
    end
    else begin
      exit;
    end;
  end
  else begin
    if fileexists(TelnetCaptureFile) = true then begin
      // Ask about appending
      if Application.MessageBox(Pchar('Do you want to append to the current ' + TelnetCaptureFile + '?'), 'Append To Existing File?', 4) = 6 then begin
        // Start capture here
        IPTerminal1.CaptureFile := TelnetCaptureFile;
        IPTerminal1.Capture := cmAppend;
        bitbtnCaptureOff.Visible := True;
        bitbtnCapture.Visible := False;
        ImageDisk.Visible := True;
        try
        IPTerminal1.SetFocus except
        end;
        exit;
      end
      else begin
        TelnetCaptureFile := '';
        BitBtnCaptureClick(self);
        exit;
      end;

    end
    else begin
      // Ask if should use the filename
      if Application.MessageBox(Pchar('Use filename ' + TelnetCaptureFile + '?'), 'Capture file name', 4) = 6 then begin
      end
      else begin
        TelnetCaptureFile := '';
        BitBtnCaptureClick(self);
        exit;
      end;
    end;
  end;
  // Start capture here
  IPTerminal1.CaptureFile := TelnetCaptureFile;
  IPTerminal1.Capture := cmOn;
  bitbtnCaptureOff.Visible := True;
  bitbtnCapture.Visible := False;
  ImageDisk.Visible := True;
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.BitBtnCaptureOffClick(Sender: TObject);
begin
  IPTerminal1.Capture := cmOff;
  bitbtnCaptureOff.Visible := False;
  bitbtnCapture.Visible := True;
  ImageDisk.Visible := False;
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.SwitchSites1Click(Sender: TObject);
begin
  frmMain.Connect1Click(Application);
end;

procedure TfrmTelnet.File1Click(Sender: TObject);
begin
  if frmMain.SwitchSites1.Enabled = True then
    SwitchSites1.Enabled := True
  else
    SwitchSites1.Enabled := False;
  PrintSelected1.Enabled := False;
  if IPTerminal1.HaveSelection = True then begin
    if IPTerminal1.GetSelectedText <> '' then begin
      PrintSelected1.Enabled := True;
    end;
  end;
end;

procedure TfrmTelnet.BitBtn2Click(Sender: TObject);
begin
  bitBtn1.Enabled := true;
  IPTerminal1.Font.Size := IPTerminal1.Font.Size - 1;
  if IPTerminal1.Font.Size <= 8 then begin
    bitbtn2.enabled := false;
  end;
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.BitBtn1Click(Sender: TObject);
begin
  bitbtn2.enabled := true;
  IPTerminal1.Font.Size := IPTerminal1.Font.Size + 1;
  if IPTerminal1.Font.Size >= 122 then begin
    bitbtn1.enabled := false;
  end;
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.BitBtn4Click(Sender: TObject);
begin
  FontDialog1.Font := IPTerminal1.Font;
  FontDialog1.Execute;
  IPTerminal1.Font := FontDialog1.Font;
  IPTerminal1.Repaint;
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.FormCreate(Sender: TObject);
begin
  //FMListBoxTerminalType
  FMListBoxTerminalType:= TFMListBox.Create(PanelTerminal);
  FMListBoxTerminalType.Left:=480;
  FMListBoxTerminalType.Top:=48;
  FMListBoxTerminalType.Width:=121;
  FMListBoxTerminalType.Height:=97;
  FMListBoxTerminalType.ItemHeight:=13;
  FMListBoxTerminalType.Sorted:=False;
  FMListBoxTerminalType.TabOrder:=2;
  FMListBoxTerminalType.Visible:=False;
  FMListBoxTerminalType.FMCtrlOptions:=[];
  FMListBoxTerminalType.FMLister:=FMListerTerminalType;
  //FMGetsNewPersonFile
  FMGetsNewPersonFile.FieldNumbers.SetText(PChar('.01'+#13+#10+'2'+#13+#10+'11'+#13+#10+'9.3'+#13+#10+'200.04'+#13+#10+'8983.13'));
  FMGetsNewPersonFile.RPCBroker:=frmMain.RPCBroker1;
  doublebuffered := true; // Prevent flickering
end;

procedure TfrmTelnet.BitBtn5Click(Sender: TObject);
var
  x: integer;
  DtTm: string;
  Division: string;
begin
  {Get Date}
  frmMain.RPCBroker1.RemoteProcedure := 'DVBAB DATETIME';
  try
    frmMain.RPCBroker1.Call;
  except
    on EBrokerError do begin
      ANURemoteProcedureCallInProgress := False;
      frmMain.AnimateLogo(False);
      frmMain.StatusBarLoadPt.Caption := 'RPC DVBAB DATETIME could not be accessed!';
      frmMain.StatusBarLoadPt.Repaint;
      Application.Processmessages;
      ShowMessageCAPRI('DVBAB DATETIME could not be accessed!');
    end;
  end;
  DtTm := frmMain.RPCBroker1.Results[0];

  {Get Division}
  frmMain.RPCBroker1.RemoteProcedure := 'XWB GET VARIABLE VALUE';
  frmMain.RPCBroker1.Param[0].Value := 'DUZ("2")';
  frmMain.RPCBroker1.Param[0].PType := reference;
  try
    frmMain.RPCBroker1.Call;
  except
    on EBrokerError do begin
      ANURemoteProcedureCallInProgress := False;
      frmMain.AnimateLogo(False);
      frmMain.StatusBarLoadPt.Caption := 'RPC XWB GET VARIABLE VALUE could not be accessed!';
      frmMain.StatusBarLoadPt.Repaint;
      Application.Processmessages;
      ShowMessageCAPRI('XWB GET VARIABLE VALUE could not be accessed!');
    end;
  end;
  Division := frmMain.RPCBroker1.Results[0];

  formPrint.QRMemoReport.Font.Size := 11;

  formPrint.QRMemoReport.Lines.clear;
  formPrint.QRLabel1.Caption := 'Telnet Print Screen';
  formPrint.QRLabel3.Caption := '';
  formPrint.QRLabel4.Caption := 'Printed on: ' + DtTm;
  formPrint.QRLabel6.Caption := 'Division: ' + Division;
  formPrint.QRLabel5.Caption := 'System: ' + Uppercase(ANUStrServer);

  formPrint.QRMemoReport.ParentRichEdit := nil;

  for x := 1 to IPTerminal1.Rows do
    formPrint.QRMemoReport.Lines.Add(IPTerminal1.Line[x]);

  formPrint.buttonPrintClick(Application);
  try
  IPTerminal1.SetFocus except
  end;
end;

procedure TfrmTelnet.PrintSelected1Click(Sender: TObject);
var
  x: integer;
  DtTm: string;
  Division: string;
begin
  {Get Date}
  frmMain.RPCBroker1.RemoteProcedure := 'DVBAB DATETIME';
  try
    frmMain.RPCBroker1.Call;
  except
    on EBrokerError do begin
      ANURemoteProcedureCallInProgress := False;
      frmMain.AnimateLogo(False);
      frmMain.StatusBarLoadPt.Caption := 'RPC DVBAB DATETIME could not be accessed!';
      frmMain.StatusBarLoadPt.Repaint;
      Application.Processmessages;
      ShowMessageCAPRI('DVBAB DATETIME could not be accessed!');
    end;
  end;
  DtTm := frmMain.RPCBroker1.Results[0];

  {Get Division}
  frmMain.RPCBroker1.RemoteProcedure := 'XWB GET VARIABLE VALUE';
  frmMain.RPCBroker1.Param[0].Value := 'DUZ("2")';
  frmMain.RPCBroker1.Param[0].PType := reference;
  try
    frmMain.RPCBroker1.Call;
  except
    on EBrokerError do begin
      ANURemoteProcedureCallInProgress := False;
      frmMain.AnimateLogo(False);
      frmMain.StatusBarLoadPt.Caption := 'RPC XWB GET VARIABLE VALUE could not be accessed!';
      frmMain.StatusBarLoadPt.Repaint;
      Application.Processmessages;
      ShowMessageCAPRI('XWB GET VARIABLE VALUE could not be accessed!');
    end;
  end;
  Division := frmMain.RPCBroker1.Results[0];

  formPrint.QRMemoReport.Font.Size := 11;

  formPrint.QRMemoReport.Lines.clear;
  formPrint.QRLabel1.Caption := 'Telnet Print Selected';
  formPrint.QRLabel3.Caption := '';
  formPrint.QRLabel4.Caption := 'Printed on: ' + DtTm;
  formPrint.QRLabel6.Caption := 'Division: ' + Division;
  formPrint.QRLabel5.Caption := 'System: ' + Uppercase(ANUStrServer);

  formPrint.QRMemoReport.ParentRichEdit := nil;

  formPrint.QRMemoReport.Lines.Add(IPTerminal1.GetSelectedText);

  formPrint.buttonPrintClick(Application);
end;

procedure TfrmTelnet.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  frmMain.SaveCAPRISettings
end;

procedure TfrmTelnet.SaveSettings1Click(Sender: TObject);
begin
  frmMain.SaveCAPRISettings;
end;

procedure TfrmTelnet.Colors1Click(Sender: TObject);
begin
  frmTelnetColors := TfrmTelnetColors.Create(frmTelnet);
  frmTelnetColors.ColorCombo2.Selection := ipc_TermBufForeColor;
  frmTelnetColors.ColorCombo1.Selection := ipc_TermBufBackColor;
  if frmTelnetColors.ShowModal = mrOK then begin
    ipc_TermBufForeColor := frmTelnetColors.ColorCombo2.Selection;
    ipc_TermBufBackColor := frmTelnetColors.ColorCombo1.Selection;
    ShowmessageCAPRI('Please shut down and restart CAPRI to activate your changes.');
  end;
  frmTelnetColors.Release;
end;

procedure TfrmTelnet.FormDestroy(Sender: TObject);
begin
  FreeAndNil(FMListBoxTerminalType);
end;

end.
