unit untAddtnlMMthds;

interface

uses SysUtils, MFunStr, Classes;

function FMInternalToDelphiDate(FMIntDate: String): String;
function DelphiDateToFMInternal(DDate: TDateTime): String;
function DelphiDateTimeToHorolog(DDate: TDateTime): String;
function Occurrences(const subString, strVal: String): Integer;
{ The following functions were pulled out of the CAPRI SUPPORT unit to
  support the MANAGE REPORTS form implemented in CAPRI. }
function FMDateTimeConvert(DateTimeST: String): String;
function FMDateTimeConvertDetail(DateTimeST: String; v: Integer): String;
function FMTimeConvert(FMTimeSt: String): String;
function GetMonthName(month: Integer): String;
function GetMonthNumber(month: Integer): String;

implementation

uses
  DateUtils;

function FMInternalToDelphiDate(FMIntDate: String): String;
begin
  Result := '';
  FMIntDate := Piece(FMIntDate, '.');  { In case date and time sent }
  try
    { Ensure valid Date sent }
    if ((Length(FMIntDate) = 7) and (StrToInt(FMIntDate) > 0)) then
    begin
      Result := Copy(FMIntDate, 4, 2);  { Month}
      Result := Result + '/' + Copy(FMIntDate, 6, 2);  { Day }
      Result := Result + '/' + IntToStr(StrToInt(Copy(FMIntDate, 1, 3)) + 1700);  { Year }
    end;  { End If FM Date component is Valid (7 Chars Ie. 3110101) }
  except
    { Do Nothing on EConvertError Exception }
  end;  { End Try/Except Block }
end;

function DelphiDateToFMInternal(DDate: TDateTime): String;
var
  i, num: Integer;
  strDte, strVal: String;
begin
  Result := '';
  strDte := DateToStr(DDate);
  num := Occurrences('/', strDte);
  for i := 1 to num do
  begin
    strVal := Piece(strDte, '/', i);
    case i of
      1,2 : { Month / Day }
        if Length(strVal) = 1 then
          Result := Result + '0' + strVal
        else
          Result := Result + strVal;
      3 : { Year }
        Result := IntToStr((StrToInt(strVal) - 1700)) + Result;
    end;  { End Case Selection }
  end;  { End For Loop }
end;

{
LMS - Proper Horolog Value being from a TDateTime with the "seconds" part
       DelphiDateToFMInternal is an unnecessary and wrong kludge. We're stuck with it because
         all the data is in that format. Use this function for new storage of date/time.
}
function DelphiDateTimeToHorolog(DDate: TDateTime): String;
begin
  //CacheBaseDateTime := -21548; // = StrToDateTime('12/31/1840');        // for conversion to and from Horolog
  Result := IntToStr(DaysBetween(DDate, -21548 {CacheBaseDateTime})) + ',' + IntToStr(SecondOfTheDay(DDate));
end;

function Occurrences(const subString, strVal: String): Integer;
{ Models some of the $l Functionality in M }
begin
  if (Length(subString) = 0) or (Length(strVal) = 0) then
    Result := 0
  else
    Result := (((Length(strVal) - Length(StringReplace(strVal, subString, '',
                                   [rfReplaceAll]))) div Length(substring)) + 1);
end;

{ The following functions were pulled out of the CAPRI SUPPORT unit to
  support the MANAGE REPORTS form implemented in and copied from CAPRI (177). }
function FMDateTimeConvert(DateTimeST: String): String;
begin
  Result := FMDateTimeConvertDetail(DateTimeST, 1);
end;

function FMDateTimeConvertDetail(DateTimeST: String; v: Integer): String;
var
  Code: Integer;
  DateInt: Integer;
  FMDate: String;
  fmMonth: Integer;
  fmYear: Integer;
  sgDay: String;
  sgMonth: String;
  sgTime: String;
  sgYear: String;
begin
  Result := '';
  if v < 1 then
    v := 1;
  if v > 3 then
    v := 3;
  DateTimeST := Trim(DateTimeST);
  if DateTimeST <> '' then
  begin
    sgDay := '';
    sgMonth := '';
    sgTime := '';
    sgYear := '';
    FMDate := Piece(DateTimeST, '.', 1);
    {validate Date string}
    Val(FMDate, DateInt, code);
    if (code = 0) and
       (DateInt >= 1000000) and
       (DateInt <= 9991231) then  //valid
    begin
      {get year}
      Val(Copy(DateTimeST, 1, 3), fmYear, code);
      fmYear := fmYear + 1700;
      Str(fmYear, sgYear);

      {get month}
      Val(Copy(DateTimeST, 4, 2), fmMonth, code);
      if code <> 0 then
        fmMonth := 0;

      {get day}
      sgDay := Copy(DateTimeST, 6, 2);

      case v of
      1:  {Format:  JAN 01, 2001@12:35}
        begin
          {convert month}
          sgMonth := GetMonthName(fmMonth);
          {filter day}
          if sgDay = '00' then
            sgDay := ''
          else if sgDay[1] = '0' then
            sgDay := Copy(sgDay, 2, 1);
          {get time}
          sgTime := FMTimeConvert(Piece(DateTimeSt, '.', 2));

          {format result}
          if sgDay <> '' then
            Result := sgMonth + ' ' + sgDay + ', ' + sgYear
          else if Trim(sgMonth) <> '' then
            Result := sgMonth + ', ' + sgYear
          else
            Result := sgYear;
          if sgTime <> '' then
            Result := Result + '@' + sgTime;
        end;  {v = 1}
      2:  {Format: 01/01/2001@12:35}
        begin
          {convert month}
          sgMonth := GetMonthNumber(fmMonth);
          {get time}
          sgTime := FMTimeConvert(Piece(DateTimeSt, '.', 2));
          {format result}
          if sgTime <> '' then
            Result := sgMonth + '/' + sgDay + '/' + sgYear + '@' + sgTime
          else
            Result := sgMonth + '/' + sgDay + '/' + sgYear;
        end;  {v = 2}
      3:  {Format: 01/01/2001}
        begin
          {convert month}
          sgMonth := GetMonthNumber(fmMonth);
          Result := sgMonth  + '/' + sgDay + '/' + sgYear;
        end;  {v = 3}
      end;  {case v}
    end  {valid FM DateTimeSt}
    else
      Result := DateTimeST;
  end;  {DateTimeSt <> ''}
end;

function GetMonthName(month: Integer): String;
var
  sgMonth: String;
begin
  case Month of
    1: sgMonth := 'JAN';
    2: sgMonth := 'FEB';
    3: sgMonth := 'MAR';
    4: sgMonth := 'APR';
    5: sgMonth := 'MAY';
    6: sgMonth := 'JUN';
    7: sgMonth := 'JUL';
    8: sgMonth := 'AUG';
    9: sgMonth := 'SEP';
    10: sgMonth := 'OCT';
    11: sgMonth := 'NOV';
    12: sgMonth := 'DEC';
  else  {default}
    sgMonth := '   ';
  end;
  Result := sgMonth;
end;

function GetMonthNumber(Month: Integer): String;
var
  sgMonth: String;
begin
  case Month of
    1: sgMonth := '01';
    2: sgMonth := '02';
    3: sgMonth := '03';
    4: sgMonth := '04';
    5: sgMonth := '05';
    6: sgMonth := '06';
    7: sgMonth := '07';
    8: sgMonth := '08';
    9: sgMonth := '09';
    10: sgMonth := '10';
    11: sgMonth := '11';
    12: sgMonth := '12';
  else  {default}
    sgMonth := '00';
  end;
  Result := sgMonth;
end;

function FMTimeConvert(FMTimeSt: String): String;
var
  Pad, Hours, Minutes: String;
  FMTimeInt, code: Integer;
begin
  Result := '';
  Val(FMTimeSt, FMTimeInt, code);
  if (code = 0) and (FMTimeInt > 0) then
  begin
    Pad := StringOfChar('0', 6);
    FMTimeSt := FMTimeSt + Copy(Pad, 1, 6 - Length(FMTimeSt));
    Hours := Copy(FMTimeSt, 1, 2);
    Minutes := Copy(FMTimeSt, 3, 2);
    Result := Hours + ':' + Minutes;
  end;
end;

end.
