unit untClsInsuf;

interface

uses
  FMCmpnts, MFunStr, TRpcb, Classes, SysUtils, untMiscMthds;

type
  TInsExam = class(TObject)  // VA FileMan file # 396.4
    private
      _Broker: TRPCBroker;
    protected
      FIfn: String;
      FRefNum: String;      // .01
      FRequest: String;     // .02
      FExamType: String;    // .03
      FStatus: String;     // .04
      FExamDate: String;    // .06
      FResults: String;     // 70
      FContractor: String;  // 100
      FTrnsfrDate: String;  // 101
      FRcvdDate: String;    // 102
    public
      constructor Create;
      destructor Destroy; override;
      function Clear: Boolean;
      function Copy(AExam: TInsExam): Boolean;
      function LoadFromVista(AFMGets: TFMGets; AIfn: String): Boolean;
      function GetIdn(): String;
      property Ifn: String read FIfn;
      property RefNum: String read FRefNum;
      property Request: String read FRequest;
      property ExamType: String read FExamType;
      property Status: String read FStatus write FStatus;
      property ExamDate: String read FExamDate write FExamDate;
      property Results: String read FResults write FResults;
      property Contractor: String read FContractor;
      property TransferDate: String read FTrnsfrDate;
      property DateRecieved: String read FRcvdDate write FRcvdDate;
  end;

type
  TInsRequest = class(TObject)
    private
      _Broker: TRPCBroker;
      function GetExam(Index: Integer): TInsExam;
      procedure SetExam(Index: Integer; const Value: TInsExam);
    protected
      FCount: Integer;
      FIfn: String;
      FExams: array of TInsExam;
    public
      constructor Create;
      destructor Destroy; override;
      function Add(AExam: TInsExam): Integer;
      function Delete(Index: Integer): Boolean;
      function Clear: Boolean;
      function GetExams(ALister: TFMLister; AIfn: String): Boolean;
      function GetIdn(Vendor: String): String;
      function PrimaryExam(Vendor: String): TInsExam;
      property Count: Integer read FCount;
      property Ifn: String read FIfn;
      property Exams[Index: Integer]: TInsExam read GetExam write SetExam;
  end;

implementation

{ TInsExam }

function TInsExam.Clear: Boolean;
begin
  FIfn := '-1';
  FRefNum := '';      // .01
  FRequest := '';     // .02
  FExamType := '';    // .03
  FStatus := '';     // .04
  FExamDate := '';    // .06
  FResults := '';     // 70
  FContractor := '';  // 100
  FTrnsfrDate := '';  // 101
  FRcvdDate := '';    // 102
end;

function TInsExam.Copy(AExam: TInsExam): Boolean;
begin
  FIfn := AExam.FIfn;
  FRefNum := AExam.FRefNum;
  FRequest := AExam.FRequest;
  FExamDate := AExam.FExamDate;
  FExamType := AExam.FExamType;
  FStatus := AExam.FStatus;
  FResults := AExam.FResults;
  FContractor := AExam.FContractor;
  FTrnsfrDate := AExam.FTrnsfrDate;
  FRcvdDate := AExam.FRcvdDate;
  Result := True;
end;

constructor TInsExam.Create;
begin
  Clear;
end;

destructor TInsExam.Destroy;
begin
  _Broker := nil;
  inherited;
end;

function TInsExam.GetIdn: String;
begin
  Result := '';
  if (Assigned(_Broker) and (CompareText(FRequest, '') <> 0) and
    (CompareText(FIfn, '') <> 0)) then
      Result := Piece(_Broker.User.Division, '^')
       + '-' + Piece(FRequest, '^') + '-' + FIfn;
end;

function TInsExam.LoadFromVista(AFMGets: TFMGets; AIfn: String): Boolean;
var
  FMGets: TFMGets;
begin
  _Broker := AFMGets.RPCBroker;
  FMGets := TFMGets.Create(nil);
  FMGets.RPCBroker := _Broker;
  FMGets.FileNumber := '396.4';
  FMGets.FieldNumbers.Clear;
  FMGets.AddField('.01');
  FMGets.AddField('.02');
  FMGets.AddField('.03');
  FMGets.AddField('.04');
  FMGets.AddField('.06');
  FMGets.AddField('70');
  FMGets.AddField('100');
  FMGets.AddField('101');
  FMGets.AddField('102');
  FMGets.IENS := AIfn;

  try
    FMGets.GetData;
  except
    Result := False;
  end;

  FIfn := NumStr(FMGets.IENS);
  FRefNum := FMGets.GetField('.01').FMDBExternal;
  FRequest := FMGets.GetField('.02').FMDBInternal + '^' + FMGets.GetField('.02').FMDBExternal;
  FExamType := FMGets.GetField('.03').FMDBInternal + '^' + FMGets.GetField('.03').FMDBExternal;
  FStatus := FMGets.GetField('.04').FMDBInternal;
  FExamDate := FMGets.GetField('.06').FMDBInternal + '^' + FMGets.GetField('.06').FMDBExternal;
  FResults := FMGets.GetField('70').FMDBInternal;
  FContractor := FMGets.GetField('100').FMDBInternal + '^' + FMGets.GetField('100').FMDBExternal;
  FTrnsfrDate := FMGets.GetField('101').FMDBInternal + '^' + FMGets.GetField('101').FMDBExternal;
  FRcvdDate := FMGets.GetField('102').FMDBInternal + '^' + FMGets.GetField('102').FMDBExternal;

  FMGets.Free;
  Result := True;
end;

{ TInsRequest }

function TInsRequest.Add(AExam: TInsExam): Integer;
begin
  Result := FCount;
  FExams[Result] := TInsExam.Create;
  FExams[Result].Copy(AExam);
  Inc(FCount);
  SetLength(FExams, FCount + 1);
end;

function TInsRequest.Clear: Boolean;
begin
  while (FCount > 0) do Delete(FCount - 1);
  Result := True;
end;

constructor TInsRequest.Create;
begin
  FCount := 0;
  SetLength(FExams, 1);
end;

function TInsRequest.Delete(Index: Integer): Boolean;
begin
  FExams[Index].Free;
  Dec(FCount);
end;

destructor TInsRequest.Destroy;
begin
  _Broker := nil;
  Clear;
  inherited;
end;

function TInsRequest.GetExam(Index: Integer): TInsExam;
begin
  Result := FExams[Index];
end;

function TInsRequest.GetExams(ALister: TFMLister; AIfn: String): Boolean;
var
  i: Integer;
  U: String;
  Exam: TInsExam;
  slExams: TStringList;
  Lister: TFMLister;
begin
  U := '^';
  slExams := TStringList.Create;
  _Broker := ALister.RPCBroker;
  FIfn := AIfn;
  Exam := TInsExam.Create;
  Lister := TFMLister.Create(nil);
  Lister.RPCBroker := ALister.RPCBroker;
  Lister.FileNumber := '396.4';
  Lister.FMIndex := 'C';
  Lister.Delimiter := '^';
  Lister.ListerFlags := [lfInternal];
  Lister.Screen := 'I $P(^(0),U,2)="' + AIfn + '"';
//  Lister.FieldNumbers.Add('.01');
  Lister.FieldNumbers.Add('.02');
  Lister.FieldNumbers.Add('.03');
  Lister.FieldNumbers.Add('.04');
  Lister.FieldNumbers.Add('.06');
//  Lister.FieldNumbers.Add('70');
  Lister.FieldNumbers.Add('100');
  Lister.FieldNumbers.Add('101');
  Lister.FieldNumbers.Add('102');
//  Lister.DisplayFields.Add('.01');
  Lister.DisplayFields.Add('.02');
  Lister.DisplayFields.Add('.03');
  Lister.DisplayFields.Add('.04');
  Lister.DisplayFields.Add('.06');
//  Lister.DisplayFields.Add('70');
  Lister.DisplayFields.Add('100');
  Lister.DisplayFields.Add('101');
  Lister.DisplayFields.Add('102');
  Lister.GetList(slExams);
  i := 0;
  while (i < Lister.Results.Count) do
  begin
    Exam.Clear;
    Exam.FIfn := Lister.Results[i];
    Exam.FRefNum := Piece(slExams[i], U, 1);
    Exam.FRequest := Piece(slExams[i], U, 2);
    Exam.FExamType := Piece(slExams[i], U, 3);
    Exam.FStatus := Piece(slExams[i], U, 4);
    Exam.FExamDate := Piece(slExams[i], U, 5);
//    Exam.FResults := Piece(slExams[i], U, 6);
    Exam.FContractor := Piece(slExams[i], U, 6);
    Exam.FTrnsfrDate := Piece(slExams[i], U, 7);
    Exam.FRcvdDate := Piece(slExams[i], U, 8);
    Add(Exam);
    Inc(i);
  end;
  Exam.Free;
  Lister.Free;
  slExams.Free;
end;

function TInsRequest.GetIdn(Vendor: String): String;

  function FirstExamIfn(Vndr: String): String;
  var
    i: Integer;
    Done: Boolean;
  begin
    Result := '-1';
    Done := False;
    i := 0;
    while (i < FCount) and (not Done) do
    begin
      if (CompareText(Vndr, FExams[i].FContractor) = 0) then
      begin
        Result := FIfn;
        Done := True;
      end;
      Inc(i);
    end;
  end;

  begin
  Result := '-1';
  if (Assigned(_Broker) and (StrToInt(FIfn) >= 0) and
    (StrToInt(FirstExamIfn(Vendor)) >= 0)) then
      Result := _Broker.User.Division + '-' + FIfn + '-' + FirstExamIfn(Vendor);
end;

function TInsRequest.PrimaryExam(Vendor: String): TInsExam;
  // Primary Exam simply is the first exam in the Request's exam list for that contractor.
  //  If an exam was to be assigned to a different contractor, then the primary would change.
  //  This is most likely only relevant in SFTP scenario
var
  i: Integer;
  Done : Boolean;
begin
  i := 0;
  Result := nil;
  Done := False;
  while ((i < FCount) and (not Done)) do
  begin
    if (CompareText(Vendor, Piece(Exams[i].FContractor, '^')) = 0) then
    begin
      Result := Exams[i];
      Done := True;
    end;
    Inc(i);
  end;
end;

procedure TInsRequest.SetExam(Index: Integer; const Value: TInsExam);
begin
  FExams[Index] := Value;
end;

end.
