unit untMiscMthds;

interface

uses Windows, Forms, Classes, Controls, CheckLst, Registry, XMLIntf, XMLDoc, SysUtils,
  TlHelp32, ShellAPI, StdCtrls, ComCtrls, MFunStr, untAddtnlMMthds, Fmcntrls,
  StrUtils;

type
  TStrArray = Array of String;
  QuickCopyError = class(Exception);

type
  TVersionParam = (vpCompanyName, vpFileDescription, vpFileVersion,
    vpInternalName, vpLegalCopyright, vpLegalTradeMarks,
    vpOriginalFilename, vpProductName, vpProductVersion, vpComments);

function GetCheckedCount(chkdLstBx: TCheckListBox): Integer;
function GetCheckedEnabledCount(chkdLstBx: TCheckListBox): Integer;
function GetCheckedOpenCount(chkdLstBx: TCheckListBox): Integer;
function GetCheckedOpenOneVendorCount(chkdLstBx: TCheckListBox): Integer;
function GetItemIndex(fmCmbBxObj: TFMComboBox; IENVal: String): Integer;
function GetTempDir: TFileName;
function GetWindowsDir: String;
// function GetWindowsTempDir: String;   // LMS - CodeCR496 - 2014-07-14 Cosolidated with GetTempDir and consolidate/refactor all GetTempDir into one location, so all can benefit from Win 7 specific handling.
function GetRgstryVal(KeyLoc, Key: String): String;
function SetRgstryVal(KeyLoc, Key, KeyVal: String): Boolean;
function DeleteRstryVal(KeyLoc: String): Boolean;
function ProcessExists(exeName: String): Boolean;

function StringToArray(const strVal: String; const dlmtr: String): TStrArray;

function ValidateTypeAndInstance(aObj: Pointer; ofClass: TClass): Boolean;

procedure FreeAndClearTStrings(aList: TStrings);

procedure EmptyListAndFreeObjects(aList: TList);
procedure FreeAndNilListWithObjects(var aList: TList);

procedure EmptyTStringsAndFreeObjects(aList: TStrings);
procedure FreeAndNilTStringsWithObjects(var aList: TStrings);

function VersionInfo(P: TVersionParam): String;
function NumStr(X: String): String;
function FMDTtoExternal(FMDT: String): String;

function GetDivisionOfCBOC(CBOC: string): string;                               // CodeCR457 - LMS

procedure FormatXMLDoc(XMLFile: String);
procedure SaveStrXMLInFileAndXMLDoc(var strXml: string; filePrefix: string; fileSuffix: string; AXMLDom: TXMLDocument; SaveLoaded: Boolean = false); overload;
procedure SaveStrXMLInFileAndXMLDoc(var strXml: string; TempSubFolder: string; filePrefix: string; fileSuffix: string; AXMLDom: TXMLDocument; SaveLoaded: Boolean = false); overload;
function GenericGetXMLNodeText(aNode: IXMLNode; aNodeName: string = ''): string;
function FindNodeByNodeName(rootLevel: IXMLNode; NodeNameToFind: string): IXMLNode;
function FindNodeInPathByNodeName(rootLevel: IXMLNode; PathAndNodeName: string; separator: string = '^'): IXMLNode;
function FindOrInsertNodeInPathByNodeName(rootLevel: IXMLNode; PathAndNodeName: string; separator: string = '^'): IXMLNode;

procedure AddChildWithTextValue(aAttachNode: IXMLNode; cNodeName: string; cNodeText: string = '');
procedure UpdateChildWithTextValue(aAttachNode: IXMLNode; cNodeName: string; cNodeText: string = '');

procedure AddDocumentsWrapperIfMissing(var rootLevel: TXMLDocument);

function FormattedTimeStamp(ATimeStamp: TDateTime): String;

function IsDateValid(aDateStr: string): Boolean;                                // CodeCR457 - LMS

procedure QuickCopy(AFrom, ATo: TObject);                                       { Copied Out of ORFn }
procedure SaveStrToFle(const FileName: TFileName; const content: String);

procedure EnsureDirectoryExists(dirName: String);
function GetTempExmDir(DUZ: String): String;

function MakeOutputFileStreamReady(ResponseOutputFile: string): TFileStream;

function GetStringGUID: string;

// call in FormMouseWheel and pass a TScrollingWincontrol as inBox
// can use TScrollBox, TListBox
// function uses PostMessage to send WM_VSCROLL messages to the control.
function ScrollBox(inBox: TWinControl;                                          // CodeCR700 rpk 3/5/2015
  inPoint: TPoint; WheelDelta: Integer): Boolean;

// note right and bottom edges are not included in the boundary per Delphi 2006 help
function RectContains(Rect: TRect; Point: TPoint): boolean;                     // rpk 3/5/2015 {from ORFn}


implementation

uses ActiveX                                                                    // rra 819092
  , comobj
{$IFDEF CAPRIDEVELOPER}
  , dialogscapri
{$ENDIF}
  , Messages
  ;

function GetCheckedCount(chkdLstBx: TCheckListBox): Integer;
var
  i: Integer;
begin
  Result := 0;
  For i := 0 to chkdLstBx.Items.Count - 1 Do
  begin
    If chkdLstBx.Checked[i] Then
      Result := Result + 1;
  end;                                                                          { End For Loop }
end;

function GetCheckedEnabledCount(chkdLstBx: TCheckListBox): Integer;
var
  i: Integer;
begin
  Result := 0;
  For i := 0 to chkdLstBx.Items.Count - 1 Do
  begin
    If chkdLstBx.Checked[i] and chkdLstBx.ItemEnabled[i] Then                   // DONE: LMS New method to handle disabled exams
      Result := Result + 1;
  end;                                                                          { End For Loop }
end;

function GetCheckedOpenCount(chkdLstBx: TCheckListBox): Integer;
var
  i: Integer;
begin
  Result := 0;
  For i := 0 to chkdLstBx.Items.Count - 1 Do
  begin
    If chkdLstBx.Checked[i] and (pos('[OPEN', UpperCase(chkdLstBx.Items[i])) <> 0) Then // DONE: LMS New method to handle disabled but open exams
      Result := Result + 1;
  end;                                                                          { End For Loop }
end;

function GetCheckedOpenOneVendorCount(chkdLstBx: TCheckListBox): Integer;
var
  i, tPos: Integer;
  aExam, V1, V2: string;
begin
  Result := 0;
  For i := 0 to chkdLstBx.Items.Count - 1 Do
  begin
    aExam := UpperCase(chkdLstBx.Items[i]);
    tPos := pos('[OPEN', aExam);
    if chkdLstBx.Checked[i] and (tPos <> 0) then
    begin
      V1 := copy(aExam, tPos + 8, 99);
      if (V2 <> V1) and (V2 <> '') then begin
        Result := 0;                                                            // DONE: LMS different vendor (not allowed). Temporary hack to get vendor name from the checkbox list. Ideally should have and compare vendor IEN's. Best added to the object in the list in the method TExntdCntrctdBaseFormNewRsnd.PopExamList in this call -> chkdLstBxExams.Items.AddObject
        exit;
      end else begin
        V2 := V1;                                                               // vendor for this exam
        Result := Result + 1;
      end;
    end;
  end;                                                                          { End For Loop }
end;

{  Determines if the IEN Value exists in the ComboBox Records.
   Input: fmCmbBxObj - TFMComboBox to search through for IEN value.
          IENVal - IEN value to find in the ComboBox Object.
   Output: Index value where item resides or -1 if item cannot be found }

function GetItemIndex(fmCmbBxObj: TFMComboBox; IENVal: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to fmCmbBxObj.Items.Count - 1 do
  begin
    fmCmbBxObj.ItemIndex := i;
    if SameText(fmCmbBxObj.GetSelectedRecord.IEN, IENVal) then
    begin
      Result := i;
      break;
    end;                                                                        { End If IEN Value found in Combobox Object }
  end;                                                                          { End For Loop through Combobox Contents }
end;

{-----------------------------------------------------------------------------
  Function:  GetTempDir
  Author:    From http://www.latiumsoftware.com/en/delphi/00017.php
  Date:      24-Feb-2004
  Arguments: None
  Result:    TFileName
  Note:      Refactored out from CAPRI MAIN.pas and TODO a few other places
-----------------------------------------------------------------------------}

Function GetTempDir: TFileName;
type
  KNOWNFOLDERID = TGuid;
  TSHGetKnownFolderPath = function(const rfid: KNOWNFOLDERID; dwFlags: DWord; hToken: THandle; var ppSzPath: LPWSTR): HResult; StdCall;
const
  FOLDERID_LocalAppData: KNOWNFOLDERID = '{F1B32785-6FBA-4FCF-9D55-7B8E7F157091}';

  {The next two functions will setup the call and return the result.  After that the Memory will be freeand Nil}
  function ShGetKnownFolderPath(const rfid: KNOWNFOLDERID; dwFlags: DWord; hToken: THandle; var ppszPath: LPWSTR): HResult;
  var
    Shell: HModule;
    SHGetKnownFolderPath: TSHGetKnownFolderPath;
  begin
    Shell := LoadLibrary('shell32.dll');
    Win32Check(Shell <> 0);
    try
      @SHGetKnownFolderPath := GetProcAddress(Shell, 'SHGetKnownFolderPath');
      Win32Check(Assigned(SHGetKnownFolderPath));
      Result := SHGetKnownFolderPath(rfid, dwFlags, hToken, ppszPath);
    finally
      FreeLibrary(Shell);
    end;
  end;

  function GetKnownFolderPath(const rfid: KNOWNFOLDERID; dwFlags: DWord; hToken: THandle): WideString;
  var
    buffer: LPWSTR;
  begin
    buffer := nil;
    OleCheck(ShGetKnownFolderPath(rfid, dwFlags, hToken, buffer));
    try
      Result := buffer;
    finally
      CoTaskMemFree(buffer);
    end;
  end;

Var
  TmpDir: Array[0..MAX_PATH - 1] Of char;
Begin
  Try
    Result := GetKnownFolderPath(FOLDERID_LocalAppData, 0, 0) + '\CAPRI\';      //rra 819092 Find path to Local App Data
    If not DirectoryExists(Result) then                                         //rra 819092 Check if CAPRI Dir exists
      If not CreateDir(Result) then                                             //rra 819092
      Begin                                                                     //rra 819092
        SetString(Result, TmpDir, GetTempPath(MAX_PATH, TmpDir));
        If Not DirectoryExists(Result) Then
          If Not CreateDirectory(PChar(Result), Nil) Then
          Begin
            Result := IncludeTrailingPathDelimiter(GetWindowsDir) + 'TEMP';
            If Not DirectoryExists(Result) Then
              If Not CreateDirectory(Pointer(Result), Nil) Then
              Begin
                Result := ExtractFileDrive(Result) + '\TEMP';
                If Not DirectoryExists(Result) Then
                  If Not CreateDirectory(Pointer(Result), Nil) Then
                  Begin
                    Result := ExtractFileDrive(Result) + '\TMP';
                    If Not DirectoryExists(Result) Then
                      If Not CreateDirectory(Pointer(Result), Nil) Then
                      Begin
                        Raise Exception.Create(SysErrorMessage(GetLastError));
                      End;
                  End;
              End;
          End;
      End;                                                                      //rra 819092
  Except
    Result := '';
    Raise;
  End;
End;

(*
function GetWindowsTempDir: String;
var
  i: Integer;
begin
  SetLength(Result, MAX_PATH);
  GetTempPath(MAX_PATH, PChar(Result));
  Result := String(AnsiString(Result));
  i := 1;
  while (MidStr(Result, i, 1) <> #0) do   // LMS - remove any trailing nul chars.. sometimes these are returned by GetTempPath. All unneccessary..
    Inc(i);
  Result := LeftStr(Result, i - 1);
end;
*)


{ Note: COPIED from CAPRI CAPRISupport.pas Line #356+ }

function GetWindowsDir: String;
const
  (* The length of the directoy buffer. Usually 64 or even 32 is enough:)
  **
  ** Must be DWORD type.
  *)
  dwLength: WORD = 255;
var
  pcSysDir: PChar;
begin
  GetMem(pcSysDir, dwLength);
  GetWindowsDirectory(pcSysDir, dwLength);
  Result := String(pcSysDir);
  FreeMem(pcSysDir, dwLength);
end;

function GetRgstryVal(KeyLoc: String; Key: String): String;
var
  Rgstry: TRegistry;
begin
  Result := '';                                                                 { Does not exist or could not be accessed }
  Rgstry := TRegistry.Create;
  try
    Rgstry.RootKey := HKEY_CURRENT_USER;                                        { Section within Registry to Retrieve }
    Rgstry.Access := KEY_READ;
    if Rgstry.OpenKey(KeyLoc, False) then
      Result := Rgstry.ReadString(Key);
  finally
    Rgstry.Free;
  end;                                                                          { End Try/Finally Block }
end;

function SetRgstryVal(KeyLoc: String; Key: String; KeyVal: String): Boolean;
var
  Rgstry: TRegistry;
begin
  Rgstry := TRegistry.Create;
  Result := True;
  try
    { HKEY_LOCAL_MACHINE can't be used because of user's restrictions }
    { Central File Store would be better so that values don't need to be
      replicated for each machine/user that accesses the software at the site. }
    Rgstry.RootKey := HKEY_CURRENT_USER;                                        { Section within Registry to Store }
    Rgstry.Access := KEY_WRITE;
    if Rgstry.OpenKey(KeyLoc, True) then
      Rgstry.WriteString(Key, KeyVal)
    else
      Result := False;
  finally
    Rgstry.Free;
  end;                                                                          { End Try/Finally Block }
end;

function DeleteRstryVal(KeyLoc: String): Boolean;
var
  Rgstry: TRegistry;
begin
  Rgstry := TRegistry.Create;
  try
    Rgstry.RootKey := HKEY_CURRENT_USER;                                        { Section within Registry to Retrieve }
    Rgstry.Access := KEY_WRITE;
    Result := Rgstry.DeleteKey(KeyLoc);
  finally
    Rgstry.CloseKey;
    Rgstry.Free;
  end;                                                                          { End Try/Finally Block }
end;

function ProcessExists(exeName: String): Boolean;
var
  ContinueLoop: BOOL;
  FSnapshotHandle: THandle;
  FProcessEntry32: TProcessEntry32;
begin
  FSnapshotHandle := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  FProcessEntry32.dwSize := SizeOf(FProcessEntry32);
  ContinueLoop := Process32First(FSnapshotHandle, FProcessEntry32);
  Result := False;
  while (Integer(ContinueLoop) <> 0) do
  begin
    if ((UpperCase(ExtractFileName(FProcessEntry32.szExeFile)) =
      UpperCase(exeName)) or (UpperCase(FProcessEntry32.szExeFile) =
      UpperCase(exeName))) then
      Result := True;
    ContinueLoop := Process32Next(FSnapshotHandle, FProcessEntry32);
  end;                                                                          { End While Loop }
  CloseHandle(FSnapshotHandle);
end;

function StringToArray(const strVal: String; const dlmtr: String): TStrArray;
var
  i, numOcc: Integer;
begin
  SetLength(Result, 0);
  numOcc := Occurrences(dlmtr, strVal);
  if (numOcc > 0) then
  begin
    SetLength(Result, numOcc);
    for i := 1 to numOcc do
      Result[(i - 1)] := Piece(strVal, dlmtr, i);
  end;                                                                          { End If Dlmtr Occurrences is NOT 0 }
end;

function ValidateTypeAndInstance(aObj: Pointer; ofClass: TClass): Boolean;
begin
  Result := false;
  try
    Result := TObject(aObj^).InheritsFrom(ofClass);                             // InheritsFrom is a class method, so even if self was "not assigned" in the object the method would likely return properly.
  except
  end;
end;

{
LMS - Utility method to memory manage the list with its objects.
      Won't work on weird typecast stuff one could shove into the list as
        objects, when they really aren't a TObject descendant.
      Wont' work on list objects have special requirements in clean up,
        i.e. those that don't clean themselves up when their free/destroy method as TObject is called.
}

procedure EmptyListAndFreeObjects(aList: TList);
var
  ListIx: Integer;
  aObj: TObject;
  initialCount: Integer;
  selfResponsible: boolean;
  excepMsg: string;
begin
  if not ValidateTypeAndInstance(@aList, TList) then exit;
  try
    initialCount := aList.Count;
  except
    exit;
  end;
  for ListIx := initialCount - 1 downto 0 do begin
    try
      aObj := aList[ListIx];
      if assigned(aObj) then
        if ValidateTypeAndInstance(@aObj, TObject) then begin                   // new methods to avoid exceptions on invalid objects
          selfResponsible := true;
          if ValidateTypeAndInstance(@aObj, TComponent) then
            if TComponent(aObj).Owner <> nil then
              selfResponsible := false
            else if ValidateTypeAndInstance(@aObj, TControl) then
              if TControl(aObj).Parent <> nil then
                selfResponsible := false;
          if selfResponsible then begin
            if ValidateTypeAndInstance(@aObj, TCustomForm) then
              TCustomForm(aObj).Release
            else if ValidateTypeAndInstance(@aObj, TComponent) then
              TComponent(aObj).Free
            else
              aObj.Free;
          end;
        end;
    except
      on Excp: Exception do begin
        // if you got here, you must be a memory leak (Jeff Foxworthy)
      end;
    end;
    try
      aList.Items[ListIx] := nil;                                               // aList.Delete(ListIx);
    except
      on Excp: Exception do begin
        // if your cousin got here, you must be a memory leak (Jeff Foxworthy)
      end;
    end;
  end;
  try
    aList.Clear;
  except
    on Excp: Exception do begin
      // if your uncle got here, you must be a memory leak (Jeff Foxworthy)
    end;
  end;
  try
    aList.Free;
  except
    on Excp: Exception do begin
      // if your aunt got here, you must be a memory leak (Jeff Foxworthy)
    end;
  end;
end;

{
LMS - Utility method to memory manage the list with its objects and nil the pointer passed in.
      This method should be called mostly.
      If you ever have to leave the list address untouched, only then call EmptyListAndFreeObjects directly.
}

procedure FreeAndNilListWithObjects(var aList: TList);
begin
  EmptyListAndFreeObjects(aList);
  aList := nil;
end;

// A similar method could be made for TLists if needed.

procedure FreeAndClearTStrings(aList: TStrings);
var
  excepMsg: string;
begin
  try
    aList.Clear;
  except
    on Excp: Exception do
    begin
      // if your uncle got here, you must be a memory leak (Jeff Foxworthy)
{$IFDEF CapriDeveloperLeosCode}
      excepMsg := 'Error: ' + Excp.message + '.' + ''#13'' #10'' + 'Error occurred while clearing the List. List Class: ' + aList.ClassName + '. List Count: ' + IntToStr(aList.Count);
      ShowMessageCAPRI(excepMsg);
{$ENDIF}
    end;
  end;
  try
    aList.Free;
  except
    on Excp: Exception do
    begin
      // if your aunt got here, you must be a memory leak (Jeff Foxworthy)
{$IFDEF CapriDeveloperLeosCode}
      excepMsg := 'Error: ' + Excp.message + '.' + ''#13'' #10'' + 'Error occurred while freeing a List. List Class: ' + aList.ClassName;
      ShowMessageCAPRI(excepMsg);
{$ENDIF}
    end;
  end;
end;

procedure EmptyTStringsAndFreeObjects(aList: TStrings);
var
  ListIx: Integer;
  aObj: TObject;
  initialCount: Integer;
  selfResponsible: boolean;
  excepMsg: string;
begin
  if not ValidateTypeAndInstance(@aList, TList) then exit;
  try
    initialCount := aList.Count;
  except
    exit;
  end;
  for ListIx := initialCount - 1 downto 0 do begin
    try
      aObj := aList.Objects[ListIx];
      if assigned(aObj) then
        if ValidateTypeAndInstance(@aObj, TObject) then begin                   // new methods to avoid exceptions on invalid objects
          selfResponsible := true;
          if ValidateTypeAndInstance(@aObj, TComponent) then
            if TComponent(aObj).Owner <> nil then
              selfResponsible := false
            else if ValidateTypeAndInstance(@aObj, TControl) then
              if TControl(aObj).Parent <> nil then
                selfResponsible := false;
          if selfResponsible then begin
            if ValidateTypeAndInstance(@aObj, TCustomForm) then
              TCustomForm(aObj).Release
            else if ValidateTypeAndInstance(@aObj, TComponent) then
              TComponent(aObj).Free
            else
              aObj.Free;
          end;
        end;
    except on Excp: exception do
      // if you got here, you must be a memory leak (Jeff Foxworthy)
    end;
    try
      aList.Objects[ListIx] := nil;                                             // aList.Delete(ListIx);
    except on Excp: exception do
      // if you got here, you must be a memory leak (Jeff Foxworthy)
    end;
  end;
  FreeAndClearTStrings(aList);
end;

procedure FreeAndNilTStringsWithObjects(var aList: TStrings);
begin
  EmptyTStringsAndFreeObjects(aList);
  aList := nil;
end;

function VersionInfo(P: TVersionParam): String;
var
  Exe: String;
  BuffSize, Len: Cardinal;
  //i: Integer;       not used
  Buff, Value: PChar;
const
  PItems: array[0..9] of String = ('CompanyName', 'FileDescription', 'FileVersion', 'InternalName',
    'LegalCopyright', 'LegalTradeMarks', 'OriginalFilename',
    'ProductName', 'ProductVersion', 'Comments');
begin
  Exe := Application.ExeName;
  BuffSize := GetFileVersionInfoSize(PChar(Exe), BuffSize);
  Result := '';
  if BuffSize > 0 then
  begin
    Buff := AllocMem(BuffSize);
    GetFileVersionInfo(PChar(Exe), 0, BuffSize, Buff);
    if VerQueryValue(Buff, PChar('StringFileInfo\040904E4\' + PItems[Ord(P)]), Pointer(Value), Len)
    then Result := Value;                                                       { Hash Table of Version Info }
    FreeMem(Buff, BuffSize);
  end                                                                           { End If Version Info Found }
end;

function NumStr(X: String): String;
var
  i: Integer;
  C: String;
begin
  Result := '';
  i := 1;
  while (i <= Length(X)) do
  begin
    C := MidStr(X, i, 1);
    if (C >= '0') and (C <= '9') then
      Result := Result + C;
    Inc(i);
  end;
end;

procedure FormatXMLDoc(xmlFile: String);
var
  lDoc: IXMLDocument;
begin
  lDoc := TXMLDocument.Create(nil);
  try
    lDoc.LoadFromFile(xmlFile);
    lDoc.Encoding := 'ISO-8859-1';
    lDoc.XML.Text := xmlDoc.FormatXMLData(lDoc.XML.Text);
    lDoc.Active := True;
    lDoc.SaveToFile(xmlFile);
  except
{$IFDEF CAPRIDEVELOPER}
    on e: exception do                                                          // suppress all errors
      showmessageCAPRI('Error Formatting XML File: ' + xmlFile + #13#10 + e.Message);
{$ENDIF}
  end;
end;

{
Utility method to help with debugging at run time.
Inputs:
  A string to load into an XMLDocument object.
  Prefix of file name to use for saving, overriding and writing out final loaded version.
  Suffix of this file if needed.

Result:
  The specified XML string is saved to file in temp folder.
  Then if an override file exists for debugging, it replaces the contents of the XML string.
  Then the string is attempted to load into XMLDocument.
     Here if an exception occurs, developer in debug mode can still provide an override file on disk and it will be attempted to read and load one more time.
  Finally whatever was loaded into XMLDocument is written back out if SAVEXML was specified.
}

procedure SaveStrXMLInFileAndXMLDoc(var strXml: string; filePrefix: string; fileSuffix: string; AXMLDom: TXMLDocument; SaveLoaded: Boolean = false);
begin
  SaveStrXMLInFileAndXMLDoc(strXml, '', filePrefix, fileSuffix, AXMLDom, SaveLoaded);
end;

procedure SaveStrXMLInFileAndXMLDoc(var strXml: string; TempSubFolder: string; filePrefix: string; fileSuffix: string; AXMLDom: TXMLDocument; SaveLoaded: Boolean = false);
var
  aStL: TStringList;
  fileOriginal, fileLoaded, fileOverride: string;
  savePath: string;
begin
  savePath := GetTempDir;
  if TempSubFolder <> '' then begin
    savePath := savePath + TempSubFolder + '\';
    EnsureDirectoryExists(savePath);
  end;

  fileOriginal := savePath + filePrefix + 'Original' + fileSuffix + '.xml';
  fileOverride := savePath + filePrefix + 'OverRide' + fileSuffix + '.xml';
  fileLoaded := savePath + filePrefix + 'Loaded' + fileSuffix + '.xml';
{$IFDEF CAPRIDEVELOPER}
  aStl := TStringList.Create;
  aStl.text := strXML;
  aStl.SaveToFile(fileOriginal);
  FormatXMLDoc(fileOriginal);
  if fileexists(fileOverride) then
  begin
    aStL.LoadFromFile(fileOverride);
    strXML := aStl.Text;
  end;
{$ENDIF}
  try
    AXMLDom.LoadFromXML(strXml);
  except on e: exception do
{$IFDEF CAPRIDEVELOPER}
  // here get a chance to load a corrected version
      if fileexists(fileOverride) then begin
        aStL.LoadFromFile(fileOverride);
        strXML := aStl.Text;
        AXMLDom.LoadFromXML(strXml);
      end;
{$ENDIF}
  end;
  AXMLDom.Active := True;
  if SaveLoaded then begin
    AXMLDom.SaveToFile(fileLoaded);
    FormatXMLDoc(fileLoaded);
  end;
end;

// Purpose is to generically get a Node's text value, when exact type of node (or if it has child objects/attributes etc) is not known
//  Specifically useful for DateTime type fields.

function GenericGetXMLNodeText(aNode: IXMLNode; aNodeName: string = ''): string;
begin
{
    mnDateNode := rNode.ChildNodes['ManifestDate'];
    //mnDateTime := mnDateNode.Text;  This doesn't work
    mnDateTime := mnDateNode.ChildNodes[0].NodeValue;   // This does

    examValue := xNode.NodeValue;

    SSN := pNode.ChildNodes['SSN'].NodeValue;           // Text??

      TNodeType = (ntReserved, ntElement, ntAttribute, ntText, ntCData,
    ntEntityRef, ntEntity, ntProcessingInstr, ntComment, ntDocument,
    ntDocType, ntDocFragment, ntNotation);

}
  // Different versions of XML have yielded different results. This method tries to handle the above listed variations.

  try
    if (aNode.IsTextElement) then
      result := aNode.Text
    else if aNode.NodeType = ntText then begin
      result := aNode.Text;
      if (copy(result, 1, 1) = '#') and (copy(result, 2, maxint) = aNode.NodeValue) then // a scenario where .text results in a # prefixed nodevalue
        result := aNode.NodeValue;
    end;
  except
    try
      result := aNode.NodeValue
    except
      result := aNode.ChildNodes[0].NodeValue                                   // accorrding to help file, some nodes can have a single child that is a DOM text node. Some Date fields did exhibit this behavior.
      // any exception at this point is bubbled up.
    end;
  end;
end;

function FindNodeByNodeName(rootLevel: IXMLNode; NodeNameToFind: string): IXMLNode;
// DONE: LMS - refactor out
var
  iii: Integer;
begin
  for iii := 0 to rootLevel.ChildNodes.count - 1 do
  begin
    Result := rootLevel.ChildNodes[iii];
    if Result.NodeName = NodeNameToFind then
      exit;
  end;
  Result := nil;                                                                // we didn't find it.
end;

{
  Generic routine to find any child node by full path starting at a root level.
  Example path : 'cld:Claim^cld:DocumentTypeVersion^cld:ClaimID^cld:CommonData^nc:Document'
    if nodes have ^ in them, then use a different separator character or even a long Unique String
}

function FindNodeInPathByNodeName(rootLevel: IXMLNode; PathAndNodeName: string; separator: string = '^'): IXMLNode;
// DONE: LMS - refactor out
var
  iii: Integer;
  aNodeName: string;
  curLevel: IXMLNode;
begin
  Result := nil;
  curLevel := rootLevel;
  iii := 1;
  repeat
    aNodeName := MFunStr.piece(PathAndNodeName, separator, iii);
    if aNodeName = '' then break;
    curLevel := FindNodeByNodeName(curLevel, aNodeName);
    if curLevel = nil then break;                                               // we didn't find it.
    inc(iii);
  until false;                                                                  // aNodeName = '';
  Result := curLevel;
end;

function FindOrInsertNodeInPathByNodeName(rootLevel: IXMLNode; PathAndNodeName: string; separator: string = '^'): IXMLNode;
// DONE: LMS - refactor out
var
  aLevelsRemain: Integer;
  PathAndNodeNameOneUp: string;
  ix: Integer;
  maxCount: Integer;
  curIxLevel: IXMLNode;
begin
  curIxLevel := rootLevel;
  maxCount := untAddtnlMMthds.Occurrences(separator, PathAndNodeName);
  for ix := 1 to maxCount do begin
    curIxLevel := FindNodeByNodeName(rootLevel, piece(PathAndNodeName, separator, ix));
    if curIxLevel = nil then begin
      curIxLevel := rootLevel.AddChild(piece(PathAndNodeName, separator, ix));
    end;
    rootLevel := curIxLevel;
  end;
  Result := curIxLevel;
end;

procedure AddChildWithTextValue(aAttachNode: IXMLNode; cNodeName: string; cNodeText: string = '');
var
  aChildNode: IXMLNode;
begin
  aChildNode := aAttachNode.AddChild(cNodeName);
  aChildNode.Text := cNodeText;
end;

procedure UpdateChildWithTextValue(aAttachNode: IXMLNode; cNodeName: string; cNodeText: string = '');
var
  aChildNode: IXMLNode;
begin
  aChildNode := FindOrInsertNodeInPathByNodeName(aAttachNode, cNodeName);
  aChildNode.ChildNodes.Clear;
  aChildNode.Text := cNodeText;
end;

procedure AddDocumentsWrapperIfMissing(var rootLevel: TXMLDocument);
var
  topLevel: IXMLNode;
  aStream: TStringStream;
  aStr: string;
begin
  if rootLevel.DocumentElement.NodeName = 'documents' then begin
    exit;
  end;
  if rootLevel.DocumentElement.NodeName = 'document' then begin
    aStr := '<documents>'#13#10;
    aStr := aStr + rootLevel.DocumentElement.XML + #13#10'</documents>';
    aStream := TStringStream.Create(aStr);
    try
      rootLevel.ChildNodes.Clear;
      rootLevel.Resync;
      rootLevel.LoadFromStream(aStream);
      rootLevel.Active := true;
    finally
      aStream.Free;
    end;
  end;
end;

function FormattedTimeStamp(ATimeStamp: TDateTime): String;
begin
  Result := FormatDateTime('YYMMDDhhnnsszzz', ATimeStamp);
  //Result := DateTimeTo    anDate(ATimeStamp);  // too clunky for users.
end;

procedure SaveStrToFle(const FileName: TFileName; const content: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    Stream.Write(content[1], Length(content));
  finally
    Stream.Free;
  end;                                                                          { End Try/Finally Block }
end;

procedure QuickCopy(AFrom, ATo: TObject);                                       { Copied Out of ORFn }
var
  ms: TMemoryStream;
  idx: integer;
  str: array[0..1] of TStrings;
  fix: array[0..1] of boolean;

  procedure GetStrings(obj: TObject);
  begin
    if (CompareText(obj.ClassName, 'TRichEditStrings') = 0) then
      raise QuickCopyError.Create('You must pass the TRichEdit object into QuickCopy, NOT it''s Lines property.');
    if obj is TStrings then
      str[idx] := TStrings(obj)
    else
      if obj is TMemo then
        str[idx] := TMemo(obj).Lines
      else
        if obj is TListBox then
          str[idx] := TListBox(obj).Items
        else
          if obj is TRichEdit then
          begin
            with TRichEdit(obj) do
            begin
              str[idx] := Lines;
              if not PlainText then
              begin
                fix[idx] := TRUE;
                PlainText := TRUE;
              end;
            end;
          end
          else
            raise QuickCopyError.Create('Unsupported object type (' + obj.ClassName +
              ') passed into QuickCopy.');
    inc(idx);
  end;                                                                          { End Procedure GetStrings }
begin
  fix[0] := FALSE;
  fix[1] := FALSE;
  idx := 0;
  GetStrings(AFrom);
  GetStrings(ATo);
  ms := TMemoryStream.Create;
  try
    str[0].SaveToStream(ms);
    ms.Seek(0, soFromBeginning);
    str[1].LoadFromStream(ms);
  finally
    ms.Free;
  end;
  if fix[0] then TRichEdit(AFrom).PlainText := FALSE;
  if fix[1] then TRichEdit(ATo).PlainText := FALSE;
end;

function FMDTtoExternal(FMDT: String): String;
var
  x: String;
begin
  Result := '';
  if (Length(FMDT) > 0) then
  begin
    Result := MidStr(FMDT, 4, 2) + '/' + MidStr(FMDT, 6, 2) + '/';
    x := IntToStr(StrToInt(MidStr(FMDT, 1, 3)) + 1700);
    Result := Result + x;
  end;
end;

function GetDivisionOfCBOC(CBOC: string): string;
var
  i, Count: Integer;
  tCBOC, Tstr: string;
begin
  // DONE: LMS (may be later) Find any other code that's already written in CAPRI or in a lib that gets 'CBOC more definitively.
  tCBOC := Piece(CBOC, '^', 3);
  SetLength(Tstr, Length(tCBOC));
  Count := 0;
  for i := 1 to Length(tCBOC) do begin
    if (tCBOC[i] >= '0') and (tCBOC[i] <= '9') then begin
      inc(Count);
      Tstr[Count] := tCBOC[i];
    end else break;                                                             // only need the first numerics. Stop at the first non-numeric char encountered.
    // end result is 500A3 is returned as 500
  end;
  SetLength(Tstr, Count);

  // TODO -cNextRelease: LMS wait 500A3 happened to be not a valid value in divisions, institution files etc. But it is coming from ^DIC(4,24,99)="500A3^^SAMUEL STRATTON VA^n^V^3^1". Not sure yet how the broker is getting that value during sign-on.

  (* DONE: LMS as of 2013-07-16 as per George Kocis the value 500A3 etc. is station + modifiers.
                       ->  This is the field description from the data dictionary:
                       ->  DESCRIPTION:      This is the 3 digit station number plus any modifiers.
               So changing the code to just return first 3 chars (which should be numeric) of the CBOC identifier.
  *)
  Result := copy(Tstr, 0, 3);

  // DONE: LMS A count of 0 would be an error.
  if (length(result) < 1) then
    Raise Exception.CreateFmt('Invalid Division/CBOC "%s" supplied.%sFirst 3 letter have to be numbers.%sPlease contact your IRM.', [CBOC, #13#10, #13#10]);
end;

function IsDateValid(aDateStr: string): Boolean;
var
  aDate: TDateTime;
begin
  // DONE: LMS more efficient implementation
  Result := TryStrToDate(aDateStr, aDate);
end;

procedure EnsureDirectoryExists(dirName: String);
begin
  if (Not DirectoryExists(dirName)) then
  begin
    { Create SubDirectory for File Attachements }
    CreateDir(dirName);
    if (Not DirectoryExists(dirName)) then
      ForceDirectories(dirName);
  end;                                                                          { End If Directory does not exists }
end;

function GetTempExmDir(DUZ: String): String;
begin
  Result := GetTempDir;                                                         { Windows Temp Directory }
  Result := Result + 'demTranTemp\' + DUZ;
  EnsureDirectoryExists(Result);
end;

function MakeOutputFileStreamReady(ResponseOutputFile: string): TFileStream;
begin
  if fileExists(ResponseOutputFile) then
    Result := TFileStream.Create(ResponseOutputFile, fmOpenReadWrite)
  else
    Result := TFileStream.Create(ResponseOutputFile, fmCreate);
  Result.Seek(0, soFromEnd);
  //ResponseToFile.Seek(1,soFromCurrent);
end;

{===============================================================================
  GetStringGUID
    GetStringGUID is a function that generates a formatted GUID

  Return values
    GetStringGUID returns a string representing the formatted GUID

  Modification History:
    patch 186 - Method created 7/2013 -MER
    patch 187 - Method moved to untMiscMthds 7/2014 -LMS Also added status check.
 ==============================================================================}

function GetStringGUID: string;
var
  lGUID: TGUID;
  lGUIDStr: string;
begin
  Result := '';
  if CreateGUID(lGUID) = S_OK then begin
    lGUIDStr := GUIDToString(lGUID);
    lGUIDStr := StringReplace(lGUIDStr, '{', '', [rfReplaceAll, rfIgnoreCase]);
    lGUIDStr := StringReplace(lGUIDStr, '}', '', [rfReplaceAll, rfIgnoreCase]);
    Result := lGUIDStr;
  end;
end;

// from ORFn but corrected per Delphi TControl.BoundsRect help
function RectContains(Rect: TRect; Point: TPoint): boolean;
begin
  Result := ((Point.X >= Rect.Left) and
    (Point.X < Rect.Right) and
    (Point.Y >= Rect.Top) and
    (Point.Y < Rect.Bottom));
end;                                                                            // RectContains

function ScrollBox(inBox: TWinControl; inPoint: TPoint;
  WheelDelta: Integer): Boolean;
var
  childPt: TPoint;
  boundsRect: TRect;
  parentc: TWinControl;
  i, code: Integer;
begin
  Result := False;
  code := 0;
  parentc := inBox.Parent;
  // is the mouse over the scrollable box?
  if (Parentc <> nil) and (inBox <> nil) then begin
    childPt := Parentc.ScreenToClient(inPoint);
    boundsRect := inBox.BoundsRect;
    If RectContains(boundsRect, childPt) then begin
      if WheelDelta > 0 then
        code := SB_LINEUP
      else
        code := SB_LINEDOWN;
      { for I := 0 to Mouse.WheelScrollLines - 1 do
        inBox.Perform(WM_VSCROLL, code, 0);
      inBox.Perform(WM_VSCROLL, SB_ENDSCROLL, 0); }
      { for I := 0 to Mouse.WheelScrollLines - 1 do
        SendMessage(inBox.Handle, WM_VSCROLL, code, 0);
      SendMessage(inBox.Handle, WM_VSCROLL, SB_ENDSCROLL, 0); }
      for I := 0 to Mouse.WheelScrollLines - 1 do
        PostMessage(inBox.Handle, WM_VSCROLL, code, 0);
      PostMessage(inBox.Handle, WM_VSCROLL, SB_ENDSCROLL, 0);
//      inBox.ScrollBy(0, WheelDelta);
      inBox.Invalidate;
      Result := True;
    end;
  end;
end;                                                                            // ScrollBox

end.

